// -*- Funnel -*-

/**
 * $Author: ysaidji $
 * $Id: AVLTree.m4,v 1.2 2001/09/12 13:25:12 ysaidji Exp $
 */

def removeMe0 = include[AVLTreeType]
def removeMe1 = include[__AssocDefaultType__]
def removeMe2 = include[__AssocDefault__]

def AVLTree[k] (equ: Equ[k], ord: Ord[k]) : AVLTreeModule[k] = {

  let this = {
    // private values and methods
    val LEFT = -1
    val BALANCED = 0
    val RIGHT = 1

    def rd[a] (t:AVLTree[k,a]) : AVLTree[k,a] = {
      t.match[AVLTree[k,a]] {
	def Nil = System.error "AVLTree.rd: rd failed [internal error]"
	def Tree(lf,(kq,vq,dq),w) = {
	  lf.match[AVLTree[k,a]] {
	    def Nil = System.error "AVLTree.rd: rd failed [internal error]"
	    def Tree(u,(kp,vp,dp),v) = {
	      if(dp == BALANCED) this.Tree[a](u,(kp,vp,RIGHT),this.Tree[a](v,(kq,vq,LEFT),w))
	      else {
		if (dp == LEFT) this.Tree[a](u,(kp,vp,BALANCED),this.Tree[a](v,(kq,vq,BALANCED),w))
		else System.error "AVLTree.rd: rd failed [internal error]"
	      }
	    }
	  }
	}
      }
    }

    def rg[a] (t:AVLTree[k,a]) : AVLTree[k,a] = {
      t.match[AVLTree[k,a]] {
	def Nil = System.error "AVLTree.rg: rg failed [internal error]"
	def Tree(u,(kp,vp,dp),rg) = {
	  rg.match[AVLTree[k,a]] {
	    def Nil = System.error "AVLTree.rg: rg failed [internal error]"
	    def Tree(v,(kq,vq,dq),w) = {
	      if(dq == BALANCED) this.Tree[a](this.Tree[a](u,(kp,vp,RIGHT),v),(kq,vq,LEFT),w)
	      else {
		if(dq == RIGHT) this.Tree[a](this.Tree[a](u,(kp,vp,BALANCED),v),(kq,vq,BALANCED),w)
		else System.error "AVLTree.rg: rg failed [internal error]"
	      }
	    }
	  }
	}
      }
    }

    def rgd[a] (t:AVLTree[k,a]) : AVLTree[k,a] = {
      t.match[AVLTree[k,a]] {
	def Nil = System.error "AVLTree.rgd: rgd failed [internal error]"
	def Tree(lf,(kr,vr,dr),w) = {
	  lf.match[AVLTree[k,a]] {
	    def Nil = System.error "AVLTree.rgd: rgd failed [internal error]"
	    def Tree(t',(kp,vp,dp),rg) = {
	      rg.match[AVLTree[k,a]] {
		def Nil = System.error "AVLTree.rgd: rgd failed [internal error]"
		def Tree(u,(kq,vq,dq),v) = {
		  if(dq == BALANCED) this.Tree[a](this.Tree[a](t',(kp,vp,BALANCED),u),(kq,vq,BALANCED),this.Tree[a](v,(kr,vr,BALANCED),w))
		  else {
		    if(dq == RIGHT) this.Tree[a](this.Tree[a](t',(kp,vp,LEFT),u),(kq,vq,BALANCED),this.Tree[a](v,(kr,vr,BALANCED),w))
		    else this.Tree[a](this.Tree[a](t',(kp,vp,BALANCED),u),(kq,vq,BALANCED),this.Tree[a](v,(kr,vr,RIGHT),w))
		  }
		}
	      }
	    }
	  }
	}
      }
    }

    def rdg[a] (t:AVLTree[k,a]) : AVLTree[k,a] = {
      t.match[AVLTree[k,a]] {
	def Nil = System.error "AVLTree.rdg: rdg failed [internal error]"
	def Tree(t',(kr,vr,dr),rg) = {
	  rg.match[AVLTree[k,a]] {
	    def Nil = System.error "AVLTree.rdg: rdg failed [internal error]"
	    def Tree(lf,(kp,vp,dp),w) = {
	      lf.match[AVLTree[k,a]] {
		def Nil = System.error "AVLTree.rdg: rdg failed [internal error]"
		def Tree(u,(kq,vq,dq),v) = {
		  if(dq == LEFT) this.Tree[a](this.Tree[a](t',(kr,vr,BALANCED),u),(kq,vq,BALANCED),this.Tree[a](v,(kp,vp,RIGHT),w))
		  else {
		    if(dq == RIGHT) this.Tree[a](this.Tree[a](t',(kr,vr,LEFT),u),(kq,vq,BALANCED),this.Tree[a](v,(kp,vp,BALANCED),w))
		    else this.Tree[a](this.Tree[a](t',(kr,vr,BALANCED),u),(kq,vq,BALANCED),this.Tree[a](v,(kp,vp,BALANCED),w))
		  }
		}
	      }
	    }
	  }
	}
      }
    }

    val IncLeft = -1
    val NoInc = 0
    val IncRight = 1

    def insert_AVL[a] (f:k->a->a->a,k:k,x:a) (t:AVLTree[k,a]) : AVLTree[k,a] = {
      def insert_aux (t:AVLTree[k,a]) : (AVLTree[k,a],int) = {
	t.match[(AVLTree[k,a],int)] {
	  def Nil = (this.Tree[a](this.Nil[a],(k,x,BALANCED),this.Nil[a]),IncLeft)
	  def Tree(t1,(ka,va,da),t2) = {
	    if(equ.equ (ka) (k)) {
	      (this.Tree[a](t1,(ka,(f (ka) (x) (va)),da),t2),NoInc)
	    }
	    else {
	      if(ord.lt (k) (ka)) {
		val (t'1,info)=insert_aux (t1)
		if(da == BALANCED) {
		  if(info == NoInc) (this.Tree[a](t'1,(ka,va,BALANCED),t2),NoInc)
		  else (this.Tree[a](t'1,(ka,va,LEFT),t2),IncLeft)
		}
		else {
		  if(da == LEFT) {
		    if(info == NoInc) (this.Tree[a](t'1,(ka,va,LEFT),t2),NoInc)
		    else {
		      if(info == IncLeft) (rd[a] (this.Tree[a](t'1,(ka,va,BALANCED),t2)),NoInc)
		      else (rgd[a] (this.Tree[a](t'1,(ka,va,BALANCED),t2)),NoInc)
		    }
		  }
		  else {
		    if(info == NoInc) (this.Tree[a](t'1,(ka,va,RIGHT),t2),NoInc)
		    else (this.Tree[a](t'1,(ka,va,BALANCED),t2),NoInc)
		  }
		}
	      }
	      else {
		val (t'2,info)=insert_aux (t2)
		if(da == BALANCED) {
		  if(info == NoInc) (this.Tree[a](t1,(ka,va,BALANCED),t'2),NoInc)
		  else (this.Tree[a](t1,(ka,va,RIGHT),t'2),IncRight)
		}
		else {
		  if(da == LEFT) {
		    if(info == NoInc) (this.Tree[a](t1,(ka,va,LEFT),t'2),NoInc)
		    else (this.Tree[a](t1,(ka,va,BALANCED),t'2),NoInc)
		  }
		  else {
		    if(info == NoInc) (this.Tree[a](t1,(ka,va,RIGHT),t'2),NoInc)
		    else {
		      if(info == IncLeft) (rdg[a] (this.Tree[a](t1,(ka,va,BALANCED),t'2)),NoInc)
		      else (rg[a] (this.Tree[a](t1,(ka,va,BALANCED),t'2)),NoInc)
		    }
		  }
		}
	      }
	    }
	  }
	}
      }
      Predef.fst (insert_aux (t))
    }

    def bal_left[a] (ka:k,va:a,da:int) (t1:AVLTree[k,a]) (t2:AVLTree[k,a]) : AVLTree[k,a] = {
      t2.match {
	def Nil = System.error "AVLTree.bal_left: bal_left failed [internal error]"
	def Tree(_,(_,_,r),_) = {
	  if(r == LEFT) rdg[a] (this.Tree[a](t1,(ka,va,da),t2))
	  else rg[a] (this.Tree[a](t1,(ka,va,da),t2))
	}
      }
    }

    def bal_right[a] (ka:k,va:a,da:int) (t1:AVLTree[k,a]) (t2:AVLTree[k,a]) : AVLTree[k,a] = {
      t1.match {
	def Nil = System.error "AVLTree.bal_right: bal_right failed [internal error]"
	def Tree(_,(_,_,r),_) = {
	  if(r == RIGHT) rgd[a] (this.Tree[a](t1,(ka,va,da),t2))
	  else rd[a] (this.Tree(t1,(ka,va,da),t2))
	}
      }
    }

    def delete_greatest[a] (t:AVLTree[k,a]) : ((k,a,int),AVLTree[k,a],boolean) = {
      t.match[((k,a,int),AVLTree[k,a],boolean)] {
	def Nil = System.error "AVLTree.delete_greatest: delete_greatest failed [internal error]"
	def Tree(t1,(ka,va,da),t2) = {
	  if(t2.isEmpty) ((ka,va,da),t1,true)
	  else {
	    val (a',t'2,flag)=delete_greatest (t2)
	    if(da == BALANCED) {
	      if(flag) (a',this.Tree[a](t1,(ka,va,LEFT),t'2),false)
	      else (a',this.Tree[a](t1,(ka,va,da),t'2),false)
	    }
	    else {
	      if(da == RIGHT) {
		if(flag) (a',this.Tree[a](t1,(ka,va,BALANCED),t'2),true)
		else (a',this.Tree[a](t1,(ka,va,da),t'2), false)
	      }
	      else {
		if(flag) {
		  t1.match[((k,a,int),AVLTree[k,a],boolean)] {
		    def Nil = System.error "AVLTree.delete_greatest: delete_greatest failed [internal error]"
		    def Tree(_,(_,_,r),_) = {
		      (a',bal_right[a] (ka,va,da) (t1) (t'2),(r != BALANCED))
		    }
		  }
		}
		else (a',this.Tree[a](t1,(ka,va,da),t'2),false)
	      }
	    }
	  }
	}
      }
    }

    def delete_AVL[a] (k:k) (t:AVLTree[k,a]) : AVLTree[k,a] = {
      def delete_AVL_aux (t:AVLTree[k,a]) : (AVLTree[k,a],boolean) = {
	t.match[(AVLTree[k,a],boolean)] {
	  def Nil = (this.Nil[a],false) //System.error "AVLTree.delete_AVL: delete_AVL"
	  def Tree(t1,(ka,va,da),t2) = {
	    if(equ.equ (ka) (k)) {
	      if(t1.isEmpty) (t2,true)
	      else {
		if(t2.isEmpty) (t1,true)
		else {
		  val ((ka',va',da'),t'1,flag)=delete_greatest (t1)
		  if(flag) {
		    if(da == BALANCED) (this.Tree[a](t'1,(ka',va',RIGHT),t2),false)
		    else {
		      if(da == LEFT) (this.Tree[a](t'1,(ka',va',BALANCED),t2),true)
		      else {
			t2.match[(AVLTree[k,a],boolean)] {
			  def Nil = System.error "AVLTree.delete_AVL: delete_AVL[internal error]"
			  def Tree(_,(_,_,r),_) = (bal_left[a] (ka',va',da') (t'1) (t2),(r != BALANCED))
			}
		      }
		    }
		  }
		  else (this.Tree[a](t'1,(ka',va',da),t2),false)
		}
	      }
	    }
	    else {
	      if(ord.lt (k) (ka)) {
		val (t'1,flag)=delete_AVL_aux (t1)
		if(flag) {
		  if(da == BALANCED) (this.Tree[a](t'1,(ka,va,RIGHT),t2),false)
		  else {
		    if(da == LEFT) (this.Tree[a](t'1,(ka,va,BALANCED),t2),true)
		    else {
		      t2.match[(AVLTree[k,a],boolean)] {
			def Nil = System.error "AVLTree.delete_AVL: delete_AVL[internal error]"
			def Tree(_,(_,_,r),_) = (bal_left[a] (ka,va,da) (t'1) (t2),(r != BALANCED))
		      }
		    }
		  }
		}
		else (this.Tree[a](t'1,(ka,va,da),t2),false)
	      }
	      else {
		val (t'2,flag)=delete_AVL_aux (t2)
		if(flag) {
		  if(da == BALANCED) (this.Tree[a](t1,(ka,va,LEFT),t'2),false)
		  else {
		    if(da == RIGHT) (this.Tree[a](t1,(ka,va,BALANCED),t'2),true)
		    else {
		      t1.match[(AVLTree[k,a],boolean)] {
			def Nil = System.error "AVLTree.delete_AVL: delete_AVL[internal error]"
			def Tree(_,(_,_,r),_) = (bal_right[a] (ka,va,da) (t1) (t'2),(r != BALANCED))

		      }
		    }
		  }
		}
		else (this.Tree[a](t1,(ka,va,da),t'2),false)
	      }
	    }
	  }
	}
      }
      Predef.fst (delete_AVL_aux (t))
    }

    def minKey[a] (t:AVLTree[k,a]) : k = {
      t.match {
	def Nil = System.error "AVLTree.minKey: minKey[internal error]"
	def Tree(l,(ka,va,da),r) = {
	  if(l.isEmpty) ka
	  else minKey[a] (l)
	}
      }
    }

    def maxKey[a] (t:AVLTree[k,a]) : k = {
      t.match {
	def Nil = System.error "AVLTree.minKey: minKey[internal error]"
	def Tree(l,(ka,va,da),r) = {
	  if(r.isEmpty) ka
	  else maxKey[a] (r)
	}
      }
    }

    // inherence
    with AssocXDefault[k](this)
    with AssocDefault[k](this)
    with FiniteMapXDefault[k](this)
    with FiniteMapDefault[k](this)

    // -------- static methods of AssocXModule --------
    // Comparison
    def equivalent (x) (y) = equ.equ (x) (y)

    // Constructors
    def empty[a] = this.Nil[a]
    def single[a] (key) (x:a) = this.Tree[a](empty[a],(key,x,BALANCED),empty[a])
    def unionSeq[a] (ms) = unionSeqUsingReduce (ms)
    def fromSeq[a] (kxs) = fromSeqUsingInsertSeq (kxs)

    // -------- static methods of OrdAssocXModule --------
    // Comparison
    val LT = -1
    val GT = 1
    val EQ = 0
    def compare (x) (y) = if (ord.lt (x) (y)) LT else if (ord.lt (y) (x)) GT else EQ

    def unsafeFromOrdSeq (kxs) = fromSeq (kxs)

    // -------- static methods of FiniteMapXModule --------
    def unionSeqWith[a] (f) (ms) = unionSeqWithUsingReduce (f) (ms)
    def fromSeqWith[a] (f: a -> a -> a) (kvs:__Seq__[(k,a)]) = fromSeqWithUsingInsertSeqWith (f) (kvs)
    def fromSeqWithKey[a] (f: k -> a -> a -> a) (kvs:__Seq__[(k,a)]) = fromSeqWithKeyUsingInsertSeqWithKey (f) (kvs)

    // -------- static methods of FiniteMapModule --------
    def unionSeqWithKey[a] (f) (ms) = unionSeqWithKeyUsingReduce (f) (ms)

    // ----------- Nil and Cons ----------
    def Nil[a] = {
      let this : AVLTree[k,a] = {
	// Visitor
	def match visitor = visitor.Nil

        // ----------- AssocX methods -----------
        // Constructors
	def insert (key:k) (sat:a) = single[a] (key) (sat)
        def insertSeq (kxs:__Seq__[(k,a)]) = insertSeqUsingFoldr (kxs) (this)
        def union (m:__AssocX__[k,a]) = m

        // Deletions
        def delete (key:k) = this
        def deleteAll (key:k) = this
        def deleteSeq (ks:__Seq__[k]) = this

        // Observers
        def isEmpty = true
        def size = 0
        def member (key:k) = false
        def count (key:k) = 0
        def lookup (key:k) = System.error "AVLTree.lookup: lookup failed"
        def lookupM (key:k) = (false,null)
        def lookupAll (key:k) = List.empty[a]
        def lookupWithDefault (key:k) (r:a) = r
        def elements = List.empty[a]

        // Modifiers
        def adjust (f: a -> a) (key:k) = this
        def adjustAll (f: a -> a) (key:k) = this

        // Maps and folds
        def map[b] (f: a -> b) = Nil[b]
        def fold[b] (f: a -> b -> b) (c:b) = c
        def fold1 (f: a-> a -> a) = System.error "AVLTree.fold1: empty map"
	def iterate (f: a->()) = null

        // Filters and partitions
        def filter (p: a -> boolean) = this
        def partition (p: a -> boolean) = (this,this)
	def exist (p: a -> boolean) = false

	// ----------- OrdAssocX methods -----------
	// Constructors
	def unsafeInsertMin (key:k) (x:a) = insert (key) (x)
	def unsafeInsertMax (key:k) (x:a) = insert (key) (x)
	def unsafeAppend (m:__OrdAssocX__[k,a]) = m

	// Destructors
	//def minView
	def minElem = System.error "AVLTree.minElem: empty map"
	//def maxView
	def maxElem = System.error "AVLTree.maxElem: empty map"

	// Deletions
	def deleteMin = this
	def deleteMax = this

	// Maps and folds
	def foldr[b] (f: a -> b -> b) (c:b) = c
	def foldl[b] (f: b -> a -> b) (c:b) = c
	def foldr1 (f: a-> a -> a) = System.error "AVLTree.foldr1: empty map"
	def foldl1 (f: a-> a -> a) = System.error "AVLTree.foldl1: empty map"
	def iterater (f:a->()) = null
	def iteratel (f:a->()) = null

	// Filters and partitions
	def filterLT (key:k) = this
	def filterLE (key:k) = this
	def filterGT (key:k) = this
	def filterGE (key:k) = this
	def partitionLT_GE (key:k) = (this,this)
	def partitionLE_GT (key:k) = (this,this)
	def partitionLT_GT (key:k) = (this,this)

        // ----------- Assoc methods -----------
        // Observers
        def toSeq = List.empty[(k,a)]
        def keys = List.empty[k]

        // Maps and folds
        def mapWithKey[b] (f: k -> a -> b) = Nil[b]
        def foldWithKey[b] (f: k -> a -> b -> b) (c:b) = c
	def iterateWithKey (f:k->a->()) = null

        // Filters and partitions
        def filterWithKey (p: k -> a -> boolean) = this
        def partitionWithKey (p: k -> a -> boolean) = (this,this)
	def existWithKey (p:k->a->boolean) = false

	// ----------- OrdAssoc methods -----------
	// Destructors
	//def minViewWithKey
	def minElemWithKey = System.error "AVLTree.minElemWithKey: empty map"
	//def maxViewWithKey
	def maxElemWithKey = System.error "AVLTree.maxElemWithKey: empty map"
	
	// Observers
	def toOrdSeq = List.empty[(k,a)]

	// Maps and folds
	def foldrWithKey[b] (f: k -> a -> b -> b) (c:b) = c
	def foldlWithKey[b] (f: b -> k -> a -> b) (c:b) = c
	def iteraterWithKey (f: k -> a -> ()) = null
	def iteratelWithKey (f: k -> a -> ()) = null

        // ----------- FiniteMapX methods -----------
        // Constructors
        def insertWith (f: a -> a -> a) (k:k) (x:a) = single[a] (k) (x)
        def insertWithKey (f: k -> a -> a -> a) (k:k) (x:a) = single[a] (k) (x)
        def insertSeqWith (f: a -> a -> a) (kvs:__Seq__[(k,a)]) = insertSeqWithUsingInsertWith (f) (kvs) (this)
        def insertSeqWithKey (f: k -> a -> a -> a) (kvs:__Seq__[(k,a)]) = insertSeqWithKeyUsingInsertWithKey (f) (kvs) (this)
        def unionl (m:__FiniteMapX__[k,a]) = m
        def unionr (m:__FiniteMapX__[k,a]) = m
        def unionWith (f: a -> a -> a) (m:__FiniteMapX__[k,a]) = m

        // Set-like operations
        def intersectWith[b,c] (f: a -> b -> c) (m:__FiniteMapX__[k,b]) = empty[c]
        def difference[b] (m:__FiniteMapX__[k,b]) = this
        def subset[b] (m:__FiniteMapX__[k,b]) = (!(m.isEmpty))
        def subsetEq[b] (m:__FiniteMapX__[k,b]) = true

        // ----------- FiniteMap methods -----------
        // Constructors
        def unionWithKey (f: k -> a -> a -> a) (m:__FiniteMap__[k,a]) = m

        // Set-like operations
        def intersectWithKey[b,c] (f: k -> a -> b -> c) (m:__FiniteMapX__[k,b]) = empty[c]

	// private methods
	def outputPS (uid:int) = null
        def outputGraph (uid:int) (depth:int) (id:int) = {
          Console.println( "R" + uid + "-Name "+id+" 0")
	  Console.println( "B" + uid + "-Ter " + depth + " " + id )
	  (true,id)
	}
      }
      this
    }

    def Tree[a] (l:AVLTree[k,a],(k:k,v:a,delta:int),r:AVLTree[k,a]) : AVLTree[k,a] = {
      let this : AVLTree[k,a] = {
	// Visitor
	def match visitor = visitor.Tree(l,(k,v,delta),r)

        // ----------- AssocX methods -----------
        // Constructors
	def insert (key:k) (sat:a) = insertWith (x|y|x) (key) (sat)
        def insertSeq (kxs:__Seq__[(k,a)]) =  insertSeqUsingFoldr (kxs) (this)
        def union (m:__AssocX__[k,a]) =  m.foldWithKey ( k | x | m:AVLTree[k,a] | m.insert (k) (x)) (this)

        // Deletions
        def delete (key:k) = delete_AVL[a] (key) (this)
        def deleteAll (key:k) = delete (key)
        def deleteSeq (ks:__Seq__[k]) = deleteSeqUsingFoldr[a] (ks) (this)

        // Observers
        def isEmpty = false
        def size =  1 + (l.size) + (r.size)
        def member (key:k) = Predef.fst (lookupM (key))
        def count (key:k) = countUsingMember (this) (key)
        def lookup (key:k) = {
	  val cmp=compare (key) (k)
	  if(cmp == LT) l.lookup (key)
	  else if(cmp == GT) r.lookup (key)
	  else v
	}
        def lookupM (key:k) = {
	  val cmp=compare (key) (k)
	  if(cmp == LT) l.lookupM (key)
	  else if(cmp == GT) r.lookupM (key)
	  else (true,v)
	}
        def lookupAll (key:k) = lookupAllUsingLookupM (this) (key)
        def lookupWithDefault (key:k) (r:a) = lookupWithDefaultUsingLookupM (r) (this) (key)
        def elements = elementsUsingFold (this)

        // Modifiers
        def adjust (f: a -> a) (key:k) = insertWith (x | y | f (y)) (key) (null)
        def adjustAll (f: a -> a) (key:k) = adjust (f) (key)

        // Maps and folds
        def map[b] (f: a -> b) = Tree[b](l.map[b] (f),(k,f (v),delta),r.map[b] (f))
        def fold[b] (f: a -> b -> b) (c:b) = r.fold[b] (f) (f (v) (l.fold[b] (f) (c)))
        def fold1 (f: a-> a -> a) = r.fold[a] (f) (l.fold[a] (f) (v))
	def iterate (f: a -> ()) = {
	  (l.iterate) (f)
	  f (v)
	  (r.iterate) (f)
	}

        // Filters and partitions
        def filter (p: a -> boolean) = foldWithKey (k | x | m:AVLTree[k,a] | if(p (x)) m.insert (k) (x) else m) (empty[a])
        def partition (p: a -> boolean) = partitionUsingFilter[a] (p) (this)
	def exist (p: a -> boolean) = {
	  (p (v) || (l.exist) (p)) || (r.exist) (p)
	}


	// ----------- OrdAssocX methods -----------
	// Constructors
	def unsafeInsertMin (key:k) (x:a) = insert (key) (x)
	def unsafeInsertMax (key:k) (x:a) = insert (key) (x)
	def unsafeAppend (m:__OrdAssocX__[k,a]) = union (m)

	// Destructors
	//def minView

	def minElem = if(l.isEmpty) v else l.minElem

	//def maxView

	def maxElem = if(r.isEmpty) v else r.maxElem

	// Deletions
	def deleteMin = delete (minKey (this))
	def deleteMax = delete (maxKey (this))

	// Maps and folds
	def foldr[b] (f: a -> b -> b) (c:b) = l.foldr (f) (f (v) (r.foldr (f) (c)))
	def foldl[b] (f: b -> a -> b) (c:b) = r.foldl (f) (f (l.foldl (f) (c)) (v))
	def foldr1 (f: a-> a -> a) = {
	  if(r.isEmpty) l.foldr (f) (v)
	  else l.foldr (f) (f (v) (r.foldr1 (f)))
	}
	def foldl1 (f: a-> a -> a) = {
	  if(l.isEmpty) r.foldl (f) (v)
	  else r.foldl (f) (f (l.foldl1 (f)) (v))
	}

	def iterater (f:a->()) = {
	  (r.iterater) (f)
	  f (v)
	  (l.iterater) (f)
	}

	def iteratel (f:a->()) = {
	  (l.iterater) (f)
	  f (v)
	  (r.iterater) (f)
	}

	// Filters and partitions
	def filterLT (key:k) = {
	  val cmp=compare (k) (key)
	  if(cmp == LT) (l.insert (k) (v)).union (r.filterLT (key))
	  else l.filterLT (key)
	}
	def filterLE (key:k) = {
	  val cmp=compare (k) (key)
	  if(cmp == GT) l.filterLE (key)
	  else (l.insert (k) (v)).union (r.filterLE (key))
	}
	def filterGT (key:k) = {
	  val cmp=compare (k) (key)
	  if(cmp == GT) (r.insert (k) (v)).union (l.filterGT (key))
	  else r.filterGT (key)
	}
	def filterGE (key:k) = {
	  val cmp=compare (k) (key)
	  if(cmp == LT) r.filterGE (key)
	  else (r.insert (k) (v)).union (l.filterGE (key))
	}

	def partitionLT_GE (key:k) = ( filterLT (key), filterGE (key) )
	def partitionLE_GT (key:k) = ( filterLE (key), filterGT (key) )
	def partitionLT_GT (key:k) = ( filterLT (key), filterGT (key) )

        // ----------- Assoc methods -----------
        // Observers
        def toSeq = toSeqUsingFoldWithKey[a] (this)
        def keys = keysUsingFoldWithKey[a] (this)

        // Maps and folds
        def mapWithKey[b] (f: k -> a -> b) = Tree[b](l.mapWithKey[b] (f),(k,f (k) (v),delta),r.mapWithKey[b] (f))
        def foldWithKey[b] (f: k -> a -> b -> b) (c:b) = r.foldWithKey (f) (f (k) (v) (l.foldWithKey (f) (c)))

	def iterateWithKey (f:k->a->()) = {
	  (l.iterateWithKey) (f)
	  f (k) (v)
	  (r.iterateWithKey) (f)	  
	}


        // Filters and partitions
        def filterWithKey (p: k -> a -> boolean) = foldWithKey (k | x | m:AVLTree[k,a] | if (p (k) (x)) (m.insert (k) (x)) else m) (empty[a])
        def partitionWithKey (p: k -> a -> boolean) = (filterWithKey (p) , filterWithKey ( k | x | (!(p (k) (x))))) 

	def existWithKey (p: k -> a -> boolean) = {
	  (p (k) (v) || (l.existWithKey) (p)) || (r.existWithKey) (p)
	}

	// ----------- OrdAssoc methods -----------
	// Destructors
	//def minViewWithKey
	def minElemWithKey = if(l.isEmpty) (k,v) else l.minElemWithKey
	//def maxViewWithKey
	def maxElemWithKey = if(r.isEmpty) (k,v) else r.maxElemWithKey
	
	// Observers
	def toOrdSeq = foldrWithKey ( k | x | l:List[(k,a)] | List.cons (k,x) (l) ) (List.Nil[(k,a)])

	// Maps and folds
	def foldrWithKey[b] (f: k -> a -> b -> b) (c:b) = l.foldrWithKey (f) (f (k) (v) (r.foldrWithKey (f) (c)))
	def foldlWithKey[b] (f: b -> k -> a -> b) (c:b) = r.foldlWithKey (f) (f (l.foldlWithKey (f) (c)) (k) (v))

 	def iteraterWithKey (f:k->a->()) = {
	  (r.iteraterWithKey) (f)
	  f (k) (v)
	  (l.iteraterWithKey) (f)
	}

	def iteratelWithKey (f:k->a->()) = {
	  (l.iteratelWithKey) (f)
	  f (k) (v)
	  (r.iteratelWithKey) (f)
	}

	// ----------- FiniteMapX methods -----------
        // Constructors
        def insertWith (f: a -> a -> a) (key:k) (sat:a) = insert_AVL[a] ((k|f),key,sat) (this)
        def insertWithKey (f: k -> a -> a -> a) (k:k) (x:a) = insert_AVL[a] (f,k,x) (this)
        def insertSeqWith (f: a -> a -> a) (kvs:__Seq__[(k,a)]) = insertSeqWithUsingInsertWith (f) (kvs) (this)
        def insertSeqWithKey (f: k -> a -> a -> a) (kvs:__Seq__[(k,a)]) = insertSeqWithKeyUsingInsertWithKey (f) (kvs) (this)
        def unionl (m:__FiniteMapX__[k,a]) = unionWith ( x | y | (x) ) (m)
        def unionr (m:__FiniteMapX__[k,a]) = unionWith ( x | y | (y) ) (m)
        def unionWith (f: a -> a -> a) (m:__FiniteMapX__[k,a]) = unionWithUsingInsertWith (f) (this) (m)

        // Set-like operations
        def intersectWith[b,c] (f: a -> b -> c) (m:__FiniteMapX__[k,b]) = intersectWithUsingLookupM[a,b,c] (f) (this) (m)
        def difference[b] (m:__FiniteMapX__[k,b]) = differenceUsingDelete[a,b] (this) (m)
        def subset[b] (m:__FiniteMapX__[k,b]) = subsetUsingSubsetEq[a,b] (this) (m)
        def subsetEq[b] (m:__FiniteMapX__[k,b]) = subsetEqUsingMember[a,b] (this)
(m)

        // ----------- FiniteMap methods -----------
        // Constructors
        def unionWithKey (f: k -> a -> a -> a) (m:__FiniteMap__[k,a]) = unionWithKeyUsingInsertWithKey (f) (this) (m)

        // Set-like operations
        def intersectWithKey[b,c] (f: k -> a -> b -> c) (m:__FiniteMapX__[k,b]) = intersectWithKeyUsingLookupM[a,b,c] (f) (this) (m)


	// private methods
        def outputPS (uid:int) = {
          val _ = outputGraph (uid) (0) (1)
          null
        }
 
        def outputGraph (uid:int) (depth:int) (id:int) = {
 
          Console.println( "R" + uid + "-Name " + id + " " + k + "," + v+":"+delta)
 
          Console.println( "A" + uid + "-Sta " + depth + " " + id )
 
          val (ba,aid) = (l.outputGraph) (uid) (depth+1) (id+1)
          if (ba) {
            Console.println( "C" + uid + "-Pri 32 " + id + " " + (id+1) )
          } else {
          }    
          val (bb,bid) = (r.outputGraph) (uid) (depth+1) (aid+1)
          if (bb) {
            Console.println( "C" + uid + "-Pri 32 " + id + " " + (aid+1) )
          } else {
          }
 
          (true,bid)
         } 
      }
      this
    }
  }
  this
}
