newtype Equ[k] = {
  def equ (x:k) (y:k): boolean
}

newtype AssocX [ k <: Eq[k], a ] = {

  // Constructors
  def insert (key:k) (x:a) : AssocX[k,a]
  def insertSeq (kxs:Seq[(k,a)]) : AssocX[k,a]
  def union (m:AssocX[k,a]) : AssocX[k,a]

  // Deletions
  def delete (key:k) : AssocX[k,a]
  def deleteAll (key:k) : AssocX[k,a]
  def deleteSeq (ks:Seq[k]) : AssocX[k,a]

  // Observers
  def isEmpty: boolean
  def size: int
  def member (key:k) : boolean
  def count (key:k) : int
  def lookup (key:k) : a
  //def lookupM
  def lookupAll (key:k) : Seq[a]
  def lookupWithDefault (key:k) (d:a) : a
  def elements : Seq[a]

  // Modifiers
  def adjust (f: a -> a) (key:k) : AssocX[k,a]
  def adjustAll (f: a -> a) (key:k) : AssocX[k,a]

  // Maps and folds
  def map[b] (f: a -> b) : AssocX[k,a]
  def fold[b] (f: a -> b -> b) (c:b) : b
  def fold1 (f: a-> a -> a) : a

  // Filters and partitions
  def filter (p: a -> boolean) : AssocX[k,a]
  def partition (p: a -> boolean) : (AssocX[k,a],AssocX[k,a])

}

type AssocXModule = {

  // Comparison
  def equivalent (x:k) (y:k) : boolean

  // Constructors
  def empty : AssocX[k,a]
  def single (key:k) (x:a) : AssocX[k,a]
  def unionSeq (ms:Seq[AssocX[k,a]]) : AssocX[k,a]
  def fromSeq (kxs:Seq[(k,a)]) : AssocX[k,a]

}

newtype Ord[k] = {
  def lt (x:k) (y:k): boolean
  def leq (x:k) (y:k): boolean
}

type EqOrd[ k <: Eq[k] & Ord[k] ] = Eq[k] & Ord[k]

newtype OrdAssocX[ k <: EqOrd[k], a  ] <: AssocX[k,a] = {

  with AssocX[k,a]

  // Constructors
  def unsafeInsertMin (key:k) (x:a) : OrdAssocX[k,a]
  def unsafeInsertMax (key:k) (x:a) : OrdAssocX[k,a]
  def unsafeAppend (m:OrdAssocX[k,a]) : OrdAssocX[k,a]

  // Destructors
  //def minView
  def minElem : a
  //def maxView
  def maxElem : a

  // Deletions
  def deleteMin :  OrdAssocX[k,a]
  def deleteMax :  OrdAssocX[k,a]

  // Maps and folds
  def foldr[b] (f: a -> b -> b) (c:b) : b
  def foldl[b] (f: b -> a -> b) (c:b) : b
  def foldr1 (f: a-> a -> a) : a
  def foldl1 (f: a-> a -> a) : a

  // Filters and partitions
  def filterLT (key:k) : OrdAssocX[k,a]
  def filterLE (key:k) : OrdAssocX[k,a]
  def filterGT (key:k) : OrdAssocX[k,a]
  def filterGE (key:k) : OrdAssocX[k,a]
  def partitionLT_GE (key:k) : (OrdAssocX[k,a],OrdAssocX[k,a])
  def partitionLE_GT (key:k) : (OrdAssocX[k,a],OrdAssocX[k,a])
  def partitionLT_GT (key:k) : (OrdAssocX[k,a],OrdAssocX[k,a])

}

type OrdAssocXModule = {

  // Comparison
  def LT : int
  def GT : int
  def EQ : int
  def compare (x:k) (y:k): int // should return LT, GT ou EQ

  def unsafeFromOrdSeq (kxs:Seq[(k,a)]) : OrdAssocX[k,a]

}

newtype Assoc[ k <: Eq[k], a  ] <: AssocX[k,a] = {

  with AssocX[k,a]

  // Observers
  def toSeq : Seq[(k,a)]
  def keys : Seq[k]

  // Maps and folds
  def mapWithKey[b] (f: k -> a -> b) : Assoc[k,b]
  def foldWithKey[b] (f: k -> a -> b -> b) : (c:b) : b

  // Filters and partitions
  def filterWithKey (p: k -> a -> boolean) : Assoc[k,a]
  def partitionWithKey (p: k -> a -> boolean) : (Assoc[k,a], Assoc[k,a])

}

newtype OrdAssoc[ k <: EqOrd[k], a ] <: Assoc[k,a] & OrdAssocX[k,a] = {

  with Assoc[k,a]
  with OrdAssocX[k,a]

  // Destructors
  //def minViewWithKey
  def minElemWithKey : (k,a)
  //def maxViewWithKey
  def maxElemWithKey : (k,a)

  // Observers
  def toOrdSeq : Seq[(k,a)]

  // Maps and folds
  def foldrWithKey[b] (f: k -> a -> b -> b) (c:b) : b
  def foldlWithKey[b] (f: b -> k -> a -> b) (c:b) : b

}

newtype FiniteMapX [ k <: Eq[k], a ] <: AssocX[k,a] = {

  with AssocX[k,a]

  // Constructors
  def insertWith (f: a -> a -> a) (k:k) (x:a) : FiniteMapX[k,a]
  def insertWithKey (f: k -> a -> a -> a) (k:k) (x:a) : FiniteMapX[k,a]
  def insertSeqWith (f: a -> a -> a) (kvs:Seq[(k,a)]) : FiniteMapX[k,a]
  def insertSeqWithKey (f: k -> a -> a -> a) (kvs:Seq[(k,a)]) : FiniteMapX[k,a]
  def unionl (m:FiniteMapX[k,a]) : FiniteMapX[k,a]
  def unionr (m:FiniteMapX[k,a]) : FiniteMapX[k,a]
  def unionWith (f: a -> a -> a) (m:FiniteMapX[k,a]) : FiniteMapX[k,a]
  def fromSeqWith (f: a -> a -> a) (kvs:Seq[(k,a)]) : FiniteMapX[k,a]
  def fromSeqWithKey (f: k -> a -> a -> a) (kvs:Seq[(k,a)]) : FiniteMapX[k,a]

  // Set-like operations
  def intersectWith[b,c] (f: a -> b -> c) (m:FiniteMapX[k,b]) : FiniteMapX[k,c]
  def difference[b] (m:FiniteMapX[k,b]) : FiniteMapX[k,a]
  def subset[b] (m:FiniteMapX[k,b]) : boolean
  def subsetEq[b] (m:FiniteMapX[k,b]) : boolean

}

type FiniteMapXModule = {

  def unionSeqWith (f: a -> a -> a) (ms:Seq[FiniteMapX[k,a]]) : FiniteMapX[k,a]

}

newtype FiniteMap [ k <: Eq[k], a ] <: FiniteMapX[k,a] & Assoc[k,a] = {

  with FiniteMapX[k,a]
  with Assoc[k,a]

  // Constructors
  def unionWithKey (f: k -> a -> a -> a) (m:FiniteMap[k,a]) : FiniteMap[k,a]

  // Set-like operations
  def intersectWithKey[b,c] (f: k -> a -> b -> c) (m:FiniteMapX[k,b]) : FiniteMapX[k,c]

}

type FiniteMapModule = {
  def unionSeqWithKey (f: k -> a -> a -> a) (ms:Seq[FiniteMap[k,a]]) : FiniteMap[k,a]

}

newtype OrdFiniteMapX[ k <: EqOrd[k], a ] <: FiniteMapX[k,a] & OrdAssocX[k,a] = {

  with FiniteMapX[k,a]
  with OrdAssocX[k,a]

}

newtype OrdFiniteMap[ k <: EqOrd[k], a ] <: FiniteMap[k,a] & OrdAssoc[k,a] = {

  with FiniteMap[k,a]
  with OrdAssoc[k,a]

}
