// -*- Funnel -*-

/**
 * $Author: ysaidji $
 * $Id: AssocList.m4,v 1.4 2001/07/02 11:59:08 ysaidji Exp $
 */

def removeMe0 = include[AssocListType]
def removeMe1 = include[__AssocDefaultType__]
def removeMe2 = include[__AssocDefault__]

def AssocList[k] (equ: Equ[k]) : AssocListModule[k] = {

  let this = {

    // inherence
    with AssocXDefault[k](this)
    with AssocDefault[k](this)
    with FiniteMapXDefault[k](this)
    with FiniteMapDefault[k](this)

    // -------- static methods of AssocXModule --------
    // Comparison
    def equivalent (x) (y) = equ.equ (x) (y)
    // Constructors
    def empty[a] = Nil[a]
    def single[a] (key) (x:a) = Cons[a](key,x,Nil[a])
    def unionSeq[a] (ms) = ms.foldr (m1:__AssocX__[k,a] | m2:AssocList[k,a] | m2.union (m1)) (Nil[a])
    def fromSeq[a] (kxs) = kxs.foldr ( (k:k,x:a) | m:AssocList[k,a] | m.insert (k) (x) ) (Nil[a])

    // -------- static methods of FiniteMapXModule --------
    def unionSeqWith[a] (f) (ms) = unionSeqWithUsingFoldr[a] (f) (ms)

    // -------- static methods of FiniteMapModule --------
    def unionSeqWithKey[a] (f) (ms) = unionSeqWithKeyUsingFoldr[a] (f) (ms)

    // -------- Private static methods --------

    // ----------- Nil and Cons ----------

    def Nil[a] = {
      let this : AssocList[k,a] = {
	// private methods
	def lookupM (key:k) = (false,null)
	def lhead = System.error "AssocList.lookup: lhead failed"
	def ltail = this

        // ----------- AssocX methods -----------
        // Constructors
	def insert (key:k) (x:a) = Cons[a](key,x,this) 
        def insertSeq (kxs:__Seq__[(k,a)]) = fromSeq[a] (kxs)
        def union (m:__AssocX__[k,a])  = m

        // Deletions
        def delete (key:k) = this
        def deleteAll (key:k)  = this
        def deleteSeq (ks:__Seq__[k])  = this

        // Observers
        def isEmpty = true
        def size = 0
        def member (key:k)  = false
        def count (key:k)  = 0
        def lookup (key:k)  = System.error "AssocList.lookup: lookup failed"
        //def lookupM
        def lookupAll (key:k) = List.empty[a]
        def lookupWithDefault (key:k) (d:a) = d
        def elements = List.empty[a]

        // Modifiers
        def adjust (f: a -> a) (key:k) = this
        def adjustAll (f: a -> a) (key:k) = this

        // Maps and folds
        def map[b] (f: a -> b) = Nil[b]
        def fold[b] (f: a -> b -> b) (c:b) = c
        def fold1 (f: a-> a -> a)  = System.error "AssocList.fold1: empty map"

        // Filters and partitions
        def filter (p: a -> boolean) = this
        def partition (p: a -> boolean) = (this,this)

        // ----------- Assoc methods -----------
        // Observers
        def toSeq = List.empty[(k,a)]
        def keys = List.empty[k]

        // Maps and folds
        def mapWithKey[b] (f: k -> a -> b) = Nil[b]
        def foldWithKey[b] (f: k -> a -> b -> b) (c:b)  = c

        // Filters and partitions
        def filterWithKey (p: k -> a -> boolean) = this
        def partitionWithKey (p: k -> a -> boolean) = (this,this)

        // ----------- FiniteMapX methods -----------
        // Constructors
        def insertWith (f: a -> a -> a) (k:k) (x:a) = Cons[a](k,x,this)
        def insertWithKey (f: k -> a -> a -> a) (k:k) (x:a) = Cons[a](k,x,this)
        def insertSeqWith (f: a -> a -> a) (kvs:__Seq__[(k,a)]) =  fromSeqWithUsingInsertSeqWith[a] (f) (kvs)
        def insertSeqWithKey (f: k -> a -> a -> a) (kvs:__Seq__[(k,a)]) = fromSeqWithKeyUsingInsertSeqWithKey[a] (f) (kvs)
        def unionl (m:__FiniteMapX__[k,a]) = m
        def unionr (m:__FiniteMapX__[k,a]) = m
        def unionWith (f: a -> a -> a) (m:__FiniteMapX__[k,a]) = unionWithUsingInsertWith[a] (f) (this) (m)
        def fromSeqWith (f: a -> a -> a) (kvs:__Seq__[(k,a)]) = fromSeqWithUsingInsertSeqWith[a] (f) (kvs)
        def fromSeqWithKey (f: k -> a -> a -> a) (kvs:__Seq__[(k,a)]) = fromSeqWithKeyUsingInsertSeqWithKey[a] (f) (kvs)

        // Set-like operations
        def intersectWith[b,c] (f: a -> b -> c) (m:__FiniteMapX__[k,b]) = intersectWithUsingLookupM[a,b,c] (f) (this) (m)
        def difference[b] (m:__FiniteMapX__[k,b]) = differenceUsingDelete[a,b] (this) (m)
        def subset[b] (m:__FiniteMapX__[k,b]) = subsetUsingSubsetEq[a,b] (this) (m)
        def subsetEq[b] (m:__FiniteMapX__[k,b]) = subsetEqUsingMember[a,b] (this) (m)

        // ----------- FiniteMap methods -----------
        // Constructors
        def unionWithKey (f: k -> a -> a -> a) (m:__FiniteMap__[k,a]) = unionWithKeyUsingInsertWithKey (f) (this) (m)

        // Set-like operations
        def intersectWithKey[b,c] (f: k -> a -> b -> c) (m:__FiniteMapX__[k,b]) = intersectWithKeyUsingLookupM[a,b,c] (f) (this) (m)
      }
      this
    }

    def Cons[a] (k:k, x:a, m:AssocList[k,a]) = {
      let this : AssocList[k,a] = {
	// private methods
	def lookupM (key:k) = {
	  if (equivalent (k) (key)) (true,x) else m.lookupM (key)
	}
	def lhead = (k,x)
	def ltail = m

        // ----------- AssocX methods -----------
        // Constructors
	def insert (key:k) (x:a) = Cons[a](key,x,this)
        def insertSeq (kxs:__Seq__[(k,a)]) = kxs.foldr ((k,x) | m:AssocList[k,a] | m.insert(k) (x)) (this)
        def union (m2:__AssocX__[k,a]) = Cons[a]( k, x, m.union(m2) )

        // Deletions
        def delete (key:k) = deleteAll (key)
        def deleteAll (key:k) = {
	  if (equivalent (k) (key)) m.deleteAll (key)
	  else Cons[a]( k, x, m.deleteAll(key) )
	}
        def deleteSeq (ks:__Seq__[k]) = deleteSeqUsingFoldr[a] (ks) (this)

        // Observers
        def isEmpty = false
        def size = 1 + m.size
        def member (key:k) = (equivalent (key) (k)) || m.member(key)
        def count (key:k) = if (equivalent (k) (key)) 1 else m.count(key)
        def lookup (key:k) = if (equivalent (k) (key)) x else m.lookup(key)
        //def lookupM
        def lookupAll (key:k) = if (equivalent (k) (key)) List.single[a] (x) else m.lookupAll(key)
        def lookupWithDefault (key:k) (d:a)  = if (equivalent (k) (key)) x else m.lookupWithDefault (key) (d)
        def elements = List.cons[a] (x) ((m.delete(k)).elements) 

        // Modifiers
        def adjust (f: a -> a) (key:k) = if (equivalent (k) (key)) Cons[a](key, f (x), m) else Cons[a](k, x, m.adjust (f) (key))
        def adjustAll (f: a -> a) (key:k) = adjust (f) (key)

        // Maps and folds
        def map[b] (f: a -> b) = Cons[b]( k, f (x), m.map (f) )
        def fold[b] (f: a -> b -> b) (c:b) = (m.delete (k)).fold (f) (f (x) (c))
        def fold1 (f: a-> a -> a) = (m.delete (k)).fold (f) (x)

        // Filters and partitions
        def filter (p: a -> boolean) = {
	  if (p (x))
	    Cons[a]( k, x, (m.delete(k)).filter (p) )
	  else
	    (m.delete(k)).filter (p)
	}

        def partition (p: a -> boolean) = {
	  val (m1,m2) = (m.delete(k)).partition (p);
	  if (p (x))
	    (Cons[a]( k, x, m1 ), m2)
	  else
	    (m1, Cons[a]( k, x, m2 ))	  
	}

        // ----------- Assoc methods -----------
        // Observers
        def toSeq = List.Cons[(k,a)]( (k,x), (m.delete(k)).toSeq )
        def keys = List.Cons[k]( k, (m.delete(k)).keys )

        // Maps and folds
        def mapWithKey[b] (f: k -> a -> b) = Cons[b]( k, f (k) (x), m.mapWithKey (f) )

        //def foldWithKey[b] (f: k -> a -> b -> b) (c:b) = (m.delete (k)).foldWithKey (f) (f (k) (x) (c))

        def foldWithKey[b] (f: k -> a -> b -> b) (c:b) = {

	  def exist (ks:List[k]) (k:k) : boolean = ks.match {
	    def Nil = false
	    def Cons(x,xs) = if (equivalent (x) (k)) true else exist (xs) (k)
	  }
	  
	  def fwk (ks:List[k]) (m:AssocList[k,a]) (f: k -> a -> b -> b) (c:b) : b = {
	    if (m.isEmpty) c
	    else {
	      val (k,x) = m.lhead
	      if (!(exist (ks) (k))) {
		val ks' = List.cons (k) (ks)
		fwk (ks') (m.ltail) (f) (f (k) (x) (c))
	      } else {
		fwk (ks) (m.ltail) (f) (c)
	      }
	    }
	  }

	  fwk (List.empty[k]) (this) (f) (c)
	}

        // Filters and partitions
        def filterWithKey (p: k -> a -> boolean) = {
	  if (p (k) (x))
	    Cons[a]( k, x, (m.delete(k)).filterWithKey (p) )
	  else
	    (m.delete(k)).filterWithKey (p)
	}

        def partitionWithKey (p: k -> a -> boolean) = {
	  val (m1,m2) = (m.delete(k)).partitionWithKey (p);
	  if (p (k) (x))
	    (Cons[a]( k, x, m1 ), m2)
	  else
	    (m1, Cons[a]( k, x, m2 ))	  
	}

        // ----------- FiniteMapX methods -----------
        // Constructors
        def insertWith (f: a -> a -> a) (k:k) (x:a) = insertWithUsingLookupM (f) (k) (x) (this)
        def insertWithKey (f: k -> a -> a -> a) (k:k) (x:a) = insertWithKeyUsingInsertWith (f) (k) (x) (this)
        def insertSeqWith (f: a -> a -> a) (kvs:__Seq__[(k,a)]) = insertSeqWithUsingInsertWith (f) (kvs) (this)
        def insertSeqWithKey (f: k -> a -> a -> a) (kvs:__Seq__[(k,a)]) = insertSeqWithKeyUsingInsertWithKey (f) (kvs) (this)
        def unionl (m:__FiniteMapX__[k,a]) = union (m)
        def unionr (m:__FiniteMapX__[k,a]) = m.union (this)
        def unionWith (f: a -> a -> a) (m:__FiniteMapX__[k,a]) = unionWithUsingInsertWith (f) (this) (m)
        def fromSeqWith (f: a -> a -> a) (kvs:__Seq__[(k,a)]) = fromSeqWithUsingInsertSeqWith (f) (kvs)
        def fromSeqWithKey (f: k -> a -> a -> a) (kvs:__Seq__[(k,a)]) = fromSeqWithKeyUsingInsertSeqWithKey (f) (kvs)

        // Set-like operations
        def intersectWith[b,c] (f: a -> b -> c) (m:__FiniteMapX__[k,b]) = intersectWithUsingLookupM[a,b,c] (f) (this) (m)
        def difference[b] (m:__FiniteMapX__[k,b]) = differenceUsingDelete[a,b] (this) (m)
        def subset[b] (m:__FiniteMapX__[k,b]) = subsetUsingSubsetEq[a,b] (this) (m)
        def subsetEq[b] (m:__FiniteMapX__[k,b]) = subsetEqUsingMember[a,b] (this) (m)

        // ----------- FiniteMap methods -----------
        // Constructors
        def unionWithKey (f: k -> a -> a -> a) (m:__FiniteMap__[k,a]) = unionWithKeyUsingInsertWithKey (f) (this) (m)

        // Set-like operations
        def intersectWithKey[b,c] (f: k -> a -> b -> c) (m:__FiniteMapX__[k,b]) = intersectWithKeyUsingLookupM[a,b,c] (f) (this) (m)
     }
      this
    }
   

  }
  this

}
