// -*- Funnel -*-

/**
 * $Author: ysaidji $
 * $Id: AssocListType.m4,v 1.2 2001/06/29 16:39:01 ysaidji Exp $
 */

newtype AssocList[k,a] = {

  // ----------- AssocX methods -----------
  // Constructors
  def insert (key:k) (x:a) : AssocList[k,a]
  def insertSeq (kxs:__Seq__[(k,a)]) : AssocList[k,a]
  def union (m:__AssocX__[k,a]) : AssocList[k,a]

  // Deletions
  def delete (key:k) : AssocList[k,a]
  def deleteAll (key:k) : AssocList[k,a]
  def deleteSeq (ks:__Seq__[k]) : AssocList[k,a]

  // Observers
  def isEmpty: boolean
  def size: int
  def member (key:k) : boolean
  def count (key:k) : int
  def lookup (key:k) : a
  //def lookupM
  def lookupAll (key:k) : __Seq__[a]
  def lookupWithDefault (key:k) (d:a) : a
  def elements : __Seq__[a]

  // Modifiers
  def adjust (f: a -> a) (key:k) : AssocList[k,a]
  def adjustAll (f: a -> a) (key:k) : AssocList[k,a]

  // Maps and folds
  def map[b] (f: a -> b) : AssocList[k,b]
  def fold[b] (f: a -> b -> b) (c:b) : b
  def fold1 (f: a-> a -> a) : a

  // Filters and partitions
  def filter (p: a -> boolean) : AssocList[k,a]
  def partition (p: a -> boolean) : (AssocList[k,a],AssocList[k,a])
 
  // ----------- Assoc methods -----------
  // Observers
  def toSeq : __Seq__[(k,a)]
  def keys : __Seq__[k]

  // Maps and folds
  def mapWithKey[b] (f: k -> a -> b) : AssocList[k,b]
  def foldWithKey[b] (f: k -> a -> b -> b) (c:b) : b

  // Filters and partitions
  def filterWithKey (p: k -> a -> boolean) : AssocList[k,a]
  def partitionWithKey (p: k -> a -> boolean) : (AssocList[k,a], AssocList[k,a])

  // ----------- FiniteMapX methods -----------
  // Constructors
  def insertWith (f: a -> a -> a) (k:k) (x:a) : AssocList[k,a]
  def insertWithKey (f: k -> a -> a -> a) (k:k) (x:a) : AssocList[k,a]
  def insertSeqWith (f: a -> a -> a) (kvs:__Seq__[(k,a)]) : AssocList[k,a]
  def insertSeqWithKey (f: k -> a -> a -> a) (kvs:__Seq__[(k,a)]) : AssocList[k,a]
  def unionl (m:__FiniteMapX__[k,a]) : AssocList[k,a]
  def unionr (m:__FiniteMapX__[k,a]) : AssocList[k,a]
  def unionWith (f: a -> a -> a) (m:__FiniteMapX__[k,a]) : AssocList[k,a]
  def fromSeqWith (f: a -> a -> a) (kvs:__Seq__[(k,a)]) : AssocList[k,a]
  def fromSeqWithKey (f: k -> a -> a -> a) (kvs:__Seq__[(k,a)]) : AssocList[k,a]

  // Set-like operations
  def intersectWith[b,c] (f: a -> b -> c) (m:__FiniteMapX__[k,b]) : AssocList[k,c]
  def difference[b] (m:__FiniteMapX__[k,b]) : AssocList[k,a]
  def subset[b] (m:__FiniteMapX__[k,b]) : boolean
  def subsetEq[b] (m:__FiniteMapX__[k,b]) : boolean

  // ----------- FiniteMap methods -----------
  // Constructors
  def unionWithKey (f: k -> a -> a -> a) (m:__FiniteMap__[k,a]) : AssocList[k,a]

  // Set-like operations
  def intersectWithKey[b,c] (f: k -> a -> b -> c) (m:__FiniteMapX__[k,b]) : AssocList[k,c]

  // ----------- private methods -----------
  def lookupM (key:k) : (boolean,a)
  def lhead : (k,a)
  def ltail : AssocList[k,a]
    

}

type AssocListModule[k] = {

  // -------- Constructors -------- 
  def Nil[a] : AssocList[k,a]
  def Cons[a] (k:k, x:a, m:AssocList[k,a]) : AssocList[k,a]

  // -------- static methods of AssocXModule --------
  // Comparison
  def equivalent (x:k) (y:k) : boolean

  // Constructors
  def empty[a] : AssocList[k,a]
  def single[a] (key:k) (x:a) : AssocList[k,a]
  def unionSeq[a] (ms:__Seq__[__AssocX__[k,a]]) : AssocList[k,a]
  def fromSeq[a] (kxs:__Seq__[(k,a)]) : AssocList[k,a]

  // -------- static methods of FiniteMapXModule --------
  def unionSeqWith[a] (f: a -> a -> a) (ms:__Seq__[__FiniteMapX__[k,a]]) : AssocList[k,a]

  // -------- static methods of FiniteMapModule --------
  def unionSeqWithKey[a] (f: k -> a -> a -> a) (ms:__Seq__[__FiniteMap__[k,a]]) : AssocList[k,a]

  // -------- Private static methods --------

}

newtype Equ[k] = {
  def equ (x:k) (y:k): boolean
}

newtype Ord[k] = {
  def lt (x:k) (y:k): boolean
  def leq (x:k) (y:k): boolean
}


newtype AssocListType = {
  def dummy:any
}
