// ------------- the Assoc test -------------------


//def removeMe = include[AVLTree]
def removeMe = include[RedBlackTree]

// some relations
val IntEq : Equ[int] = {
  def equ (x:int) (y:int) : boolean = (x == y)
}
val IntOrd : Ord[int] = {
  def lt (x:int) (y:int): boolean = (x < y)
  def leq (x:int) (y:int): boolean = (x <= y)
}

// -------- Put you implementation here --------------
/////////////////////////////////////////////////////////////////////

//type Assoc[k,a] = AVLTree[k,a]
//let IntAssoc : AVLTreeModule[int] = AVLTree[int](IntEq, IntOrd)
type Assoc[k,a] = RedBlackTree[k,a]
let IntAssoc : RedBlackTreeModule[int] = RedBlackTree[int](IntEq, IntOrd)

/////////////////////////////////////////////////////////////////////

// ----- the generators ---------
let gIntInt : Gen[(int,int)] = {
  def gen = ( Gen.gInt.gen, Gen.gInt.gen )
}

let gListIntInt : Gen[List[(int,int)]] = Gen.gList[(int,int)] (gIntInt)

let gAssocIntInt : Gen[Assoc[int,int]] = {
  def gen = IntAssoc.fromSeq[int] ( gListIntInt.gen )
}

let gListAssocIntInt : Gen[List[Assoc[int,int]]] = Gen.gList[Assoc[int,int]] (gAssocIntInt)

// ------- some helpers ---------

def lmerge (xs:List[(int,int)]) (ys:List[(int,int)]) : List[(int,int)] = {
  if (xs.isEmpty) ys
  else if (ys.isEmpty) xs
  else {
    val x = Predef.fst(xs.lhead)
    val y = Predef.fst(ys.lhead)
    if ( x < y ) List.cons[(int,int)] (xs.lhead) (lmerge (xs.ltail) (ys))
    else if ( x == y ) (lmerge (xs.ltail) (ys))
    else List.cons[(int,int)] (ys.lhead) (lmerge (xs) (ys.ltail))
  }
}

def lookup (x:int) (xs:List[(int,int)]) : (boolean,int) = {
  if (xs.isEmpty) (false,null)
  else {
    if (x == Predef.fst(xs.lhead)) (true,Predef.snd(xs.lhead))
    else lookup (x) (xs.ltail)
  }
}

def lookupAll (x:int) (xs:List[(int,int)]) : List[int] = {
  def loop  (x:int) (xs:List[(int,int)]) (lst:List[int]) = {
    if (xs.isEmpty) List.empty[int]
    else {
      if (x == Predef.fst(xs.lhead)) {
	(List.append[int]) (lst) ((List.cons[int]) (Predef.snd(xs.lhead)) (lookupAll (x) (xs.ltail)))
      } else (List.append[int]) (lst) (lookupAll (x) (xs.ltail))
    }
  }
  loop (x) (xs) (List.empty[int])
}

def insert (xy:(int,int)) (xs:List[(int,int)]) : List[(int,int)] = {
  if (xs.isEmpty) List.single (xy)
  else if ( Predef.fst(xy) < Predef.fst(xs.lhead) ) List.cons (xy) (xs)
  else if ( Predef.fst(xy) == Predef.fst(xs.lhead) ) List.cons (xy) (xs.ltail)
  else List.cons (xs.lhead) ( insert (xy) (xs.ltail) )
}

def delete (xy:(int,int)) (xs:List[(int,int)]) : List[(int,int)] = {
  if (xs.isEmpty) xs
  else if ( Predef.fst(xy) == Predef.fst(xs.lhead) ) xs.ltail
  else List.cons (xs.lhead) ( delete (xy) (xs.ltail) )
}

def quicksort[a] (comp:a->a->boolean) (lst:List[a]) : List[a] = 
  if (lst.isEmpty || lst.ltail.isEmpty) lst
  else ( val (i1,i2) = lst.ltail.partition( comp (lst.lhead) ); 
	 (List.append[a]) ((quicksort[a]) (comp) (i1)) (List.cons[a] (lst.lhead) ((quicksort[a]) (comp) (i2))) )
def pptT (a:(int,int)) (b:(int,int)): boolean = (Predef.fst(b) < Predef.fst(a))
def sortT (xs:List[(int,int)]) : List[(int,int)] = quicksort[(int,int)] (pptT) (xs) 

def ppt (a:int) (b:int): boolean = (b < a)
def sort (xs:List[int]) : List[int] = quicksort[int] (ppt) (xs) 

let QTAssoc = {

  with Gen
  with Result
  with Predef

  type b = (int,int) // binding

  // List comparison
  def equList (xs:List[int]) (ys:List[int]) = (xs.zip (ys)).foldl ( b | (x,y) | (x==y) && b  ) (true)

  def equTuple (xs:List[(int,int)]) (ys:List[(int,int)]) = {
    def eqt (x:int,y:int) (x':int,y':int) = (x==x') && (y==y')
    if ( (xs.size) == (ys.size) ) (xs.zip (ys)).foldl ( b | (x,y) | (eqt (x) (y)) && b  ) (true)
    else false
  }

  // Collections comparison
  def equ (xs:Assoc[int,int]) (ys:Assoc[int,int]) = {
    equTuple (xs.toOrdSeq) (ys.toOrdSeq)
  }

  // Debug helpers
  def printAssoc (m:Assoc[int,int]) : () = {
      (m.toOrdSeq).iterate ( xy:b | Console.print("["+ fst(xy) + "/" + snd(xy) + "]")  )
  }

  ////////////////////////////////////////////////////////////////////
  ////    AssocX operations
  ////////////////////////////////////////////////////////////////////

  // ------ single ----------
  def prop_single (xy:b) : boolean = {
    equTuple ((IntAssoc.single[int] (fst(xy)) (snd(xy))).toOrdSeq) (List.single (xy))
  }
  def prop_single_test : Result = QuickCheck.qc0b[b] (prop_single) (gIntInt)

  // ------ fromSeq ----------
  def prop_fromSeq (xys:List[b]) : boolean = {
    equ (IntAssoc.fromSeq (xys)) 
      (xys.foldr ( xy | m:Assoc[int,int] | m.insert (fst(xy)) (snd(xy))) (IntAssoc.empty))
  }
  def prop_fromSeq_test : Result = QuickCheck.qc0b[List[b]] (prop_fromSeq) (gListIntInt)
  
  // ------ insert ----------
  def prop_insert (xy:b) (m:Assoc[int,int]) : boolean = {
    val m' = m.insert (fst(xy)) (snd(xy))
    val b = equTuple (m'.toOrdSeq) (insert (xy) (m.toOrdSeq))
    b
  }
  def prop_insert_test : Result = QuickCheck.qc1b[b,Assoc[int,int]] (prop_insert) (gIntInt) (gAssocIntInt)


  // ------ union ----------
  def prop_union (xs:Assoc[int,int]) (ys:Assoc[int,int]) : boolean = {
    val a = (xs.union (ys))
    val b = equTuple (a.toOrdSeq) (lmerge (xs.toOrdSeq) (ys.toOrdSeq))
    b
  }
  def prop_union_test : Result = QuickCheck.qc1b[Assoc[int,int],Assoc[int,int]] (prop_union) (gAssocIntInt) (gAssocIntInt)

  // ------ unionSeq ----------
  def prop_unionSeq (xss:List[Assoc[int,int]]) : boolean = {
    equ (IntAssoc.unionSeq (xss)) (xss.foldr (x|m:Assoc[int,int]|m.union (x)) (IntAssoc.empty))
  }
  def prop_unionSeq_test : Result = QuickCheck.qc0b[List[Assoc[int,int]]] (prop_unionSeq) (gListAssocIntInt)

  // ------ delete ----------
  def prop_delete (x:int) (m:Assoc[int,int]) : boolean = {
    equTuple ((m.delete (x)).toOrdSeq) (delete (x,null) (m.toOrdSeq))
  }
  def prop_delete_test : Result = QuickCheck.qc1b[int,Assoc[int,int]] (prop_delete) (gInt) (gAssocIntInt)

  // ------ deleteAll ----------
  def prop_deleteAll (x:int) (m:Assoc[int,int]) : boolean = {
    val temp = ((m.toOrdSeq).filter (xy|(x !=fst(xy))))
    equTuple ((m.deleteAll (x)).toOrdSeq) (temp)
  }
  def prop_deleteAll_test : Result = QuickCheck.qc1b[int,Assoc[int,int]] (prop_delete) (gInt) (gAssocIntInt)


  // ------ deleteSeq ----------
  def prop_deleteSeq (xs:List[int]) (ys:Assoc[int,int]) : boolean = {
    equ (ys.deleteSeq (xs)) (xs.foldr (x|m:Assoc[int,int]|m.delete (x)) (ys))
  }
  def prop_deleteSeq_test : Result = QuickCheck.qc1b[List[int],Assoc[int,int]] (prop_deleteSeq) (gListInt) (gAssocIntInt)

  // ------ isEmpty_size ----------
  def prop_isEmpty_size (xs:Assoc[int,int]) : boolean = {
    ((xs.isEmpty) == (0 == xs.size))
      && ((xs.size) == ((xs.toOrdSeq).size))
  }
  def prop_isEmpty_size_test : Result = QuickCheck.qc0b[Assoc[int,int]] (prop_isEmpty_size) (gAssocIntInt)


  // ------ member_count ----------
  def prop_member_count (xs:Assoc[int,int]) (x:int) : boolean = {
    val c = xs.count (x)
    ((xs.member (x)) == (c > 0))
      && ( c == ((xs.toOrdSeq).filter (xy|(x==fst(xy)))).size )
  }
  def prop_member_count_test : Result = QuickCheck.qc1b[Assoc[int,int],int] (prop_member_count) (gAssocIntInt) (gInt)


  ////////////////////////////////////////////////////////////////////
  ////    Assoc operations
  ////////////////////////////////////////////////////////////////////

  // ------ toSeq ----------
  def prop_toSeq (xs:Assoc[int,int]) : boolean = {
    equTuple (sortT (xs.toSeq)) (xs.toOrdSeq)
  }
  def prop_toSeq_test : Result = QuickCheck.qc0b[Assoc[int,int]] (prop_toSeq) (gAssocIntInt)

  // ------ lookup ----------
  def prop_lookup (xs:Assoc[int,int]) (x:int) : boolean = {
    if ( xs.member (x) ) {
      val (b,y) = lookup (x) (xs.toOrdSeq)
      val ys = lookupAll (x) (xs.toOrdSeq)
      (b) && ((xs.lookup (x)) == y)
        && ((xs.lookupWithDefault (x) (999)) == (y))
        && equList (sort (xs.lookupAll (x))) (sort(ys))
    } else {
      ((xs.lookupWithDefault (x) (999)) ==  (999))
        && equList (xs.lookupAll (x)) (List.empty[int])
    }
  }
  def prop_lookup_test : Result = QuickCheck.qc1b[Assoc[int,int],int] (prop_lookup) (gAssocIntInt) (gInt)

  // ------ fold ----------
  def prop_fold (xs:Assoc[int,int]) : boolean = {
    def plus (x:int) (y:int) : int = x + y
    //def plusT (k1:int,x:int) (k2:int,y:int) : int = plus (x) (y)
    val a = sort ((xs.toOrdSeq).foldl (l:List[int] | (x:int,y:int) | List.cons (y) (l)) (List.empty[int]))
    val sa = a.foldl (plus) (0)
    equList (sort (xs.fold (List.cons[int]) (List.empty[int]))) (a)
      && ( xs.isEmpty || ((xs.fold1 (plus)) == (sa)) )
  }
  def prop_fold_test : Result = QuickCheck.qc0b[Assoc[int,int]] (prop_fold) (gAssocIntInt)


  // ------ filter_partition ----------
  def prop_filter_partition (xs:Assoc[int,int]) : boolean = {
    def p (x:int) : boolean = ( (x % 3) == 2 )
    val (a,b) = xs.partition (p)
    equTuple ((xs.filter (p)).toOrdSeq) ((xs.toOrdSeq).filter ((x:int,y:int) | p(y)))
      && equ (a) (xs.filter (p))
      && equ (b) (xs.filter (x|(!p (x))))
  }
  def prop_filter_partition_test : Result = QuickCheck.qc0b[Assoc[int,int]] (prop_filter_partition) (gAssocIntInt)

  ////////////////////////////////////////////////////////////////////
  ////    OrdAssocX operations
  ////////////////////////////////////////////////////////////////////

  // ------ deleteMin_Max ----------
  def prop_deleteMin_Max (xs:Assoc[int,int]) : boolean = {
    equTuple ((xs.deleteMin).toOrdSeq) ((xs.toOrdSeq).ltail)
      && equTuple ((xs.deleteMax).toOrdSeq) ((xs.toOrdSeq).rtail)
  }
  def prop_deleteMin_Max_test : Result = QuickCheck.qc0b[Assoc[int,int]] (prop_deleteMin_Max) (gAssocIntInt)

  // ------ unsafeInsertMin_Max ----------
  def prop_unsafeInsertMin_Max (i:int) (xs:Assoc[int,int]) : boolean = {
    if (xs.isEmpty) {
      equ (xs.unsafeInsertMin (0)(42)) (IntAssoc.single[int] (0)(42))
        && equ (xs.unsafeInsertMax (0)(42)) (IntAssoc.single[int] (0)(42))
    } else {
      def odd (x:int) : boolean = ((x%2) == 1)
      val lo' = fst((xs.toOrdSeq).lhead)
      val hi' = fst((xs.toOrdSeq).rhead)
      val lo = lo' - ( if (odd (i)) 1 else 0 )
      val hi = hi' + ( if (odd (i)) 1 else 0 )
      equ (xs.unsafeInsertMin (lo)(42)) (xs.insert (lo)(42))
        && equ (xs.unsafeInsertMax (hi)(42)) (xs.insert (hi)(42))
    }
  }
  def prop_unsafeInsertMin_Max_test : Result = QuickCheck.qc1b[int,Assoc[int,int]] (prop_unsafeInsertMin_Max) (gInt) (gAssocIntInt)

  // ------ unsafeFromOrdSeq ----------
  def prop_unsafeFromOrdSeq (xs:List[b]) : boolean = {
    val xs' = sortT (xs)
    equTuple ((IntAssoc.unsafeFromOrdSeq (xs')).toOrdSeq) ((IntAssoc.fromSeq (xs')).toOrdSeq)
  }
  def prop_unsafeFromOrdSeq_test : Result = QuickCheck.qc0b[List[b]] (prop_unsafeFromOrdSeq) (gListIntInt)


  // ------ unsafeAppend ----------
  def prop_unsafeAppend (i:int) (xs:Assoc[int,int]) (ys:Assoc[int,int]) : boolean = {
    def odd (x:int) : boolean = ((x%2) == 1)
    if ( (xs.isEmpty) || (ys.isEmpty) ) {
      equ (xs.unsafeAppend (ys)) (xs.union (ys))
    } else {
      val lo = fst((xs.toOrdSeq).lhead)
      val hi = fst((xs.toOrdSeq).rhead)      
      val delta = hi - lo + ( if (odd (i)) 1 else 0 )
      val ys' = IntAssoc.fromSeq ((ys.toSeq).map[(int,int)] ((x:int,y:int) | (x+delta,y) ))
      equ (xs.unsafeAppend (ys')) (xs.union (ys'))      
    }
  }
  def prop_unsafeAppend_test : Result = QuickCheck.qc2b[int,Assoc[int,int],Assoc[int,int]] (prop_unsafeAppend) (gInt) (gAssocIntInt) (gAssocIntInt)


  // ------ filter ----------
  def prop_filter (x:int) (xs:Assoc[int,int]) : boolean = {
    equTuple ((xs.filterLT (x)).toOrdSeq) ((xs.toOrdSeq).filter ((y,a)|y<x))
      && equTuple ((xs.filterLE (x)).toOrdSeq) ((xs.toOrdSeq).filter ((y,a)|y<=x))
      && equTuple ((xs.filterGT (x)).toOrdSeq) ((xs.toOrdSeq).filter ((y,a)|y>x))
      && equTuple ((xs.filterGE (x)).toOrdSeq) ((xs.toOrdSeq).filter ((y,a)|y>=x))
  }
  def prop_filter_test : Result = QuickCheck.qc1b[int,Assoc[int,int]] (prop_filter) (gInt) (gAssocIntInt)

  // ------ partition ----------
  def prop_partition (x:int) (xs:Assoc[int,int]) : boolean = {
    def equTL (a:Assoc[int,int],b:Assoc[int,int]) (a':Assoc[int,int],b':Assoc[int,int]) : boolean = {
      (equ (a) (a')) && (equ (b) (b')) 
    }
    equTL (xs.partitionLT_GE (x)) ( xs.filterLT (x) , xs.filterGE (x) )
      && equTL (xs.partitionLE_GT (x)) ( xs.filterLE (x) , xs.filterGT (x) )
      && equTL (xs.partitionLT_GT (x)) ( xs.filterLT (x) , xs.filterGT (x) )
  }
  def prop_partition_test : Result = QuickCheck.qc1b[int,Assoc[int,int]] (prop_partition) (gInt) (gAssocIntInt)

  ////////////////////////////////////////////////////////////////////
  ////    OrdAssoc operations
  ////////////////////////////////////////////////////////////////////

  // ------ minElem_maxElem ----------
  def prop_minElem_maxElem (xs:Assoc[int,int]) : Result = {
    implies (!(xs.isEmpty)) ( |
      ((xs.minElem) == snd((xs.toOrdSeq).lhead))
        && ((xs.maxElem) == snd((xs.toOrdSeq).rhead))
    )
  }
  def prop_minElem_maxElem_test : Result = QuickCheck.qc0r[Assoc[int,int]] (prop_minElem_maxElem) (gAssocIntInt)


  // ------ foldr_foldl ----------
  def prop_foldr_foldl (xs:Assoc[int,int]) : boolean = {
    def flip[a,b] (f:a->b->b) : b->a->b = (y:b | x:a | f (x) (y))
    def :: (x:int) (xs:List[int]) = List.cons (x) (xs)
    val a = ((xs.toOrdSeq).map ( (x,y) | y ))
    equList (xs.foldr (::) (List.empty[int])) (a)
      && equList (xs.foldl (flip (::)) (List.empty[int])) (a.reverse)
  }
  def prop_foldr_foldl_test : Result = QuickCheck.qc0b[Assoc[int,int]] (prop_foldr_foldl) (gAssocIntInt)

  ////////////////////////////////////////////////////////////////////
  ////    FiniteMapX operations
  ////////////////////////////////////////////////////////////////////

  // ------ intersect_difference ----------
  def prop_intersectWith_difference (xs:Assoc[int,int]) (ys:Assoc[int,int]) : boolean = {
    equ (xs.intersectWith[int] (x|y|y) (ys)) (ys.filterWithKey (k | a | xs.member(k)))
      && equ (ys.difference (xs)) (ys.filterWithKey (k | a | !(xs.member(k))))
  }
  def prop_intersectWith_difference_test : Result = QuickCheck.qc1b[Assoc[int,int],Assoc[int,int]] (prop_intersectWith_difference) (gAssocIntInt) (gAssocIntInt)

  // ------ subset_subsetEq ----------
  def prop_subset_subsetEq (xs:Assoc[int,int]) (ys:Assoc[int,int]) : boolean = {
    (xs.subset (ys)) == ((xs.subsetEq (ys)) && (!(equ (xs) (ys))))
      && (xs.subsetEq (ys)) == (equ (xs.intersectWith[int] (x|y|x) (ys)) (xs))
  }
  def prop_subset_subsetEq_test : Result = QuickCheck.qc1b[Assoc[int,int],Assoc[int,int]] (prop_subset_subsetEq) (gAssocIntInt) (gAssocIntInt)

  ////////////////////////////////////////////////////////////////////
  ////    FiniteMap operations
  ////////////////////////////////////////////////////////////////////

  // ------ fromSeqWith ----------
  def prop_fromSeqWith (xs:List[b]) : boolean = {
    equ (IntAssoc.fromSeqWith[int] (x|y|x) (xs)) (IntAssoc.fromSeq (xs))
  }
  def prop_fromSeqWith_test : Result = QuickCheck.qc0b[List[b]] (prop_fromSeqWith) (gListIntInt)

  // ------ insertWith ----------
  def prop_insertWith (xy:b) (xs:Assoc[int,int]) : boolean = {
    val x = fst(xy)
    val y = snd(xy)
    equ (xs.insertWith (x|y|x) (x) (y)) (xs.insert (x) (y))
  }
  def prop_insertWith_test : Result = QuickCheck.qc1b[b,Assoc[int,int]] (prop_insertWith) (gIntInt) (gAssocIntInt)

  // ------ insertSeqWith ----------
  def prop_insertSeqWith (xys:List[b]) (xs:Assoc[int,int]) : boolean = {
    equ (xs.insertSeqWith (x|y|x) (xys)) (xs.insertSeq (xys))
  }
  def prop_insertSeqWith_test : Result = QuickCheck.qc1b[List[b],Assoc[int,int]] (prop_insertSeqWith) (gListIntInt) (gAssocIntInt)

  // ------ unionl_unionr_unionWith ----------
  def prop_unionl_unionr_unionWith (xs:Assoc[int,int]) (ys:Assoc[int,int]) : boolean = {
    equ (xs.unionl (ys)) (ys.unionr (xs))
      && equ (xs.unionr (ys)) (ys.unionl (xs))
      && equ (xs.unionWith (x|y|x) (ys)) (xs.unionl (ys))
  }
  def prop_unionl_unionr_unionWith_test : Result = QuickCheck.qc1b[Assoc[int,int],Assoc[int,int]] (prop_unionl_unionr_unionWith) (gAssocIntInt) (gAssocIntInt)

  // ------ unionSeqWith ----------
  def prop_unionSeqWith (ms:List[Assoc[int,int]]) : boolean = {
    equ (IntAssoc.unionSeqWith[int] (x|y|y) (ms)) (IntAssoc.unionSeq[int] (ms))
  }
  def prop_unionSeqWith_test : Result = QuickCheck.qc0b[List[Assoc[int,int]]] (prop_unionSeqWith) (gListAssocIntInt)
 

  // ------------ the test sequence -------------------
  def run = {
    val _ = QuickCheck.quickcheck (|prop_single_test) ("prop_single_test")
    val _ = QuickCheck.quickcheck (|prop_fromSeq_test) ("prop_fromSeq_test")
    val _ = QuickCheck.quickcheck (|prop_insert_test) ("prop_insert_test")
    val _ = QuickCheck.quickcheck (|prop_union_test) ("prop_union_test")
    val _ = QuickCheck.quickcheck (|prop_unionSeq_test) ("prop_unionSeq_test")
    val _ = QuickCheck.quickcheck (|prop_delete_test) ("prop_delete_test")
    val _ = QuickCheck.quickcheck (|prop_deleteAll_test) ("prop_deleteAll_test")
    val _ = QuickCheck.quickcheck (|prop_deleteSeq_test) ("prop_deleteSeq_test")
    val _ = QuickCheck.quickcheck (|prop_isEmpty_size_test) ("prop_isEmpty_size_test")
    val _ = QuickCheck.quickcheck (|prop_member_count_test) ("prop_member_count_test")
    val _ = QuickCheck.quickcheck (|prop_toSeq_test) ("prop_toSeq_test")
    val _ = QuickCheck.quickcheck (|prop_lookup_test) ("prop_lookup_test")
    val _ = QuickCheck.quickcheck (|prop_fold_test) ("prop_fold_test")
    val _ = QuickCheck.quickcheck (|prop_filter_partition_test) ("prop_filter_partition_test")
    val _ = QuickCheck.quickcheck (|prop_deleteMin_Max_test) ("prop_deleteMin_Max_test")
    val _ = QuickCheck.quickcheck (|prop_unsafeInsertMin_Max_test) ("prop_unsafeInsertMin_Max_test")
    val _ = QuickCheck.quickcheck (|prop_unsafeFromOrdSeq_test) ("prop_unsafeFromOrdSeq_test")
    val _ = QuickCheck.quickcheck (|prop_unsafeAppend_test) ("prop_unsafeAppend_test")
    val _ = QuickCheck.quickcheck (|prop_filter_test) ("prop_filter_test")
    val _ = QuickCheck.quickcheck (|prop_partition_test) ("prop_partition_test")
    val _ = QuickCheck.quickcheck (|prop_minElem_maxElem_test) ("prop_minElem_maxElem_test")
    val _ = QuickCheck.quickcheck (|prop_foldr_foldl_test) ("prop_foldr_foldl_test")
    val _ = QuickCheck.quickcheck (|prop_intersectWith_difference_test) ("prop_intersectWith_difference_test")
    val _ = QuickCheck.quickcheck (|prop_subset_subsetEq_test) ("prop_subset_subsetEq_test")
    val _ = QuickCheck.quickcheck (|prop_fromSeqWith_test) ("prop_fromSeqWith_test")
    val _ = QuickCheck.quickcheck (|prop_insertWith_test) ("prop_insertWith_test")
    val _ = QuickCheck.quickcheck (|prop_insertSeqWith_test) ("prop_insertSeqWith_test")
    val _ = QuickCheck.quickcheck (|prop_unionl_unionr_unionWith_test) ("prop_unionl_unionr_unionWith_test")
    val _ = QuickCheck.quickcheck (|prop_unionSeqWith_test) ("prop_unionSeqWith_test")

  }

}


// -------- the test run -----------------

QTAssoc.run
