// -*- Funnel -*-

/**
 * $Author: ysaidji $
 * $Id: RedBlackTree.m4,v 1.3 2001/09/12 13:25:11 ysaidji Exp $
 */

def removeMe0 = include[RedBlackTreeType]
def removeMe1 = include[__AssocDefaultType__]
def removeMe2 = include[__AssocDefault__]

def RedBlackTree[k] (equ: Equ[k], ord: Ord[k]) : RedBlackTreeModule[k] = {

  let this = {

    // inherence
    with AssocXDefault[k](this)
    with AssocDefault[k](this)
    with FiniteMapXDefault[k](this)
    with FiniteMapDefault[k](this)

    // -------- static methods of AssocXModule --------
    // Comparison
    def equivalent (x) (y) = equ.equ (x) (y)
    // Constructors
    def empty[a] = Nil[a]
    def single[a] (key) (x:a) = Tree[a](BLACK, Nil, key, x, Nil)
    def unionSeq[a] (ms) = unionSeqUsingReduce (ms)
    def fromSeq[a] (kxs) = fromSeqUsingInsertSeq (kxs)

    // -------- static methods of OrdAssocXModule --------
    // Comparison
    def LT = 0
    def GT = 1
    def EQ = 2
    def compare (x) (y) = if (ord.lt (x) (y)) LT else if (ord.lt (y) (x)) GT else EQ

    def unsafeFromOrdSeq (kxs) = fromSeq (kxs)

    // -------- static methods of FiniteMapXModule --------
    def unionSeqWith[a] (f) (ms) = unionSeqWithUsingReduce (f) (ms)
    def fromSeqWith[a] (f: a -> a -> a) (kvs:__Seq__[(k,a)]) = fromSeqWithUsingInsertSeqWith (f) (kvs)
    def fromSeqWithKey[a] (f: k -> a -> a -> a) (kvs:__Seq__[(k,a)]) = fromSeqWithKeyUsingInsertSeqWithKey (f) (kvs)

    // -------- static methods of FiniteMapModule --------
    def unionSeqWithKey[a] (f) (ms) = unionSeqWithKeyUsingReduce (f) (ms)

    // -------- Private static methods --------

    def BLACK : boolean = false
    def RED : boolean = true

    // ----------- Nil and Cons ----------

    def Nil[a] = {
      let this : RedBlackTree[k,a] = {

	// ----------- private methods -----------
	def outputPS (uid:int) = null
	def outputGraph (uid:int) (depth:int) (id:int) = (false,id)

	// accessors
        val color = BLACK
	val a = this
	val k = null
	val x = null
	val b = this

        // ----------- AssocX methods -----------
        // Constructors
	def insert (k:k) (x:a) = single[a] (k) (x)
        def insertSeq (kxs:__Seq__[(k,a)]) = insertSeqUsingFoldr (kxs) (this)
        def union (m:__AssocX__[k,a]) = m

        // Deletions
        def delete (key:k) = this
        def deleteAll (key:k) = this
        def deleteSeq (ks:__Seq__[k]) = this

        // Observers
        def isEmpty = true
        def size = 0
        def member (key:k) = false
        def count (key:k) = 0
        def lookup (key:k) = System.error "RedBlackTree.lookup: lookup failed"
        def lookupM (key:k) = (false,null)
        def lookupAll (key:k) = List.empty[a]
        def lookupWithDefault (key:k) (d:a) = d
        def elements = List.empty[a]

        // Modifiers
        def adjust (f: a -> a) (key:k) = this
        def adjustAll (f: a -> a) (key:k) = this

        // Maps and folds
        def map[b] (f: a -> b) = Nil[b]
        def fold[b] (f: a -> b -> b) (c:b) = c
        def fold1 (f: a-> a -> a) = System.error "RedBlackTree.fold1: empty map"
	def iterate (f: a->()) = null

        // Filters and partitions
        def filter (p: a -> boolean) = this
        def partition (p: a -> boolean) = (this,this)
	def exist (p: a -> boolean) = false

	// ----------- OrdAssocX methods -----------
	// Constructors
	def unsafeInsertMin (key:k) (x:a) = insert (key) (x)
	def unsafeInsertMax (key:k) (x:a) = insert (key) (x)
	def unsafeAppend (m:__OrdAssocX__[k,a]) = m

	// Destructors
	//def minView
	def minElem = System.error "RedBlackTree.minElem: empty map"
	//def maxView
	def maxElem = System.error "RedBlackTree.maxElem: empty map"

	// Deletions
	def deleteMin = this
	def deleteMax = this

	// Maps and folds
	def foldr[b] (f: a -> b -> b) (c:b) = c
	def foldl[b] (f: b -> a -> b) (c:b) = c
	def foldr1 (f: a-> a -> a) = System.error "RedBlackTree.foldr1: empty map"
	def foldl1 (f: a-> a -> a) = System.error "RedBlackTree.foldl1: empty map"
	def iterater (f:a->()) = null
	def iteratel (f:a->()) = null

	// Filters and partitions
	def filterLT (key:k) = this
	def filterLE (key:k) = this
	def filterGT (key:k) = this
	def filterGE (key:k) = this
	def partitionLT_GE (key:k) = (this,this)
	def partitionLE_GT (key:k) = (this,this)
	def partitionLT_GT (key:k) = (this,this)

        // ----------- Assoc methods -----------
        // Observers
        def toSeq = List.empty[(k,a)]
        def keys = List.empty[k]

        // Maps and folds
        def mapWithKey[b] (f: k -> a -> b) = Nil[b]
        def foldWithKey[b] (f: k -> a -> b -> b) (c:b) = c
	def iterateWithKey (f:k->a->()) = null

        // Filters and partitions
        def filterWithKey (p: k -> a -> boolean) = this
        def partitionWithKey (p: k -> a -> boolean) = (this,this)
	def existWithKey (p:k->a->boolean) = false

	// ----------- OrdAssoc methods -----------
	// Destructors
	//def minViewWithKey
	def minElemWithKey = System.error "RedBlackTree.minElemWithKey: empty map"
	//def maxViewWithKey
	def maxElemWithKey = System.error "RedBlackTree.maxElemWithKey: empty map"
	
	// Observers
	def toOrdSeq = List.empty[(k,a)]

	// Maps and folds
	def foldrWithKey[b] (f: k -> a -> b -> b) (c:b) = c
	def foldlWithKey[b] (f: b -> k -> a -> b) (c:b) = c
	def iteraterWithKey (f: k -> a -> ()) = null
	def iteratelWithKey (f: k -> a -> ()) = null

        // ----------- FiniteMapX methods -----------
        // Constructors
        def insertWith (f: a -> a -> a) (k:k) (x:a) = insert (k) (x)
        def insertWithKey (f: k -> a -> a -> a) (k:k) (x:a) = insert (k) (x)

        def insertSeqWith (f: a -> a -> a) (kvs:__Seq__[(k,a)]) =  insertSeqWithUsingInsertWith (f) (kvs) (this)
        def insertSeqWithKey (f: k -> a -> a -> a) (kvs:__Seq__[(k,a)]) = insertSeqWithKeyUsingInsertWithKey (f) (kvs) (this) 
        def unionl (m:__FiniteMapX__[k,a]) = m
        def unionr (m:__FiniteMapX__[k,a]) = m
        def unionWith (f: a -> a -> a) (m:__FiniteMapX__[k,a]) = m

        // Set-like operations
        def intersectWith[b,c] (f: a -> b -> c) (m:__FiniteMapX__[k,b]) = empty[c]
        def difference[b] (m:__FiniteMapX__[k,b]) = this
        def subset[b] (m:__FiniteMapX__[k,b]) = if (!(m.isEmpty)) true else false
        def subsetEq[b] (m:__FiniteMapX__[k,b]) = true

        // ----------- FiniteMap methods -----------
        // Constructors
        def unionWithKey (f: k -> a -> a -> a) (m:__FiniteMap__[k,a]) = m

        // Set-like operations
        def intersectWithKey[b,c] (f: k -> a -> b -> c) (m:__FiniteMapX__[k,b]) = empty[c]
      }
      this
    }

    def Tree[a] (color':boolean, a':RedBlackTree[k,a], k':k, x':a, b':RedBlackTree[k,a]) : RedBlackTree[k,a] = {
      let this : RedBlackTree[k,a] = {

	// ----------- private methods -----------
	// accessors
        val color = color'
	val a = a'
	val k = k'
	val x = x'
	val b = b'

	def outputPS (uid:int) = {
	  val _ = outputGraph (uid) (0) (1)
	  null
	}

	def outputGraph (uid:int) (depth:int) (id:int) = {

	  Console.println( "D" + uid + "-Name " + id + " " + k + "," + x)
	  
	  if (BLACK == color) {
	    Console.println( "A" + uid + "-Sta " + depth + " " + id )
	  } else {
	    Console.println( "B" + uid + "-Ter " + depth + " " + id )
	  }
	  
	  val (ba,aid) = (a.outputGraph) (uid) (depth+1) (id+1)
	  if (ba) {
	    Console.println( "C" + uid + "-Pri 32 " + id + " " + (id+1) )
	  } else {
	  }
	  
	  val (bb,bid) = (b.outputGraph) (uid) (depth+1) (aid+1)
	  if (bb) {
	    Console.println( "C" + uid + "-Pri 32 " + id + " " + (aid+1) )
	  } else {
	  }

	  (true,bid)

	}


	def balance (color:boolean, a:RedBlackTree[k,a], tk:k, tx:a, b:RedBlackTree[k,a]) : RedBlackTree[k,a] = {
	  
	  type T = RedBlackTree[k,a]

	  let body = Tree(color,a,tk,tx,b)

	  def balanced(a:T, b:T, c:T, d:T, xk:k, xx:a, yk:k, yx:a, zk:k, zx:a) = {
	    val x = Tree(BLACK, a, xk, xx, b)
	    val z = Tree(BLACK, c, zk, zx, d)
	    Tree(RED, x, yk, yx, z)
	  }

	  if (BLACK == color) {
	    if ( ( RED == a.color ) && ( RED == b.color ) ) {
	      Tree(RED, (Tree(BLACK,a.a,a.k,a.x,a.b)), tk, tx, (Tree(BLACK,b.a,b.k,b.x,b.b)))
	    } else if ( ( RED == a.color ) && ( RED == (a.a).color ) ) {
	      val y = a
	      val x = y.a
	      balanced( x.a, x.b, y.b, b, x.k, x.x, y.k, y.x, tk, tx )
	    } else if ( ( RED == a.color ) && ( RED == (a.b).color ) ) {
	      val x = a
	      val y = x.b
	      balanced( x.a, y.a, y.b, b, x.k, x.x, y.k, y.x, tk, tx )
	    } else if ( ( RED == b.color ) && ( RED == (b.a).color ) ) {
	      val z = b
	      val y = z.a
	      balanced( a, y.a, y.b, z.b, tk, tx, y.k, y.x, z.k, z.x )
	    } else if ( ( RED == b.color ) && ( RED == (b.b).color ) ) {
	      val y = b
	      val z = y.b
	      balanced( a, y.a, z.a, z.b, tk, tx, y.k, y.x, z.k, z.x )
	    } else {
	      body
	    }
	  } else {
	    body
	  }

	}

        // ----------- AssocX methods -----------
        // Constructors
	def insert (key:k) (sat:a) = insertWith ( x | y | x ) (key) (sat)

        def insertSeq (kxs:__Seq__[(k,a)]) = insertSeqUsingFoldr (kxs) (this)
        def union (m:__AssocX__[k,a]) = m.foldWithKey ( k | x | m:RedBlackTree[k,a] | m.insert (k) (x)) (this)

        // Deletions
        def delete (key:k) = {

	  type T = RedBlackTree[k,a]

	  def app (a:T) (b:T) : T = {
	    if (a.isEmpty) b
	    else if (b.isEmpty) a
	    else if ((RED == (a.color)) && (RED == (b.color))) {
	      val t = app (a.b) (b.a)
	      if (RED == (t.color)) Tree(RED, (Tree(RED,a.a,a.k,a.x,t.a)), t.k, t.x, (Tree(RED,t.b,b.k,b.x,b.b)))
	      else Tree(RED,(a.a),(a.k),(a.x),(Tree(RED,t,b.k,b.x,b.b)))	      
	    } else if ((BLACK == (a.color)) && (BLACK == (b.color))) {
	      val t = app (a.b) (b.a)
	      if (RED == (t.color)) Tree(RED, (Tree(BLACK,a.a,a.k,a.x,t.a)), t.k, t.x, (Tree(BLACK,t.b,b.k,b.x,b.b)))
	      else balleft (a.a) (a.k) (a.x) (Tree(BLACK,t,b.k,b.x,b.b))
	    } else if (RED == (b.color)) {
	      Tree(RED,(app (a)(b.a)),b.k,b.x,b.b)
	    } else {
	      Tree(RED,a.a,a.k,a.x,(app (a.b) (b)))
	    }
	  }

	  def sub1 (t:T) : T = {
	    if ((!(t.isEmpty)) && (BLACK == (t.color))) Tree(RED,t.a,t.k,t.x,t.b)
	    else {	      
	      System.error ("invariance violation" )
	    }
	  }

	  def balleft (bl:T) (k:k) (x:a) (t:T) : T = {
	    if (RED == (bl.color)) Tree( RED,  Tree(BLACK,bl.a,bl.k,bl.x,bl.b), k, x, t )
	    else if (BLACK == (t.color)) balance ((BLACK), (bl), (k), (x), (Tree(RED,t.a,t.k,t.x,t.b)))
	    else {
	      val a  = (t.a).a
	      val yk = (t.a).k
	      val yx = (t.a).x
	      val b  = (t.a).b
	      val zk = t.k
	      val zx = t.x
	      val c  = t.b
	      val newa = Tree(BLACK,bl,k,x,a)
	      val newb = balance ((BLACK), (b), (zk), (zx), (sub1 (c)))
	      Tree(RED,newa,yk,yx,newb)
	    }
	  }

	  def balright (a:T) (k:k) (x:a) (bl:T) : T = {
	    if (RED == (bl.color)) Tree(RED,a,k,x,Tree(BLACK,bl.a,bl.k,bl.x,bl.b))
	    else if (BLACK == (a.color)) balance ((BLACK), (Tree(RED,a.a,a.k,a.x,a.b)), (k), (x), (bl))
	    else {
	      val a' = a.a
	      val xk = a.k
	      val xx = a.x
	      val b  = (a.b).a
	      val yk = (a.b).k
	      val yx = (a.b).x
	      val c  = (a.b).b
	      val zk = k
	      val zx = x
	      val newa = balance ((BLACK), (sub1 (a')), (xk), (xx), (b))
	      val newb = Tree(BLACK,c,zk,zx,bl)
	      Tree(RED,newa,yk,yx,newb)
	    }
	  }


	  def delfromLeft (a:T) (y:k) (x:a) (b:T) : T = {
	    if ( (!(a.isEmpty)) && (BLACK == (a.color))) balleft (del (a)) (y) (x) (b)
	    else Tree(RED, del (a), y, x, b)
	  }

	  def delfromRight (a:T) (y:k) (x:a) (b:T) : T = {
	    if ((!(b.isEmpty)) && (BLACK == (b.color))) balright (a) (y) (x) (del (b))
	    else Tree(RED, a, y, x, del (b))
	  }

	  def del (t:T) : T = {
	    if (t.isEmpty) t
	    else {
	      val d = compare (key) (t.k)
	      if (LT == d) delfromLeft (t.a) (t.k) (t.x) (t.b)
	      else if (GT == d) delfromRight (t.a) (t.k) (t.x) (t.b)
	      else app (t.a) (t.b)
	    }
	  }

	  val r = del (this)
	  if (r.isEmpty) r else Tree(BLACK, r.a, r.k, r.x, r.b)
	}


        def deleteAll (key:k) = delete (key)
        def deleteSeq (ks:__Seq__[k]) = deleteSeqUsingFoldr[a] (ks) (this)

        // Observers
        def isEmpty = false
        def size = 1 + (a.size) + (b.size)

        def member (key:k) = {
	  val (b,x) = lookupM (key)
	  b
	}

        def count (key:k) = countUsingMember (this) (key)

        def lookup (key:k) = {
	  val d = compare (key) (k)
	  if (LT == d) (a.lookup) (key)
	  else if (GT == d) (b.lookup) (key)
	  else x
	}

        def lookupM (key:k) = {
	  val d = compare (key) (k)
	  if (LT == d) (a.lookupM) (key)
	  else if (GT == d) (b.lookupM) (key)
	  else (true,x)	  
	}

        def lookupAll (key:k) = lookupAllUsingLookupM (this) (key)
        def lookupWithDefault (key:k) (d:a) = lookupWithDefaultUsingLookupM (d) (this) (key)
        def elements = elementsUsingFold (this)

        // Modifiers
        def adjust (f: a -> a) (key:k) = insertWith ( x | y | f (y) ) (key) (null)
        def adjustAll (f: a -> a) (key:k) = adjust (f) (key)

        // Maps and folds
        def map[b] (f: a -> b) = {
	  val a' = (a.map) (f)
	  val b' = (b.map) (f)
	  Tree[b](color,a',k,f (x),b')
	}

        def fold[b] (f: a -> b -> b) (c:b) = {
	  val ra = (a.fold) (f) (c)
	  val r = f (x) (ra)
	  (b.fold) (f) (r)
	}

        def fold1 (f: a-> a -> a) = {
	  val ra = (a.fold) (f) (x)
	  (b.fold) (f) (ra)
	}

	def iterate (f: a -> ()) = {
	  (a.iterate) (f)
	  f (x)
	  (b.iterate) (f)
	}

        // Filters and partitions
        def filter (p: a -> boolean) = foldWithKey (k | x | m:RedBlackTree[k,a] | if (p (x)) (m.insert (k) (x)) else m) (empty[a])
        def partition (p: a -> boolean) = partitionUsingFilter[a] (p) (this)
	def exist (p: a -> boolean) = {
	  (p (x) || (a.exist) (p)) || (b.exist) (p)
	}


	// ----------- OrdAssocX methods -----------
	// Constructors
	def unsafeInsertMin (key:k) (x:a) = insert (key) (x)
	def unsafeInsertMax (key:k) (x:a) = insert (key) (x)
	def unsafeAppend (m:__OrdAssocX__[k,a]) = union (m)

	// Destructors
	//def minView

	def minElem = if (a.isEmpty) x else (a.minElem)

	//def maxView

	def maxElem = if (b.isEmpty) x else (b.maxElem)

	// Deletions
	def deleteMin = {
	  def minKey (m:RedBlackTree[k,a]) (k:k) : k = if (m.isEmpty) k else minKey (m.a) (m.k)
	  delete (minKey (a) (k))
	}

	def deleteMax = {
	  def maxKey (m:RedBlackTree[k,a]) (k:k) : k = if (m.isEmpty) k else maxKey (m.b) (m.k)
	  delete (maxKey (b) (k))
	}

	// Maps and folds
	def foldr[b] (f: a -> b -> b) (c:b) = {
	  val rb = (b.foldr) (f) (c)
	  val r = f (x) (rb)
	  (a.foldr) (f) (r)
	}

	def foldl[b] (f: b -> a -> b) (c:b) = {
	  val ra = (a.foldl) (f) (c)
	  val r = f (ra) (x)
	  (b.foldl) (f) (r)	  
	}

	def foldr1 (f: a-> a -> a) = if (b.isEmpty) (a.foldr) (f) (x) else (a.foldr) (f) (f (x) (b.foldr1 (f)))
	def foldl1 (f: a-> a -> a) = if (a.isEmpty) (b.foldl) (f) (x) else (b.foldl) (f) (f (a.foldl1 (f)) (x))

	def iterater (f:a->()) = {
	  (b.iterater) (f)
	  f (x)
	  (a.iterater) (f)
	}

	def iteratel (f:a->()) = {
	  (a.iterater) (f)
	  f (x)
	  (b.iterater) (f)
	}

	// Filters and partitions
	def filterLT (key:k) = {
	  val d = compare (k) (key)
	  if (LT == d) ((a.insert) (k) (x)).union (b.filterLT (key)) else a.filterLT (key)
	}

	def filterLE (key:k) = {
	  val d = compare (k) (key)
	  if ((LT == d) || (EQ == d)) ((a.insert) (k) (x)).union (b.filterLE (key)) else (a.filterLE) (key)	  
	}

	def filterGT (key:k) = {
	  val d = compare (k) (key)
	  if (GT == d) (b.insert (k) (x)).union (a.filterGT (key)) else (b.filterGT) (key)	  
	}

	def filterGE (key:k) = {
	  val d = compare (k) (key)
	  if ((GT == d) || (EQ == d)) ((b.insert) (k) (x)).union (a.filterGE (key)) else (b.filterGE) (key)	  
	}

	def partitionLT_GE (key:k) = ( filterLT (key), filterGE (key) )
	def partitionLE_GT (key:k) = ( filterLE (key), filterGT (key) )
	def partitionLT_GT (key:k) = ( filterLT (key), filterGT (key) )

        // ----------- Assoc methods -----------
        // Observers
        def toSeq = toSeqUsingFoldWithKey[a] (this)
        def keys = keysUsingFoldWithKey[a] (this)

        // Maps and folds
        def mapWithKey[b] (f: k -> a -> b) = {
	  val a' = (a.mapWithKey) (f)
	  val b' = (b.mapWithKey) (f)
	  Tree[b](color,a',k,f (k) (x),b')
	}

        def foldWithKey[b] (f: k -> a -> b -> b) (c:b) = {
	  val ra = (a.foldWithKey) (f) (c)
	  val r = f (k) (x) (ra)
	  (b.foldWithKey) (f) (r)
	}

	def iterateWithKey (f:k->a->()) = {
	  (a.iterateWithKey) (f)
	  f (k) (x)
	  (b.iterateWithKey) (f)	  
	}

        // Filters and partitions
        def filterWithKey (p: k -> a -> boolean) = foldWithKey (k | x | m:RedBlackTree[k,a] | if (p (k) (x)) (m.insert (k) (x)) else m) (empty[a])

        def partitionWithKey (p: k -> a -> boolean) = (filterWithKey (p) , filterWithKey ( k | x | (!(p (k) (x)))))

	def existWithKey (p: k -> a -> boolean) = {
	  (p (k) (x) || (a.existWithKey) (p)) || (b.existWithKey) (p)
	}

	// ----------- OrdAssoc methods -----------
	// Destructors
	//def minViewWithKey
	def minElemWithKey = if (a.isEmpty) (k,x) else (a.minElemWithKey)
	//def maxViewWithKey
	def maxElemWithKey = if (b.isEmpty) (k,x) else (b.maxElemWithKey)
	
	// Observers
	def toOrdSeq = foldrWithKey ( k | x | l:List[(k,a)] | List.cons (k,x) (l) ) (List.Nil[(k,a)])

	// Maps and folds
	def foldrWithKey[b] (f: k -> a -> b -> b) (c:b) = {
	  val rb = (b.foldrWithKey) (f) (c)
	  val r = f (k) (x) (rb)
	  (a.foldrWithKey) (f) (r)
	}

	def foldlWithKey[b] (f: b -> k -> a -> b) (c:b) = {
	  val ra = (a.foldlWithKey) (f) (c)
	  val r = f (ra) (k) (x)
	  (b.foldlWithKey) (f) (r)	  
	}

	
	def iteraterWithKey (f:k->a->()) = {
	  (b.iteraterWithKey) (f)
	  f (k) (x)
	  (a.iteraterWithKey) (f)
	}

	def iteratelWithKey (f:k->a->()) = {
	  (a.iteratelWithKey) (f)
	  f (k) (x)
	  (b.iteratelWithKey) (f)
	}

        // ----------- FiniteMapX methods -----------
        // Constructors
        def insertWith (f: a -> a -> a) (key:k) (sat:a) = {

	  def ins (s:RedBlackTree[k,a]) : RedBlackTree[k,a] = {

	    if (s.isEmpty) {
	      Tree[a](RED,s,key,sat,s)
	    } else {
	      val d = compare (key) (s.k)
	      
	      if      (LT == d) balance (s.color, ins (s.a), s.k, s.x, s.b)
	      else if (GT == d) balance (s.color, s.a, s.k, s.x, ins (s.b))
	      else              Tree[a] (s.color, s.a, s.k, f (sat) (s.x), s.b)
	    }
	  }
	  
	  val r = ins (this)
	  Tree(BLACK, r.a, r.k, r.x, r.b)

	}



        def insertWithKey (f: k -> a -> a -> a) (k:k) (x:a) = insertWithKeyUsingInsertWith (f) (k) (x) (this)
        def insertSeqWith (f: a -> a -> a) (kvs:__Seq__[(k,a)]) =  insertSeqWithUsingInsertWith (f) (kvs) (this)
        def insertSeqWithKey (f: k -> a -> a -> a) (kvs:__Seq__[(k,a)]) = insertSeqWithKeyUsingInsertWithKey (f) (kvs) (this) 
        def unionl (m:__FiniteMapX__[k,a]) = unionWith ( x | y | (x) ) (m)
        def unionr (m:__FiniteMapX__[k,a]) = unionWith ( x | y | (y) ) (m)
        def unionWith (f: a -> a -> a) (m:__FiniteMapX__[k,a]) = unionWithUsingInsertWith (f) (this) (m)

        // Set-like operations
        def intersectWith[b,c] (f: a -> b -> c) (m:__FiniteMapX__[k,b]) = intersectWithUsingLookupM[a,b,c] (f) (this) (m)
	


        def difference[b] (m:__FiniteMapX__[k,b]) = differenceUsingDelete[a,b] (this) (m)
        def subset[b] (m:__FiniteMapX__[k,b]) = subsetUsingSubsetEq[a,b] (this) (m)
        def subsetEq[b] (m:__FiniteMapX__[k,b]) = subsetEqUsingMember[a,b] (this) (m)

        // ----------- FiniteMap methods -----------
        // Constructors
        def unionWithKey (f: k -> a -> a -> a) (m:__FiniteMap__[k,a]) = unionWithKeyUsingInsertWithKey (f) (this) (m)

        // Set-like operations
        def intersectWithKey[b,c] (f: k -> a -> b -> c) (m:__FiniteMapX__[k,b]) = intersectWithKeyUsingLookupM[a,b,c] (f) (this) (m)
      }
      this

    }
   

  }
  this

}
