// -*- Funnel -*-

/**
 * $Author: ysaidji $
 * $Id: RedBlackTreeType.m4,v 1.2 2001/09/12 13:25:10 ysaidji Exp $
 */

newtype RedBlackTree[k,a] = {

  //with MaybeType

  // ----------- private methods -----------
  def color : boolean
  def a : RedBlackTree[k,a]
  def k : k
  def x : a
  def b : RedBlackTree[k,a]

  def outputPS (uid:int) : ()
  def outputGraph (uid:int) (depth:int) (id:int) : (boolean,int)

  // ----------- AssocX methods -----------
  // Constructors
  def insert (key:k) (x:a) : RedBlackTree[k,a]
  def insertSeq (kxs:__Seq__[(k,a)]) : RedBlackTree[k,a]
  def union (m:__AssocX__[k,a]) : RedBlackTree[k,a]

  // Deletions
  def delete (key:k) : RedBlackTree[k,a]
  def deleteAll (key:k) : RedBlackTree[k,a]
  def deleteSeq (ks:__Seq__[k]) : RedBlackTree[k,a]

  // Observers
  def isEmpty: boolean
  def size: int
  def member (key:k) : boolean
  def count (key:k) : int
  def lookup (key:k) : a
  def lookupM (key:k) : (boolean,a) // fixme Maybe[a]
  def lookupAll (key:k) : __Seq__[a]
  def lookupWithDefault (key:k) (d:a) : a
  def elements : __Seq__[a]

  // Modifiers
  def adjust (f: a -> a) (key:k) : RedBlackTree[k,a]
  def adjustAll (f: a -> a) (key:k) : RedBlackTree[k,a]

  // Maps and folds
  def map[b] (f: a -> b) : RedBlackTree[k,b]
  def fold[b] (f: a -> b -> b) (c:b) : b
  def fold1 (f: a->a->a) : a
  def iterate (f: a->()) : ()

  // Filters and partitions
  def filter (p: a -> boolean) : RedBlackTree[k,a]
  def partition (p: a -> boolean) : (RedBlackTree[k,a],RedBlackTree[k,a])
  def exist (p: a -> boolean) : boolean

  // ----------- OrdAssocX methods -----------
  // Constructors
  def unsafeInsertMin (key:k) (x:a) : RedBlackTree[k,a]
  def unsafeInsertMax (key:k) (x:a) : RedBlackTree[k,a]
  def unsafeAppend (m:__OrdAssocX__[k,a]) : RedBlackTree[k,a]

  // Destructors
  //def minView
  def minElem : a
  //def maxView
  def maxElem : a

  // Deletions
  def deleteMin :  RedBlackTree[k,a]
  def deleteMax :  RedBlackTree[k,a]

  // Maps and folds
  def foldr[b] (f: a -> b -> b) (c:b) : b
  def foldl[b] (f: b -> a -> b) (c:b) : b
  def foldr1 (f: a-> a -> a) : a
  def foldl1 (f: a-> a -> a) : a
  def iterater (f: a -> ()) : ()
  def iteratel (f: a -> ()) : ()

  // Filters and partitions
  def filterLT (key:k) : RedBlackTree[k,a]
  def filterLE (key:k) : RedBlackTree[k,a]
  def filterGT (key:k) : RedBlackTree[k,a]
  def filterGE (key:k) : RedBlackTree[k,a]
  def partitionLT_GE (key:k) : (RedBlackTree[k,a],RedBlackTree[k,a])
  def partitionLE_GT (key:k) : (RedBlackTree[k,a],RedBlackTree[k,a])
  def partitionLT_GT (key:k) : (RedBlackTree[k,a],RedBlackTree[k,a])
 
  // ----------- Assoc methods -----------
  // Observers
  def toSeq : __Seq__[(k,a)]
  def keys : __Seq__[k]

  // Maps and folds
  def mapWithKey[b] (f: k -> a -> b) : RedBlackTree[k,b]
  def foldWithKey[b] (f: k -> a -> b -> b) (c:b) : b
  def iterateWithKey (f: k -> a -> ()) : ()

  // Filters and partitions
  def filterWithKey (p: k -> a -> boolean) : RedBlackTree[k,a]
  def partitionWithKey (p: k -> a -> boolean) : (RedBlackTree[k,a], RedBlackTree[k,a])
  def existWithKey (p: k -> a -> boolean) : boolean

  // ----------- OrdAssoc methods -----------
  // Destructors
  //def minViewWithKey
  def minElemWithKey : (k,a)
  //def maxViewWithKey
  def maxElemWithKey : (k,a)

  // Observers
  def toOrdSeq : __Seq__[(k,a)]

  // Maps and folds
  def foldrWithKey[b] (f: k -> a -> b -> b) (c:b) : b
  def foldlWithKey[b] (f: b -> k -> a -> b) (c:b) : b
  def iteraterWithKey (f: k -> a -> ()) : ()
  def iteratelWithKey (f: k -> a -> ()) : ()

  // ----------- FiniteMapX methods -----------
  // Constructors
  def insertWith (f: a -> a -> a) (k:k) (x:a) : RedBlackTree[k,a]
  def insertWithKey (f: k -> a -> a -> a) (k:k) (x:a) : RedBlackTree[k,a]
  def insertSeqWith (f: a -> a -> a) (kvs:__Seq__[(k,a)]) : RedBlackTree[k,a]
  def insertSeqWithKey (f: k -> a -> a -> a) (kvs:__Seq__[(k,a)]) : RedBlackTree[k,a]
  def unionl (m:__FiniteMapX__[k,a]) : RedBlackTree[k,a]
  def unionr (m:__FiniteMapX__[k,a]) : RedBlackTree[k,a]
  def unionWith (f: a -> a -> a) (m:__FiniteMapX__[k,a]) : RedBlackTree[k,a]
 
  // Set-like operations
  def intersectWith[b,c] (f: a -> b -> c) (m:__FiniteMapX__[k,b]) : RedBlackTree[k,c]
  def difference[b] (m:__FiniteMapX__[k,b]) : RedBlackTree[k,a]
  def subset[b] (m:__FiniteMapX__[k,b]) : boolean
  def subsetEq[b] (m:__FiniteMapX__[k,b]) : boolean

  // ----------- FiniteMap methods -----------
  // Constructors
  def unionWithKey (f: k -> a -> a -> a) (m:__FiniteMap__[k,a]) : RedBlackTree[k,a]

  // Set-like operations
  def intersectWithKey[b,c] (f: k -> a -> b -> c) (m:__FiniteMapX__[k,b]) : RedBlackTree[k,c]

 

}

type RedBlackTreeModule[k] = {

  // -------- Constructors -------- 
  def Nil[a] : RedBlackTree[k,a]
  def Tree[a] (red:boolean, a:RedBlackTree[k,a], k:k, x:a, b:RedBlackTree[k,a]) : RedBlackTree[k,a]

  // -------- static methods of AssocXModule --------
  // Comparison
  def equivalent (x:k) (y:k) : boolean

  // Constructors
  def empty[a] : RedBlackTree[k,a]
  def single[a] (key:k) (x:a) : RedBlackTree[k,a]
  def unionSeq[a] (ms:__Seq__[__AssocX__[k,a]]) : RedBlackTree[k,a]
  def fromSeq[a] (kxs:__Seq__[(k,a)]) : RedBlackTree[k,a]

  // -------- static methods of OrdAssocXModule --------
  // Comparison
  def LT : int
  def GT : int
  def EQ : int
  def compare (x:k) (y:k): int // should return LT, GT ou EQ

  def unsafeFromOrdSeq[a] (kxs:__Seq__[(k,a)]) : RedBlackTree[k,a]

  // -------- static methods of FiniteMapXModule --------
  def unionSeqWith[a] (f: a -> a -> a) (ms:__Seq__[__FiniteMapX__[k,a]]) : RedBlackTree[k,a]
  def fromSeqWith[a] (f: a -> a -> a) (kvs:__Seq__[(k,a)]) : RedBlackTree[k,a]
  def fromSeqWithKey[a] (f: k -> a -> a -> a) (kvs:__Seq__[(k,a)]) : RedBlackTree[k,a]

  // -------- static methods of FiniteMapModule --------
  def unionSeqWithKey[a] (f: k -> a -> a -> a) (ms:__Seq__[__FiniteMap__[k,a]]) : RedBlackTree[k,a]

  // -------- Private static methods --------

}

newtype Equ[k] = {
  def equ (x:k) (y:k): boolean
}

newtype Ord[k] = {
  def lt (x:k) (y:k): boolean
  def leq (x:k) (y:k): boolean
}


newtype RedBlackTreeType = {
  def dummy:any
}
