// -*- Funnel -*-
/*  FUNNEL LIBRARY (EDISON PORT)
**  
**  CollectionDefault -- some default functions for Collections
**  
**  27/04/2001, Yacine Saidji
*/

/**
 * $Author: ysaidji $
 * $Id: CollectionDefault.m4,v 1.1.1.1 2001/06/26 13:20:20 ysaidji Exp $
 */

def removeMe = include[__CollectionDefaultType__]

// __begin_CollX__
def CollXDefault[a](coll:__CollXModule__[a]): CollXDefaultModule[a] = {

  def insertSeqUsingUnion (xs) (c) = c.union ( coll.fromSeq (xs)  )

  def insertSeqUsingFoldr (xs) (c) = {
    def insert (x:a) (c:__CollX__[a]): __CollX__[a] = c.insert (x)
    xs.foldr insert c
  }

  def deleteSeqUsingDelete (xs) (c) = {
    def delete (x:a) (c:__CollX__[a]): __CollX__[a] = c.delete (x)
    xs.foldr delete c
  }

  def union (h1:__CollX__[a]) (h2:__CollX__[a]) = h1.union (h2)

  def unionSeqUsingFoldl (xss) = xss.foldl union (coll.empty)
  
  def unionSeqUsingReduce (xss) = xss.reducel union (coll.empty)

  def fromSeqUsingFoldr (xs) = {
    def insert (x:a) (c:__CollX__[a]): __CollX__[a] = c.insert (x)
    xs.foldr insert (coll.empty)
  }

  def fromSeqUsingUnionSeq (sm) (xs) = {
    def singleCons (xs:__Seq__[__CollX__[a]]) (x:a) : __Seq__[__CollX__[a]] =
      sm.cons[__CollX__[a]] (coll.single (x)) xs
    coll.unionSeq (xs.foldl singleCons (sm.empty[__CollX__[a]]))
  }
}

// __begin_OrdCollX__
def OrdCollXDefault[a](coll: __OrdCollXModule__[a]): OrdCollXDefaultModule[a] = {

  def unsafeInsertMaxUsingUnsafeAppend (c) (x) = c.unsafeAppend (coll.single (x))

  def unsafeFromOrdSeqUsingUnsafeInsertMin (xs) = xs.foldr (x:a | c:__OrdCollX__[a] | c.unsafeInsertMin (x)) ((coll.empty))

}

// __begin_Coll__
def CollDefault[a](coll:__CollModule__[a]): CollDefaultModule[a] = {

  def memberUsingFold (h) (x) = h.fold ( y:a | ans:boolean | (coll.EQ == coll.compare (x) (y)) || ans) (false)

  def toSeqUsingFold (sm) (c) = c.fold (sm.cons[a]) (sm.empty[a])

  def lookupUsingLookupAll (ys) (x) = {
    val xs = ys.lookupAll (x) ;
    if (xs.isEmpty) System.error "lookup failed" else xs.lhead
  }

  def lookupWithDefaultUsingLookupAll (dflt) (ys) (x) = {
    val xs = ys.lookupAll (x) ;
    if (xs.isEmpty) dflt else xs.lhead
  }
}

// __begin_OrdColl__
def OrdCollDefault[a](coll:__OrdCollModule__[a]): OrdCollDefaultModule[a] = {

  def filterUsingOrdLists (p) (xs) = {
    (coll.unsafeFromOrdSeq) (xs.toOrdSeq.filter (p))
  }

  def partitionUsingOrdList (p) (xs) = {
    val (ys,zs) = xs.toOrdSeq.partition (p);
    ((coll.unsafeFromOrdSeq) (ys), (coll.unsafeFromOrdSeq) (zs))
  }

  def toOrdSeqUsingFoldr (sm) (c) = c.foldr (sm.cons[a]) (sm.empty[a])

  def disjointUsingToOrdList (h1) (h2) = {
    def disj (xs:__Seq__[a]) (ys:__Seq__[a]) : boolean =
      if (xs.isEmpty || ys.isEmpty) true
      else {
	val d = (coll.compare) (xs.lhead) (ys.lhead);
	if ( d == coll.LT )
	  disj (xs.ltail) (ys)
	else if ( d == coll.EQ )
	  false
	else // GT
	  disj (xs) (ys.ltail)
      }
    disj (h1.toOrdSeq) (h2.toOrdSeq)
  }
}

// __begin_SetX__
def SetXDefault[a](coll:__SetXModule__[a]): SetXDefaultModule[a] = {

  def countUsingMember (xs) (x) = if (xs.member (x)) 1 else 0

}

// __begin_Set__
def SetDefault[a](coll:__SetModule__[a]): SetDefaultModule[a] = {

  def fromSeqWithUsingInsertWith (c) (xs) = {
    xs.foldr (x:a | s:__Set__[a] | s.insertWith (c) (x)) (coll.empty)
  }
  
  def insertUsingInsertWith (x) (s) = {
    s.insertWith (x:a | y:a | x) (x)
  }
 
  def unionUsingUnionWith (s1) (s2) = {
    s1.unionWith (x:a | y:a | x) (s2)
  }

  def intersectUsingIntersectWith (s1) (s2) = { 
    s1.insertWith (x:a | y:a | x) (s2)
  }

  def unionSeqWithUsingReducer (c) (cs) = {
    cs.reducer (x:a | s:__Set__[a] | s.unionWith (c) (x)) (coll.empty)
  }

  def insertSeqWithUsingInsertWith (c) (xs) (s) = {
    xs.foldr (x:a | s:__Set__[a] | s.insertWith (c) (x)) s
  }

  def unionlUsingUnionWith (s1) (s2) = {
    s1.unionWith (x:a | y:a | x) (s2) 
  }

  def unionrUsingUnionWith (s1) (s2) = {
    s1.unionWith (x:a | y:a | y) (s2) 
  }
}

// __begin_OrdSet__
def OrdSetDefault[a](coll:__OrdSetModule__[a]): OrdSetDefaultModule[a] = {

  def differenceUsingOrdLists (sm) (xs) (ys) = {
    def diff (xs:__Seq__[a]) (ys:__Seq__[a]) : __Seq__[a] =
      if (ys.isEmpty) xs
      else {
	val d = (coll.compare) (xs.lhead) (ys.lhead);
	if ( d == coll.LT )
	  (sm.cons[a]) (xs.lhead) (diff xs.ltail ys)
	else if ( d == coll.EQ )
	  diff (xs.ltail) (ys.ltail)
	else // GT
	  diff (xs) (ys.ltail)
      }
    (coll.unsafeFromOrdSeq) (diff (xs.toOrdSeq) (ys.toOrdSeq))
  }

  def subsetOnOrdLists (xs:__Seq__[a]) (ys:__Seq__[a]) : boolean = {
    if (ys.isEmpty) false
    else if (xs.isEmpty) true
    else {
      	val d = (coll.compare) (xs.lhead) (ys.lhead);
	if      (d == coll.LT) false
	else if (d == coll.EQ) subsetOnOrdLists (xs.ltail) (ys.ltail)
	else              subsetEqOnOrdLists (xs) (ys.ltail)
    }
  }

  def subsetEqOnOrdLists (xs:__Seq__[a]) (ys:__Seq__[a]) : boolean = {
    if (ys.isEmpty) false
    else if (xs.isEmpty) true
    else {
      	val d = (coll.compare) (xs.lhead) (ys.lhead);
	if      (d == coll.LT) false
	else if (d == coll.EQ) subsetEqOnOrdLists (xs.ltail) (ys.ltail)
	else              subsetEqOnOrdLists (xs) (ys.ltail)
    }
  }

  def subsetUsingOrdLists (xs) (ys) = {
    subsetOnOrdLists (xs.toOrdSeq) (ys.toOrdSeq)
  }
 
  def subsetEqUsingOrdLists (xs) (ys) = {
    subsetEqOnOrdLists (xs.toOrdSeq) (ys.toOrdSeq)
  }

  def unionWithUsingOrdLists (sm) (c) (xs) (ys) = {
    def merge (xs:__Seq__[a]) (ys:__Seq__[a]) : __Seq__[a] = {
      if (xs.isEmpty) ys else if (ys.isEmpty) xs
      else {
	val d = (coll.compare) (xs.lhead) (ys.lhead);
	if      (d == coll.LT) (sm.cons[a]) (xs.lhead) (merge (x.ltail) (ys))
	else if (d == coll.EQ) (sm.cons[a]) (c (xs.lhead) (ys.lhead)) (merge (xs.ltail) (ys.ltail))
	else              (sm.cons[a]) (ys.lhead) (merge (xs) (ys.ltail))        
      }
    }
    (coll.unsafeFromOrdList) (merge (xs.toOrdSeq) (ys.toOrdSeq))
  }

  def intersectWithUsingOrdLists (sm) (c) (xs) (ys) = {
    def inter (xs:__Seq__[a]) (ys:__Seq__[a]) : __Seq__[a] = {
      if (xs.isEmpty || ys.isEmpty) __Seq__[a].empty
      else {
	val d = (coll.compare) (xs.lhead) (ys.lhead);
	if      (d == coll.LT) inter (xs.ltail) ys
	else if (d == coll.EQ) (sm.cons[a]) (c (xs.lhead) (ys.lhead)) (inter (xs.ltail) (ys.ltail))
	else              inter (xs) (ys.ltail)        
      }
    }
    (coll.unsafeFromOrdList) (inter (xs.toOrdSeq) (ys.toOrdSeq))    
  }
}

// divert(0)
newtype __CollectionDefault__ = {
  def empty:any
}























