// -*- Funnel -*-
/*  FUNNEL LIBRARY (EDISON PORT)
**  
**  Leftist Heap -- support for heaps in Funnel
**  
**  27/04/2001, Yacine Saidji
*/

/**
 * $Author: ysaidji $
 * $Id: LeftistHeap.m4,v 1.2 2001/09/12 13:25:10 ysaidji Exp $
 */

// trick to get the compiler to import some types
def removeMe0 = include[LeftistHeapType]
def removeMe1 = include[__CollectionDefaultType__]
def removeMe2 = include[__CollectionDefault__]

def LeftistHeap[a] (equ: Equ[a], ord: Ord[a]) : LeftistHeapModule[a] = {

  let this = {
  
    // inherence
    with CollXDefault[a](this)
    with CollDefault[a](this)
    with OrdCollXDefault[a](this)
    with OrdCollDefault[a](this)

    // private static methods
    def node (x:a) (first: LeftistHeap[a]) (second: LeftistHeap[a]) : LeftistHeap[a] = {
      if (first.rank >= second.rank)
        Tree(second.rank+1, first, x, second)
      else
        Tree(first.rank+1, second, x, first)
    }

    // -------- static methods of CollXModule --------
    // Comparison
    def equivalent (x) (y) = equ.equ (x) (y) 
    // Constructors
    def empty = Nil
    def single (x) = Tree(1,Nil,x,Nil)
    def unionSeq (s) =  unionSeqUsingReduce (s)
    def fromSeq (xs) = fromSeqUsingUnionSeq (List) (xs)


    // -------- static methods of OrdCollXModule -------- 
    // Comparison
    def LT = 0
    def GT = 1
    def EQ = 2
    def compare (x) (y) = if (ord.lt (x) (y)) LT else if (ord.lt (y) (x)) GT else EQ
    // Constructor
    def unsafeFromOrdSeq (xs) = unsafeFromOrdSeqUsingUnsafeInsertMin (xs)

    // ----------- Empty and Tree ----------

    def Nil = {
      let this : LeftistHeap[a] = {

        // private methods
        def rank = 0
        def lChild = empty
        def rChild = empty

        def del (x:a) : (boolean, LeftistHeap[a]) = (false, this)

        def sz (i: int) : int = i

        def lookupM (x:a) : (boolean,a) = (false, x)

        def maxView : (boolean, LeftistHeap[a], a) = (false, this, null) // ToDo : use Maybe

        // ----------- CollX methods -----------
        // Constructors
        def insert (x:a) = Tree(1, empty, x, empty)
        def insertSeq (xs:__Seq__[a]) = insertSeqUsingUnion (xs) (this)
        def union (xs:__CollX__[a]) = xs

        // Deletions
        def delete (x:a) = this
        def deleteAll (x:a) = this
        def deleteSeq (xs:__Seq__[a]) = this

        // Observers
        def isEmpty = true
        def size = 0
        def member (x:a) = false
        def count (x:a) = 0

        // ----------- Coll methods -----------
        // Observers
        def lookup (x:a) = System.error "lookup: empty collection"
        def lookupAll (x:a) = __Seq__.empty[a]
        def lookupWithDefault (x:a) (d:a) = d
        def toSeq = __Seq__.empty[a]

        // Filters and partitions
        def filter (p) = this
        def partition (p) = (this,this)
	def exist (p) = false

        // Folds
        def fold[b] (f) (e:b) = e
        def fold1 (f) = System.error "fold1: empty collection"
	def iterate (f) = null

        // ----------- OrdCollX methods -----------
        // Constructors
        def unsafeInsertMin (x:a) = Tree(1,this,x,this)
        def unsafeInsertMax (x:a) = unsafeInsertMaxUsingUnsafeAppend (this) (x)
        def unsafeAppend (h:__OrdCollX__[a]) = h

        // Deletions
        def deleteMin = this
        def deleteMax = this

        // Filters and partitions
        def filterLT (x) = this
        def filterLE (x) = this
        def filterGT (x) = this
        def filterGE (x) = this

        def partitionLT_GE (x) = (this,this)
        def partitionLE_GT (x) = (this,this)
        def partitionLT_GT (x) = (this,this)

        // ----------- OrdCollX methods -----------
        // Destructors
        def minElem = System.error "minElem: empty collection"
        def maxElem = System.error "maxElem: empty collection"

        // Observers
        def toOrdSeq = __Seq__.empty[a]

        // Folds
        def foldr[b] (f) (e:b) = e
        def foldl[b] (f) (e:b) = e
        def foldr1 (f) = System.error "foldr1: empty collection"
        def foldl1 (f) = System.error "foldl1: empty collection"
	def iterater (f) = null
	def iteratel (f) = null

        // return the record
        //this
      }
      this

    }

    def Tree (r, lsubtree, y:a, rsubtree) = {
      let this : LeftistHeap[a] =  {

        // private methods
        def rank = r
        def lChild = lsubtree
        def rChild = rsubtree

        def del (x:a) : (boolean, LeftistHeap[a]) = {
	  val d = compare (x) (y)
	  // check with min element
	  if (d == LT) (false,this)
	  else if (d == EQ) (true, lsubtree.union (rsubtree))
	  // check with the childrens
	  else ( val (b,h) = rsubtree.del (x);
	         if (b == true) (true, node (y) (lsubtree) (h))
	         else ( val (b',h') = lsubtree.del (x);
		        if (b' == true) (true, node (y) (h') (rsubtree))
		        else (false,this)
		       )
	        )
        }

        def sz (i:int) : int = lsubtree.sz (rsubtree.sz (i+1))

        def lookupM (x:a) : (boolean,a) = {
	  val d = compare (x) (y)
	  if (d == LT) (false,x)
	  else if (d == EQ) (true,y)
	  else ( val (b,z) = lsubtree.lookupM (x);
	         if (b == true) (b,z)
	         else rsubtree.lookupM (x) )
        }

        def maxView : (boolean, LeftistHeap[a], a) = {
	  if (lsubtree.isEmpty) (true,empty,y)
	  else if (rsubtree.isEmpty) ( val (b,h,x) = lsubtree.maxView; (true,Tree(1,h,y,empty),x))
	  else (
	    val (_,a',y') = lsubtree.maxView;
	    val (_,b',z) = rsubtree.maxView;
	    val d = compare (y') (z);
	    if (d == LT) (true,node (y) (lsubtree) (b'),z) else (true,node (y) (a') (rsubtree),y') )
        }

        // ----------- CollX methods -----------
        // Constructors
        def insert (x:a) = union( Tree(1, empty, x, empty)  )

        def insertSeq (xs:__Seq__[a]) = insertSeqUsingUnion (xs) (this)

        def union (xs:__CollX__[a]) : LeftistHeap[a]= if (xs.isEmpty) Tree(r, lsubtree, y, rsubtree)
		         else if ( ord.leq (xs.minElem) (y) )
			        node (xs.minElem) (xs.lChild) (union (xs.rChild))
			      else
			        node (y) (lsubtree) ((rsubtree.union) (xs))

        // Deletions
        def delete (x:a) = {
	  val (b,h) = del (x)
	  if (b == true) h else this
        }

        def deleteAll (x:a) = {
	  val d = compare (x) (y)
	  if (d == LT) this
	  else if (d == EQ) (lsubtree.deleteAll (x)).union (rsubtree.deleteAll (x))
	  else node (y) (lsubtree.deleteAll (x)) (rsubtree.deleteAll (x))
        }

        def deleteSeq (xs:__Seq__[a]) = deleteSeqUsingDelete (xs) (this)

        // Observers
        def isEmpty = false

        def size = sz 0

        def member (x:a) = {
	  val d = compare (x) (y)
	  if (d == LT) false
	  else if (d == EQ) true
	  else ((lsubtree.member (x)) || (rsubtree.member (x)))
        }

        def count (x:a) = {
	  val d = compare (x) (y)
	  if (d == LT) 0
	  else if (d == EQ) 1 + ((lsubtree.count (x)) + (rsubtree.count (x)))
	  else ((lsubtree.count (x)) + (rsubtree.count (x)))
        }

        // ----------- Coll methods -----------
        // Observers
        def lookup (x:a) = {
	  val (b,z) = lookupM (x)
	  if (b == true) z
	  else System.error "lookup failed"
        }

        def lookupAll (x:a) = {
	  def look (h:LeftistHeap[a]) (ys:__Seq__[a]) : __Seq__[a] = {
	    if (h.isEmpty) ys
	    else {
	      val y = h.minElem;
	      val d = compare (x) (y);
	      if (d == LT) ys
	      else if (d == EQ) (__Seq__.cons[a]) (y) (look (h.rChild) (look (h.lChild) (ys)))
	      else (look (h.rChild) (look (h.lChild) (ys)))
            }
	  }
	  look (this) (__Seq__.empty[a])
        }

        def lookupWithDefault (x:a) (d:a) = {
	  val (b,z) = lookupM (x)
	  if (b == true) z
	  else d
        }

        def toSeq = {
	  def tol (h:LeftistHeap[a]) (rest:__Seq__[a]) :__Seq__[a]  = {
	    if (h.isEmpty) rest
	    else (__Seq__.cons[a]) (h.minElem) (tol (h.rChild) (tol (h.lChild) rest))
	  }
	  tol (this) (__Seq__.empty[a])
        }

        // Filters and partitions
        def filter (p:a -> boolean) = {
	  if (p (y)) node (y) (lsubtree.filter (p)) (rsubtree.filter (p))
	  else (lsubtree.filter (p)).union (rsubtree.filter (p))
        }

        def partition (p:a -> boolean) = {
	  val (a',a'') = lsubtree.partition (p)
	  val (b',b'') = rsubtree.partition (p)
	  if (p (y)) (node (y) (a') (b'), (a'').union (b''))
	  else ((a').union (b'), node (y) (a'') (b''))
        }

	def exist (p:a -> boolean) = (p(y) || (lsubtree.exist) (p)) || (rsubtree.exist) (p)

        // Folds
        def fold[b] (f: a -> b -> b) (e:b) = f y (rsubtree.fold (f) (lsubtree.fold (f) (e))) 
        def fold1 (f: a -> a -> a) = rsubtree.fold (f) (lsubtree.fold (f) (y))
	def iterate (f: a -> ()) = {
	  f (y)
	  rsubtree.iterate (f)
	  lsubtree.iterate (f)
	}

        // ----------- OrdCollX methods -----------
        // Constructor
        def unsafeInsertMin (x:a) = Tree(1,this,x,empty)
        def unsafeInsertMax (x:a) = unsafeInsertMaxUsingUnsafeAppend (this) (x)
        def unsafeAppend (h:__OrdCollX__[a]) = node (y) (lsubtree) (rsubtree.unsafeAppend (h))

        // Deletions
        def deleteMin = lsubtree.union (rsubtree)
        def deleteMax = (val (b,h,x) = maxView; if (b) h else empty)

        // Filters and partition
        def filterLT (x:a) = if (LT == compare (y) (x))
			     node y (lsubtree.filterLT (x)) (rsubtree.filterLT (x))
			   else
			     empty

        def filterLE (x:a) = (val d = compare (y) (x);
			    if ((LT == d) || (EQ == d)) 
			      node y (lsubtree.filterLE (x)) (rsubtree.filterLE (x))
			    else
			      empty)

        // ToDo : optimize the following
        def filterGT (x:a) = filter (y:a | (GT == compare (y) (x)))
        def filterGE (x:a) = filter (y:a | (val d = compare (y) (x); ((d == GT) || (d == EQ))))
        def partitionLT_GE (x:a) = partition (y:a | (LT == compare (y) (x)))
        def partitionLE_GT (x:a) = partition (y:a | (val d = compare (y) (x); ((d == LT) || (d == EQ))))
        def partitionLT_GT (x:a) = (filterLT x, filterGT x)


        // ----------- OrdCollX methods -----------
        // Destructors
        def minElem = y
        def maxElem = (val (b,h,x) = maxView; x)

        // Observers
        def toOrdSeq = toOrdSeqUsingFoldr (List) (this)

        // Folds
        def foldr[b] (f: a -> b -> b) (e:b) = f y ((lsubtree.union (rsubtree)).foldr (f) (e))
        def foldl[b] (f: b -> a -> b) (e:b) = (lsubtree.union (rsubtree)).foldl (f) (f (e) (y))
        def foldr1 (f: a -> a -> a) = if (lsubtree.isEmpty) y 
		         else f y ((lsubtree.union (rsubtree)).foldr1 (f))
        def foldl1 (f: a -> a -> a) = (lsubtree.union (rsubtree)).foldl (f) (y)

	def iterater (f: a -> ()) = foldr[()] (x | dummy | f(x)) (null)
	def iteratel (f: a -> ()) = foldl[()] (dummy | x | f(x)) (null)

        // return the record
        //this
      }
      this
    }

  }
  this

}




