// -*- Funnel -*-
/*  FUNNEL LIBRARY (EDISON PORT)
**  
**  Leftist Heap -- support for heaps in Funnel
**  
**  27/04/2001, Yacine Saidji
*/

/**
 * $Author: ysaidji $
 * $Id: LeftistHeapType.m4,v 1.2 2001/09/12 13:25:10 ysaidji Exp $
 */

// ToDo : newtype LeftistHeap[a<:EqOrd[a]] <: OrdColl[a] = {
newtype LeftistHeap[a] = {

  // ----------- CollX methods -----------
  // Constructors
  def insert (x:a): LeftistHeap[a]
  def insertSeq (xs:__Seq__[a]): LeftistHeap[a]
  def union (xs:__CollX__[a]): LeftistHeap[a]

  // Deletions
  def delete (x:a): LeftistHeap[a]
  def deleteAll (x:a): LeftistHeap[a]
  def deleteSeq (xs:__Seq__[a]): LeftistHeap[a]

  // Observers
  def isEmpty: boolean
  def size: int
  def member (x:a): boolean
  def count (x:a): int

  // ----------- Coll methods -----------
  // Observers
  def lookup (x:a): a
  //def lookupM
  def lookupAll (x:a): __Seq__[a]
  def lookupWithDefault (x:a) (d:a): a
  def toSeq: __Seq__[a]

  // Filters and partitions
  def filter (p: a -> boolean): LeftistHeap[a]
  def partition (p: a -> boolean): (LeftistHeap[a], LeftistHeap[a])
  def exist(p:a -> boolean): boolean

  // Folds
  def fold[b](f: a -> b -> b)(e: b): b
  def fold1(f: a -> a -> a): a
  def iterate(f:a ->()) : ()

  // ----------- OrdCollX methods -----------
  // Constructor
  def unsafeInsertMin (x:a): LeftistHeap[a]
  def unsafeInsertMax (x:a): LeftistHeap[a]
  def unsafeAppend (h:__OrdCollX__[a]): LeftistHeap[a]

  // Deletions
  def deleteMin: LeftistHeap[a]
  def deleteMax: LeftistHeap[a]

  // Filters and partitions
  def filterLT (x:a): LeftistHeap[a]
  def filterLE (x:a): LeftistHeap[a]
  def filterGT (x:a): LeftistHeap[a]
  def filterGE (x:a): LeftistHeap[a]
  def partitionLT_GE (x:a): (LeftistHeap[a],LeftistHeap[a])
  def partitionLE_GT (x:a): (LeftistHeap[a],LeftistHeap[a])
  def partitionLT_GT (x:a): (LeftistHeap[a],LeftistHeap[a])


  // ----------- OrdColl methods -----------
  // Destructors
  //def minView
  def minElem: a
  //def maxView
  def maxElem: a

  // Observers
  def toOrdSeq: __Seq__[a]

  // Folds
  def foldr[b](f: a -> b -> b)(e: b): b
  def foldl[b](f: b -> a -> b)(e: b): b
  def foldr1(f: a -> a -> a): a
  def foldl1(f: a -> a -> a): a
  def iterater(f:a ->()) : ()
  def iteratel(f:a ->()) : ()


  // ----------- Private methods -----------
  def rank: int
  def lChild: LeftistHeap[a]
  def rChild: LeftistHeap[a]
  def del (x:a) : (boolean, LeftistHeap[a])
  def sz (i: int) : int
  def lookupM (x:a) : (boolean,a)
  def maxView : (boolean, LeftistHeap[a], a)

}


type LeftistHeapModule[a] = {

  // -------- Constructors -------- 
  def Nil : LeftistHeap[a]
  def Tree (r: int, lsubtree: LeftistHeap[a], y: a, rsubtree: LeftistHeap[a]) : LeftistHeap[a]

  // -------- static methods of CollXModule --------
  // Comparison
  def equivalent (x:a) (y:a) : boolean
  // Constructors
  def empty: LeftistHeap[a]
  def single (x:a): LeftistHeap[a]
  def unionSeq (s:__Seq__[__CollX__[a]]): LeftistHeap[a]
  def fromSeq (xs:__Seq__[a]): LeftistHeap[a]
  

  // -------- static methods of OrdCollXModule -------- 
  // Comparison
  def LT: int
  def GT: int
  def EQ: int
  def compare (x:a) (y:a) : int
  // Constructor
  def unsafeFromOrdSeq (xs:__Seq__[a]): LeftistHeap[a]
  
  // -------- Private static methods --------
  def node (x:a) (first:LeftistHeap[a]) (second:LeftistHeap[a]) : LeftistHeap[a]
}

newtype Equ[a] = {
  def equ (x:a) (y:a): boolean
}

newtype Ord[a] = {
  def lt (x:a) (y:a): boolean
  def leq (x:a) (y:a): boolean
}


newtype LeftistHeapType = {
  def dummy:any
}













