// ------------- the Coll test -------------------


// - change the following to test another collection implementation

def removeMe = include[LeftistHeapType]
//def removeMe = include[SplayHeapType]

// some relations
val IntEq : Equ[int] = {
  def equ (x:int) (y:int) : boolean = (x == y)
}
val IntOrd : Ord[int] = {
  def lt (x:int) (y:int): boolean = (x < y)
  def leq (x:int) (y:int): boolean = (x <= y)
}

/////////////////////////////////////////////////////////////////////

/*
type Coll[a] = SplayHeap[a]
let IntColl : SplayHeapModule[int] = SplayHeap[int](IntEq, IntOrd)
*/

/////////////////////////////////////////////////////////////////////


type Coll[a] = LeftistHeap[a]
let IntColl : LeftistHeapModule[int] = LeftistHeap[int](IntEq, IntOrd)


/////////////////////////////////////////////////////////////////////

// ----- the generators ---------
let gCollInt : Gen[Coll[int]] = {
  def gen = IntColl.fromSeq ( Gen.gListInt.gen )
}

let gListCollInt : Gen[List[Coll[int]]] = Gen.gList[Coll[int]] (gCollInt)


// ------- some helpers ---------

def lmerge (xs:List[int]) (ys:List[int]) : List[int] = {
  if (xs.isEmpty) ys
  else if (ys.isEmpty) xs
  else {
    val x = xs.lhead
    val y = ys.lhead
    if ( x <= y ) List.cons[int] (x) (lmerge (xs.ltail) (ys))
    else List.cons[int] (y) (lmerge (xs) (ys.ltail))
  }
}

def insert (x:int) (xs:List[int]) : List[int] = {
  if (xs.isEmpty) List.single (x)
  else if ( x <= xs.lhead ) List.cons (x) (xs)
  else List.cons (xs.lhead) ( insert (x) (xs.ltail) )
}

def delete (x:int) (xs:List[int]) : List[int] = {
  if (xs.isEmpty) xs
  else if ( x == xs.lhead ) xs.ltail
  else List.cons (xs.lhead) ( delete (x) (xs.ltail) )
}

def quicksort[a] (comp:a->a->boolean) (lst:List[a]) : List[a] = 
  if (lst.isEmpty || lst.ltail.isEmpty) lst
  else ( val (i1,i2) = lst.ltail.partition( comp (lst.lhead) ); 
	 (List.append[a]) ((quicksort[a]) (comp) (i1)) (List.cons[a] (lst.lhead) ((quicksort[a]) (comp) (i2))) )
def ppt (a:int) (b:int): boolean = (b < a)
def sort (xs:List[int]) : List[int] = quicksort[int] (ppt) (xs) 

let QTColl = {

  with Gen
  with Result

  // List comparison
  def equList (xs:List[int]) (ys:List[int]) = (xs.zip (ys)).foldl ( b | (x,y) | (x==y) && b  ) (true)

  // Collections comparison
  def equ (xs:Coll[int]) (ys:Coll[int]) = {
    equList (xs.toOrdSeq) (ys.toOrdSeq)
  }

  ////////////////////////////////////////////////////////////////////
  ////    CollX operations
  ////////////////////////////////////////////////////////////////////

  // ------ single ----------
  def prop_single (x:int) : boolean = {
    equList ((IntColl.single (x)).toOrdSeq) (List.single (x))
  }
  def prop_single_test : Result = QuickCheck.qc0b[int] (prop_single) (gInt)


  // ------ fromSeq ----------
  def prop_fromSeq (xs:List[int]) : boolean = {
    equ (IntColl.fromSeq (xs)) (xs.foldr (x|c:Coll[int]|c.insert (x)) (IntColl.empty))
  }
  def prop_fromSeq_test : Result = QuickCheck.qc0b[List[int]] (prop_fromSeq) (gListInt)

  // ------ insert ----------
  def prop_insert (x:int) (xs:Coll[int]) : boolean = {
    equList ((xs.insert (x)).toOrdSeq) (insert (x) (xs.toOrdSeq))
  }
  def prop_insert_test : Result = QuickCheck.qc1b[int,Coll[int]] (prop_insert) (gInt) (gCollInt)


  // ------ union ----------
  def prop_union (xs:Coll[int]) (ys:Coll[int]) : boolean = {
    equList ((xs.union (ys)).toOrdSeq) (lmerge (xs.toOrdSeq) (ys.toOrdSeq))
  }
  def prop_union_test : Result = QuickCheck.qc1b[Coll[int],Coll[int]] (prop_union) (gCollInt) (gCollInt)

  // ------ unionSeq ----------
  def prop_unionSeq (xss:List[Coll[int]]) : boolean = {
    equ (IntColl.unionSeq (xss)) (xss.foldr (x|c:Coll[int]|c.union (x)) (IntColl.empty))
  }
  def prop_unionSeq_test : Result = QuickCheck.qc0b[List[Coll[int]]] (prop_unionSeq) (gListCollInt)

  // ------ delete ----------
  def prop_delete (x:int) (xs:Coll[int]) : boolean = {
    equList ((xs.delete (x)).toOrdSeq) (delete (x) (xs.toOrdSeq))
  }
  def prop_delete_test : Result = QuickCheck.qc1b[int,Coll[int]] (prop_delete) (gInt) (gCollInt)
  

  // ------ deleteAll ----------
  def prop_deleteAll (x:int) (xs:Coll[int]) : boolean = {
    equList ((xs.deleteAll (x)).toOrdSeq) ((xs.toOrdSeq).filter (y|(x !=y)))
  }
  def prop_deleteAll_test : Result = QuickCheck.qc1b[int,Coll[int]] (prop_deleteAll) (gInt) (gCollInt)

  // ------ deleteSeq ----------
  def prop_deleteSeq (xs:List[int]) (ys:Coll[int]) : boolean = {
    equ (ys.deleteSeq (xs)) (xs.foldr (x|c:Coll[int]|c.delete (x)) (ys))
  }
  def prop_deleteSeq_test : Result = QuickCheck.qc1b[List[int],Coll[int]] (prop_deleteSeq) (gListInt) (gCollInt)

  // ------ isEmpty_size ----------
  def prop_isEmpty_size (xs:Coll[int]) : boolean = {
    ((xs.isEmpty) == (0 == xs.size))
      && ((xs.size) == ((xs.toOrdSeq).size))
  }
  def prop_isEmpty_size_test : Result = QuickCheck.qc0b[Coll[int]] (prop_isEmpty_size) (gCollInt)

  // ------ member_count ----------
  def prop_member_count (xs:Coll[int]) (x:int) : boolean = {
    val c = xs.count (x)
    ((xs.member (x)) == (c > 0))
      && ( c == ((xs.toOrdSeq).filter (y|(x==y))).size )
  }
  def prop_member_count_test : Result = QuickCheck.qc1b[Coll[int],int] (prop_member_count) (gCollInt) (gInt)

  ////////////////////////////////////////////////////////////////////
  ////    Coll operations
  ////////////////////////////////////////////////////////////////////

  // ------ toSeq ----------
  def prop_toSeq (xs:Coll[int]) : boolean = {
    equList (sort (xs.toSeq)) (xs.toOrdSeq)
  }
  def prop_toSeq_test : Result = QuickCheck.qc0b[Coll[int]] (prop_toSeq) (gCollInt)

  // ------ lookup ----------
  def prop_lookup (xs:Coll[int]) (x:int) : boolean = {
    if ( xs.member (x) ) {
      ((xs.lookup (x)) == (x))
        && ((xs.lookupWithDefault (x) (999)) == (x))
        && equList (xs.lookupAll (x)) (List.copy[int] (xs.count (x)) (x))
    } else {
      ((xs.lookupWithDefault (x) (999)) ==  (999))
        && equList (xs.lookupAll (x)) (List.empty[int])
    }
  }
  def prop_lookup_test : Result = QuickCheck.qc1b[Coll[int],int] (prop_lookup) (gCollInt) (gInt)

  // ------ fold ----------
  def prop_fold (xs:Coll[int]) : boolean = {
    def plus (x:int) (y:int) : int = x + y
    equList (sort (xs.fold (List.cons[int]) (List.empty[int]))) (xs.toOrdSeq)
      && ( xs.isEmpty || ((xs.fold1 (plus)) == ((xs.toOrdSeq).foldl (plus) (0))) )
  }
  def prop_fold_test : Result = QuickCheck.qc0b[Coll[int]] (prop_fold) (gCollInt)

  // ------ filter_partition ----------
  def prop_filter_partition (xs:Coll[int]) : boolean = {
    def p (x:int) : boolean = ( (x % 3) == 2 )
    val (a,b) = xs.partition (p)
    equList ((xs.filter (p)).toOrdSeq) ((xs.toOrdSeq).filter (p))
      && equ (a) (xs.filter (p))
      && equ (b) (xs.filter (x|(!p (x))))
  }
  def prop_filter_partition_test : Result = QuickCheck.qc0b[Coll[int]] (prop_filter_partition) (gCollInt)

  ////////////////////////////////////////////////////////////////////
  ////    OrdCollX operations
  ////////////////////////////////////////////////////////////////////

  // ------ deleteMin_Max ----------
  def prop_deleteMin_Max (xs:Coll[int]) : boolean = {
    equList ((xs.deleteMin).toOrdSeq) ((xs.toOrdSeq).ltail)
      && equList ((xs.deleteMax).toOrdSeq) ((xs.toOrdSeq).rtail)
  }
  def prop_deleteMin_Max_test : Result = QuickCheck.qc0b[Coll[int]] (prop_deleteMin_Max) (gCollInt)

  // ------ unsafeInsertMin_Max ----------
  def prop_unsafeInsertMin_Max (i:int) (xs:Coll[int]) : boolean = {
    if (xs.isEmpty) {
      equ (xs.unsafeInsertMin (0)) (IntColl.single (0))
        && equ (xs.unsafeInsertMax (0)) (IntColl.single (0))
    } else {
      def odd (x:int) : boolean = ((x%2) == 1)
      val lo = (xs.minElem) - ( if (odd (i)) 1 else 0 )
      val hi = (xs.maxElem) + ( if (odd (i)) 1 else 0 )
      equ (xs.unsafeInsertMin (lo)) (xs.insert (lo))
        && equ (xs.unsafeInsertMax (hi)) (xs.insert (hi))
    }
  }
  def prop_unsafeInsertMin_Max_test : Result = QuickCheck.qc1b[int,Coll[int]] (prop_unsafeInsertMin_Max) (gInt) (gCollInt)

  // ------ unsafeFromOrdSeq ----------
  def prop_unsafeFromOrdSeq (xs:List[int]) : boolean = {
    val xs' = sort (xs)
    equList ((IntColl.unsafeFromOrdSeq (xs')).toOrdSeq) (xs')
  }
  def prop_unsafeFromOrdSeq_test : Result = QuickCheck.qc0b[List[int]] (prop_unsafeFromOrdSeq) (gListInt)


  // ------ unsafeAppend ----------
  def prop_unsafeAppend (i:int) (xs:Coll[int]) (ys:Coll[int]) : boolean = {
    def odd (x:int) : boolean = ((x%2) == 1)
    if ( (xs.isEmpty) || (ys.isEmpty) ) {
      equ (xs.unsafeAppend (ys)) (xs.union (ys))
    } else {
      val delta = (xs.maxElem) - (ys.minElem) + ( if (odd (i)) 1 else 0 )
      val ys' = IntColl.fromSeq ((ys.toSeq).map (x:int|x+delta))
      equ (xs.unsafeAppend (ys')) (xs.union (ys'))      
    }
  }
  def prop_unsafeAppend_test : Result = QuickCheck.qc2b[int,Coll[int],Coll[int]] (prop_unsafeAppend) (gInt) (gCollInt) (gCollInt)

  // ------ filter ----------
  def prop_filter (x:int) (xs:Coll[int]) : boolean = {
    equList ((xs.filterLT (x)).toOrdSeq) ((xs.toOrdSeq).filter (y|y<x))
      && equList ((xs.filterLE (x)).toOrdSeq) ((xs.toOrdSeq).filter (y|y<=x))
      && equList ((xs.filterGT (x)).toOrdSeq) ((xs.toOrdSeq).filter (y|y>x))
      && equList ((xs.filterGE (x)).toOrdSeq) ((xs.toOrdSeq).filter (y|y>=x))
  }
  def prop_filter_test : Result = QuickCheck.qc1b[int,Coll[int]] (prop_filter) (gInt) (gCollInt)


  // ------ partition ----------
  def prop_partition (x:int) (xs:Coll[int]) : boolean = {
    def equTL (a:Coll[int],b:Coll[int]) (a':Coll[int],b':Coll[int]) : boolean = (equ (a) (a')) && (equ (b) (b')) 
    equTL (xs.partitionLT_GE (x)) ( xs.filterLT (x) , xs.filterGE (x) )
      && equTL (xs.partitionLE_GT (x)) ( xs.filterLE (x) , xs.filterGT (x) )
      && equTL (xs.partitionLT_GT (x)) ( xs.filterLT (x) , xs.filterGT (x) )
  }
  def prop_partition_test : Result = QuickCheck.qc1b[int,Coll[int]] (prop_partition) (gInt) (gCollInt)



  ////////////////////////////////////////////////////////////////////
  ////    OrdColl operations
  ////////////////////////////////////////////////////////////////////

  // ------ minElem_maxElem ----------
  def prop_minElem_maxElem (xs:Coll[int]) : Result = {
    implies (!(xs.isEmpty)) ( |
      ((xs.minElem) == (xs.toOrdSeq).lhead)
        && ((xs.maxElem) == (xs.toOrdSeq).rhead)
    )
  }
  def prop_minElem_maxElem_test : Result = QuickCheck.qc0r[Coll[int]] (prop_minElem_maxElem) (gCollInt)


  // ------ foldr_foldl ----------
  def prop_foldr_foldl (xs:Coll[int]) : boolean = {
    def flip[a,b] (f:a->b->b) : b->a->b = (y:b | x:a | f (x) (y))
    def :: (x:int) (xs:List[int]) = List.cons (x) (xs)
    equList (xs.foldr (::) (List.empty[int])) (xs.toOrdSeq)
      && equList (xs.foldl (flip (::)) (List.empty[int])) ((xs.toOrdSeq).reverse)
  }
  def prop_foldr_foldl_test : Result = QuickCheck.qc0b[Coll[int]] (prop_foldr_foldl) (gCollInt)

  // ------ foldr1_foldl1 ----------
  def prop_foldr1_foldl1 (xs:Coll[int]) : Result = {
    def flip[a,b] (f:a->b->b) : b->a->b = (y:b | x:a | f (x) (y))
    def f (x:int) (y:int) : int = if (y==1333) x else (3*x-7*y)
    implies (!(xs.isEmpty)) ( |
      (xs.foldr1 (f) == xs.foldr (f) (1333))
	&& (xs.foldl1 (flip (f)) == xs.foldl (flip (f)) 1333)
    )
  }
  def prop_foldr1_foldl1_test : Result = QuickCheck.qc0r[Coll[int]] (prop_foldr1_foldl1) (gCollInt)


  // ------------ the test sequence -------------------
  def run = {
    val _ = QuickCheck.quickcheck (|prop_single_test) ("prop_single_test")
    val _ = QuickCheck.quickcheck (|prop_fromSeq_test) ("prop_fromSeq_test")
    val _ = QuickCheck.quickcheck (|prop_insert_test) ("prop_insert_test")
    val _ = QuickCheck.quickcheck (|prop_union_test) ("prop_union_test")
    val _ = QuickCheck.quickcheck (|prop_unionSeq_test) ("prop_unionSeq_test")
    val _ = QuickCheck.quickcheck (|prop_delete_test) ("prop_delete_test")
    val _ = QuickCheck.quickcheck (|prop_deleteAll_test) ("prop_deleteAll_test")
    val _ = QuickCheck.quickcheck (|prop_deleteSeq_test) ("prop_deleteSeq_test")
    val _ = QuickCheck.quickcheck (|prop_isEmpty_size_test) ("prop_isEmpty_size_test")
    val _ = QuickCheck.quickcheck (|prop_member_count_test) ("prop_member_count_test")
    val _ = QuickCheck.quickcheck (|prop_toSeq_test) ("prop_toSeq_test")
    val _ = QuickCheck.quickcheck (|prop_lookup_test) ("prop_lookup_test")
    val _ = QuickCheck.quickcheck (|prop_fold_test) ("prop_fold_test")
    val _ = QuickCheck.quickcheck (|prop_filter_partition_test) ("prop_filter_partition_test")    
    val _ = QuickCheck.quickcheck (|prop_deleteMin_Max_test) ("prop_deleteMin_Max_test")
    val _ = QuickCheck.quickcheck (|prop_unsafeInsertMin_Max_test) ("prop_unsafeInsertMin_Max_test")
    val _ = QuickCheck.quickcheck (|prop_unsafeFromOrdSeq_test) ("prop_unsafeFromOrdSeq_test")
    val _ = QuickCheck.quickcheck (|prop_unsafeAppend_test) ("prop_unsafeAppend_test")
    val _ = QuickCheck.quickcheck (|prop_filter_test) ("prop_filter_test")
    val _ = QuickCheck.quickcheck (|prop_partition_test) ("prop_partition_test")    
    val _ = QuickCheck.quickcheck (|prop_minElem_maxElem_test) ("prop_minElem_maxElem_test")
    val _ = QuickCheck.quickcheck (|prop_foldr_foldl_test) ("prop_foldr_foldl_test")
    val _ = QuickCheck.quickcheck (|prop_foldr1_foldl1_test) ("prop_foldr1_foldl1_test")

  }

}


// -------- the test run -----------------

QTColl.run
