// -*- Funnel -*-
/*  FUNNEL LIBRARY (EDISON PORT)
**  
**  Splay Heap -- support for heaps in Funnel
**  
**  30/06/2001, Yacine Saidji
*/

/**
 * $Author: ysaidji $
 * $Id: SplayHeap.m4,v 1.2 2001/09/12 13:25:09 ysaidji Exp $
 */

// trick to get the compiler to import some types
def removeMe0 = include[SplayHeapType]
def removeMe1 = include[__CollectionDefaultType__]
def removeMe2 = include[__CollectionDefault__]

def SplayHeap[a] (equ: Equ[a], ord: Ord[a]) : SplayHeapModule[a] = {

  let this : SplayHeapModule[a] = {
  
    // inherence
    with CollXDefault[a](this)
    with CollDefault[a](this)
    with OrdCollXDefault[a](this)
    with OrdCollDefault[a](this)

    // private static methods
    def node (a:SplayHeap[a], x:a, b:SplayHeap[a]) : SplayHeap[a]= Tree(a,x,b)

    // -------- static methods of CollXModule --------
    // Comparison
    def equivalent (x) (y) = equ.equ (x) (y) 

    // Constructors
    def empty = Nil
    def single (x) = Tree(Nil,x,Nil)

    def unionSeq (s) = unionSeqUsingReduce (s)
    def fromSeq (xs) = fromSeqUsingFoldr (xs)


    // -------- static methods of OrdCollXModule -------- 
    // Comparison
    def LT = 0
    def GT = 1
    def EQ = 2
    def compare (x) (y) = if (ord.lt (x) (y)) LT else if (ord.lt (y) (x)) GT else EQ

    // Constructors
    def unsafeFromOrdSeq (xs) = unsafeFromOrdSeqUsingUnsafeInsertMin (xs)

    // ----------- Empty and Tree ----------
    def Nil = {
      let this : SplayHeap[a] = {

        // private methods
	def sz (n:int) : int = n
	def cnt (n:int) (x:a) : int = n
	def look (y:a) (rest:List[a]) : List[a] = rest
	def tos (rest:List[a]) : List[a] = rest

	// visitor suppport
	def match[b] ( v:SplayHeapVisitor[a,b]) = v.Nil

	// ----------- CollX methods -----------
        // Constructors
	def insert (x:a) = Tree(this,x,this)
	def insertSeq (xs:__Seq__[a]) = insertSeqUsingFoldr (xs) (this)
	def union (xs:__CollX__[a]) = xs

	// Deletions
	def delete (x:a) = this
	def deleteAll (x:a) = this
	def deleteSeq (xs:__Seq__[a]) = this

	// Observers
	def isEmpty = true
	def size = 0
	def member (x:a) = false
	def count (x:a) = 0

        // ----------- Coll methods -----------
        // Observers
        def lookup (x:a) = System.error "lookup: empty collection"
        def lookupM (x:a) = Maybe.Nothing
        def lookupAll (x:a) = __Seq__.empty[a]
        def lookupWithDefault (x:a) (d:a) = d
        def toSeq = __Seq__.empty[a]

        // Filters and partitions
        def filter (p) = this
        def partition (p) = (this,this)
	def exist (p) = false

        // Folds
        def fold[b] (f) (e:b) = e
        def fold1 (f) = System.error "fold1: empty collection"
	def iterate (f) = null

        // ----------- OrdCollX methods -----------
        // Constructors
        def unsafeInsertMin (x:a) = Tree(this,x,this)
        def unsafeInsertMax (x:a) = Tree(this,x,this)
        def unsafeAppend (h:__OrdCollX__[a]) = h
        // Deletions
        def deleteMin = this
        def deleteMax = this

        // Filters and partitions
        def filterLT (x) = this
        def filterLE (x) = this
        def filterGT (x) = this
        def filterGE (x) = this

        def partitionLT_GE (x) = (this,this)
        def partitionLE_GT (x) = (this,this)
        def partitionLT_GT (x) = (this,this)

        // ----------- OrdCollX methods -----------
        // Destructors
	def minView = Maybe.Nothing
        def minElem = System.error "minElem: empty collection"
	def maxView = Maybe.Nothing
        def maxElem = System.error "maxElem: empty collection"

        // Observers
        def toOrdSeq = __Seq__.empty[a]

        // Folds
        def foldr[b] (f) (e:b) = e
        def foldl[b] (f) (e:b) = e
        def foldr1 (f) = System.error "foldr1: empty collection"
        def foldl1 (f) = System.error "foldl1: empty collection"
	def iterater (f) = null
	def iteratel (f) = null

	
      }
      this

    }

    def Tree (a:SplayHeap[a], x:a, b:SplayHeap[a]) = {
      let this : SplayHeap[a] =  {

        // private methods
	def sz (n:int) : int = (b.sz) ((a.sz) (1+n) )

	def cnt (n:int) (y:a) : int = {
	  val d = compare (y) (x)
	  if (d == LT) (a.cnt) (n) (y)
	  else if (d == GT) (b.cnt) (n) (y)
	  else (b.cnt) ((a.cnt) (1+n) (y)) (y)
	}

	def look (y:a) (rest:List[a]) : List[a] = {
	  val d = compare (y) (x)
	  if (d == LT) (a.look) (y) (rest)
	  else if (d == GT) (b.look) (y) (rest)
	  else (a.look) (y) (List.cons (x) ((b.look) (y) (rest)))
	}

	def tos (rest:List[a]) : List[a] = {
	  List.cons (x) ((a.tos) ((b.tos) rest))
	}

	// visitor suppport
	def match[b] ( v:SplayHeapVisitor[a,b]) = v.Tree(a,x,b)

        // ----------- CollX methods -----------
        // Constructors
	def insert (x:a) = {
	  val (a,b) = partitionLE_GT (x)
	  Tree (a,x,b)
	}

	def insertSeq (xs:__Seq__[a]) = insertSeqUsingFoldr (xs) (this)

	def union (ys:__CollX__[a]) : SplayHeap[a] = {
	  val (c,d) = (ys.partitionLE_GT) (x)
	  Tree((c.union) (a), x, (d.union) (b))
	}

	// Deletions
	def delete (x:a) = {
	  val (a,b) = partitionLE_GT (x)
	  (a.maxView).match[SplayHeap[a]] {
	    def Nothing = b
	    def Just(a',y) = {
	      if ( GT == compare (x) (y) ) Tree(a',y,b)
	      else (a'.unsafeAppend) (b)
	    }
	  }
	}

	def deleteAll (x:a) = {
	  val (a,b) = partitionLT_GT (x)
	  (a.unsafeAppend) (b)
      	}

	def deleteSeq (xs:__Seq__[a]) = deleteSeqUsingDelete (xs) (this)

	// Observers
	def isEmpty = false

	def size = sz (0)

	def member (y:a) = {
	  val d = compare (y) (x)
	  if (d == LT) (a.member) (y)
	  else (d == EQ) || (b.member) (y)
	}

	def count (x:a) = cnt (0) (x)

        // ----------- Coll methods -----------
        // Observers
	def lookup (y:a) = {
	  val d = compare (y) (x)
	  if (d == LT) (a.lookup) (y)
	  else if (d == GT) (b.lookup) (y)
	  else x
	}

	def lookupM (y:a) = {
	  val d = compare (y) (x)
	  if (d == LT) (a.lookupM) (y)
	  else if (d == GT) (b.lookupM) (y)
	  else Maybe.Just(x)
	}

	def lookupAll (x:a) = look (x) (List.empty)

	
	def lookupWithDefault (y:a) (d:a) = {
	  val d' = compare (y) (x)
	  if (d' == LT) (a.lookupWithDefault) (y) (d)
	  else if (d' == GT) (b.lookupWithDefault) (y) (d)
	  else x
	}

	def toSeq = tos (List.empty)

        // Filters and partitions
        def filter (p:a -> boolean) = {
	  if (p (x)) Tree((a.filter) (p), x, (b.filter) (p))
	  else ((a.filter) (p)).unsafeAppend ((b.filter) (p))
	}

        def partition (p:a -> boolean) = {
	  val (a0,a1) = (a.partition) (p)
	  val (b0,b1) = (b.partition) (p)
	  if (p (x)) ( Tree(a0,x,b0) , (a1.unsafeAppend) (b1) )
	  else ( (a0.unsafeAppend) (b0) , Tree(a1,x,b1) )
	}

	def exist (p) = (p(x) || (a.exist) (p)) || (b.exist) (p)

        // Folds
	def fold[b] (f: a -> b -> b) (e:b) = f (x) ((a.fold) (f) ((b.fold) (f) (e)))
	def fold1 (f: a -> a -> a) = (a.fold) (f) (b.fold (f) (x))
	def iterate (f) = {
	  f (x);
	  (a.iterate) (f)
	  (b.iterate) (f)
	}

        // ----------- OrdCollX methods -----------
        // Constructor
	def unsafeInsertMin (x:a) = Tree(Nil,x,this)
	def unsafeInsertMax (x:a) = Tree(this,x,Nil)
	def unsafeAppend (b:__OrdCollX__[a]) = {
	  (this.maxView).match[SplayHeap[a]] {
	    def Nothing = b
	    def Just(a,x) = Tree(a,x,b)
	  }
	}

        // Deletions
        def deleteMin = {
	  (a.match)[SplayHeap[a]] {
	    def Nil = b
	    def Tree(a',x',b') = (a'.match)[SplayHeap[a]] {
	      def Nil = node(b',x,b)
	      def Tree(a'',x'',b'') = node(a'.deleteMin, x', node(b',x,b))
	    }
	  }
	}

        def deleteMax = {
	  (b.match)[SplayHeap[a]] {
	    def Nil = a
	    def Tree(b,y,c') = (c'.match)[SplayHeap[a]] {
	      def Nil = node(a,x,b)
	      def Tree(c,z,d) = node(node(a,x,b), y, c'.deleteMax)
	    }
	  }
	}

        // Filters and partition
	def filterLT (k:a) = {
	  val d = compare (x) (k)
	  if ((GT == d) || (EQ == d)) (a.filterLT) (k)
	  else {
	    (b.match)[SplayHeap[a]] {
	      def Nil = this
	      def Tree(ba,y,bb) = {
		val d = compare (y) (k)
		if ((GT == d) || (EQ == d)) node(a, x, (ba.filterLT) (k))
		else node (node(a,x,ba), y, (bb.filterLT) (k))
	      }
	    }
	  }
	}

	def filterLE (k:a) = {
	  val d = compare (x) (k)
	  if ((GT == d)) (a.filterLE) (k)
	  else {
	    (b.match)[SplayHeap[a]] {
	      def Nil = this
	      def Tree(ba,y,bb) = {
		val d = compare (y) (k)
		if ((GT == d)) node(a, x, (ba.filterLE) (k))
		else node (node(a,x,ba), y, (bb.filterLE) (k))
	      }
	    }
	  }
	}


	def filterGT (k:a) = {
	  val d = compare (x) (k)
	  if ((LT == d) || (EQ == d)) (b.filterGT) (k)
	  else {
	    (a.match)[SplayHeap[a]] {
	      def Nil = this
	      def Tree(aa,y,ab) = {
		val d = compare (y) (k)
		if ((LT == d) || (EQ == d)) node((ab.filterGT) (k), x, b)
		else node ((aa.filterGT) (k), y, node(ab,x,b))
	      }
	    }
	  }
	}


	def filterGE (k:a) = {
	  val d = compare (x) (k)
	  if ((LT == d)) (b.filterGE) (k)
	  else {
	    (a.match)[SplayHeap[a]] {
	      def Nil = this
	      def Tree(aa,y,ab) = {
		val d = compare (y) (k)
		if ((LT == d)) node((ab.filterGE) (k), x, b)
		else node ((aa.filterGE) (k), y, node(ab,x,b))
	      }
	    }
	  }
	}


	def partitionLT_GE (k:a) = {
	  val d = compare (x) (k)
	  if ((GT == d) || (EQ == d)) {
	    (a.match)[(SplayHeap[a],SplayHeap[a])] {
	      def Nil = (empty,this)
	      def Tree(aa,y,ab) = {
		val d = compare (y) (k)
		if ((GT == d) || (EQ == d)) {
		  val (small,big) = (aa.partitionLT_GE) (k)
		  (small, node(big, y, node(ab,x,b)))
		} else {
		  val (small,big) = (ab.partitionLT_GE) (k)
		  (node(aa,y,small), node(big,x,b))
		}
	      }
	    }
	  } else {
	    (b.match)[(SplayHeap[a],SplayHeap[a])] {
	      def Nil = (this,empty)
	      def Tree(ba,y,bb) = {
		val d = compare (y) (k)
		if ((GT == d) || (EQ == d)) {
		  val (small,big) = (ba.partitionLT_GE) (k)
		  (node(a,x,small), node(big,y,bb))
		} else {
		  val (small,big) = (bb.partitionLT_GE) (k)
		  (node(node(a,x,ba), y, small), big)
		}
	      }
	    }
	  }
	}


	def partitionLE_GT (k:a) = {
	  val d = compare (x) (k)
	  if ((GT == d)) {
	    (a.match)[(SplayHeap[a],SplayHeap[a])] {
	      def Nil = (empty,this)
	      def Tree(aa,y,ab) = {
		val d = compare (y) (k)
		if ((GT == d)) {
		  val (small,big) = (aa.partitionLE_GT) (k)
		  (small, node(big, y, node(ab,x,b)))
		} else {
		  val (small,big) = (ab.partitionLE_GT) (k)
		  (node(aa,y,small), node(big,x,b))
		}
	      }
	    }
	  } else {
	    (b.match)[(SplayHeap[a],SplayHeap[a])] {
	      def Nil = (this,empty)
	      def Tree(ba,y,bb) = {
		val d = compare (y) (k)
		if ((GT == d)) {
		  val (small,big) = (ba.partitionLE_GT) (k)
		  (node(a,x,small), node(big,y,bb))
		} else {
		  val (small,big) = (bb.partitionLE_GT) (k)
		  (node(node(a,x,ba), y, small), big)
		}
	      }
	    }
	  }
	}


	def partitionLT_GT (k:a) = {
	  val d = compare (x) (k)
	  if ((GT == d)) {
	    (a.match)[(SplayHeap[a],SplayHeap[a])] {
	      def Nil = (empty,this)
	      def Tree(aa,y,ab) = {
		val d = compare (y) (k)
		if ((GT == d)) {
		  val (small,big) = (aa.partitionLT_GT) (k)
		  (small, node(big, y, node(ab,x,b)))
		} else if ((LT == d)) {
		  val (small,big) = (ab.partitionLT_GT) (k)
		  (node(aa, y, small), node(big, x, b))
		} else {
		  ((aa.filterLT) (k), node((ab.filterGT) (k), x, b))
		}
	      }
	    }
	  } else if ((LT == d)) {
	    (b.match)[(SplayHeap[a],SplayHeap[a])] {
	      def Nil = (this, empty)
	      def Tree(ba,y,bb) = {
		val d = compare (y) (k)
		if ((GT == d)) {
		  val (small,big) = (ba.partitionLT_GT) (k)
		  (node(a, x, small), node(big, y, bb))
		} else if ((LT == d)) {
		  val (small,big) = (bb.partitionLT_GT) (k)
		  (node(node(a, x, ba), y, small),  big)
		} else {
		  (node(a, x, (ba.filterLT) (k)), (bb.filterGT) (k))
		}
	      }
	    }
	  } else {
	    ((a.filterLT) (k), (b.filterGT) (k))
	  }
	}




        // ----------- OrdCollX methods -----------
        // Destructors

	def minView = {
	  def minv (a:SplayHeap[a]) (x:a) (b:SplayHeap[a]) : (a,SplayHeap[a])= {
	    (a.match)[(a,SplayHeap[a])] {
	      def Nil = (x,b)
	      def Tree(a',x',b') = (a'.match)[(a,SplayHeap[a])] {
		def Nil = (x', node(b',x,b))
		def Tree(a'',x'',b'') = {
		  val (w,ab) = minv (a'') (x'') (b'')
		  (w, node(ab, x',node(b', x, b)))
		}
	      }
	    }
	  }
	  val (y,ys) = minv (a) (x) (b)
	  Maybe.Just(y,ys)
	}

        def minElem = {
	  def minel (xs:SplayHeap[a]) (x:a) : a = {
	    (xs.match)[a] {
	      def Nil = x
	      def Tree(a,x,b) = minel (a) (x)
	    }
	  }
	  minel (a) (x)
	}

	def maxView = {
	  def maxv (a:SplayHeap[a]) (x:a) (b:SplayHeap[a]) : (SplayHeap[a],a)= {
	    (b.match)[(SplayHeap[a],a)] {
	      def Nil = (a,x)
	      def Tree(b,y,c) = (c.match)[(SplayHeap[a],a)] {
		def Nil = (node(a,x,b), y)
		def Tree(c,z,d) = {
		  val (cd,w) = maxv (c) (z) (d)
		  (node(node(a,x,b), y, cd), w)
		}
	      }
	    }
	  }
	  val (ys,y) = maxv (a) (x) (b)
	  Maybe.Just(ys,y)
	}

        def maxElem = {
	  def maxel (x:a) (xs:SplayHeap[a]) : a = {
	    (xs.match)[a] {
	      def Nil = x
	      def Tree(a,x,b) = maxel (x) (b)
	    }
	  }
	  maxel (x) (b)
	}

        // Observers
        def toOrdSeq = {
	  def tos (xs:SplayHeap[a]) (rest:List[a]) : List[a] = {
	    (xs.match)[List[a]] {
	      def Nil = rest
	      def Tree(a,x,b) = tos (a) (List.cons (x) (tos (b) (rest)))
	    }
	  }
	  tos (this) (List.empty)
	}

        // Folds
        def foldr[b] (f: a -> b -> b) (e:b) = (a.foldr) (f) (f (x) ((b.foldr) (f) (e)))
        def foldl[b] (f: b -> a -> b) (e:b) = (b.foldl) (f) (f ((a.foldl) (f) (e)) (x))

        def foldr1 (f: a -> a -> a) = {
	  def myfold (f:a->a->a) (x:a) (xs:SplayHeap[a]) : a = {
	    (xs.match)[a] {
	      def Nil = x
	      def Tree(a,y,b) = f (x) ((a.foldr) (f) (myfold (f) (y) (b)))
	    }
	  }
	  (a.foldr) (f) (myfold (f) (x) (b))
	}

        def foldl1 (f: a -> a -> a) = {
	  def myfold (f:a->a->a) (xs:SplayHeap[a]) (y:a) : a = {
	    (xs.match)[a] {
	      def Nil = y
	      def Tree(a,x,b) = f ((b.foldl) (f) (myfold (f) (a) (x))) (y)
	    }
	  }
	  (b.foldl) (f) (myfold (f) (a) (x))
	}

	def iterater (f) = {
	  (a.iterater) (f)
	  f (x)
	  (b.iterater) (f)
	}

	def iteratel (f) = {
	  (b.iterater) (f)
	  f (x)
	  (a.iterater) (f)
	}

      }
      this

    }

  }
  this

}
