// -*- Funnel -*-
/*  FUNNEL LIBRARY (EDISON PORT)
**  
**  Splay Heap -- support for heaps in Funnel
**  
**  30/06/2001, Yacine Saidji
*/

/**
 * $Author: ysaidji $
 * $Id: SplayHeapType.m4,v 1.2 2001/09/12 13:25:09 ysaidji Exp $
 */

newtype SplayHeap[a] = {

  with MaybeType

  // ----------- CollX methods -----------
  // Constructors
  def insert (x:a): SplayHeap[a]
  def insertSeq (xs:__Seq__[a]): SplayHeap[a]
  def union (xs:__CollX__[a]): SplayHeap[a]

  // Deletions
  def delete (x:a): SplayHeap[a]
  def deleteAll (x:a): SplayHeap[a]
  def deleteSeq (xs:__Seq__[a]): SplayHeap[a]

  // visitor suppport
  def match[b] ( v:SplayHeapVisitor[a,b]) : b

  // Observers
  def isEmpty: boolean
  def size: int
  def member (x:a): boolean
  def count (x:a): int

  // ----------- Coll methods -----------
  // Observers
  def lookup (x:a): a
  def lookupM (x:a) : Maybe[a]
  def lookupAll (x:a): __Seq__[a]
  def lookupWithDefault (x:a) (d:a): a
  def toSeq: __Seq__[a]

  // Filters and partitions
  def filter (p: a -> boolean): SplayHeap[a]
  def partition (p: a -> boolean): (SplayHeap[a], SplayHeap[a])
  def exist(p:a -> boolean): boolean

  // Folds
  def fold[b](f: a -> b -> b)(e: b): b
  def fold1(f: a -> a -> a): a
  def iterate(f:a ->()) : ()

  // ----------- OrdCollX methods -----------
  // Constructor
  def unsafeInsertMin (x:a): SplayHeap[a]
  def unsafeInsertMax (x:a): SplayHeap[a]
  def unsafeAppend (h:__OrdCollX__[a]): SplayHeap[a]

  // Deletions
  def deleteMin: SplayHeap[a]
  def deleteMax: SplayHeap[a]

  // Filters and partitions
  def filterLT (x:a): SplayHeap[a]
  def filterLE (x:a): SplayHeap[a]
  def filterGT (x:a): SplayHeap[a]
  def filterGE (x:a): SplayHeap[a]
  def partitionLT_GE (x:a): (SplayHeap[a],SplayHeap[a])
  def partitionLE_GT (x:a): (SplayHeap[a],SplayHeap[a])
  def partitionLT_GT (x:a): (SplayHeap[a],SplayHeap[a])


  // ----------- OrdColl methods -----------
  // Destructors
  def minView: Maybe[(a,SplayHeap[a])]
  def minElem: a
  def maxView: Maybe[(SplayHeap[a],a)]
  def maxElem: a

  // Observers
  def toOrdSeq: __Seq__[a]

  // Folds
  def foldr[b](f: a -> b -> b)(e: b): b
  def foldl[b](f: b -> a -> b)(e: b): b
  def foldr1(f: a -> a -> a): a
  def foldl1(f: a -> a -> a): a
  def iterater(f:a ->()) : ()
  def iteratel(f:a ->()) : ()

  // ----------- Private methods -----------
  def sz (n:int) : int
  def cnt (n:int) (x:a) : int
  def look (y:a) (rest:List[a]) : List[a]
  def tos (rest:List[a]) : List[a]

}

newtype SplayHeapVisitor[a, +b] = {
  def Nil: b
  def Tree(a: SplayHeap[a], x:a, b: SplayHeap[a]): b
}

type SplayHeapModule[a] = {

  // -------- Constructors -------- 
  def Nil : SplayHeap[a]
  def Tree (a: SplayHeap[a], x: a, b: SplayHeap[a]) : SplayHeap[a]

  // -------- static methods of CollXModule --------
  // Comparison
  def equivalent (x:a) (y:a) : boolean
  // Constructors
  def empty: SplayHeap[a]
  def single (x:a): SplayHeap[a]
  def unionSeq (s:__Seq__[__CollX__[a]]): SplayHeap[a]
  def fromSeq (xs:__Seq__[a]): SplayHeap[a]
  

  // -------- static methods of OrdCollXModule -------- 
  // Comparison
  def LT: int
  def GT: int
  def EQ: int
  def compare (x:a) (y:a) : int
  // Constructor
  def unsafeFromOrdSeq (xs:__Seq__[a]): SplayHeap[a]
  
}

newtype Equ[a] = {
  def equ (x:a) (y:a): boolean
}

newtype Ord[a] = {
  def lt (x:a) (y:a): boolean
  def leq (x:a) (y:a): boolean
}


newtype SplayHeapType = {
  def dummy:any
}


