// -*- Funnel -*-
/*  FUNNEL LIBRARY (EDISON PORT)
**  
**  List -- support for lists in Funnel
**  
**  12/04/2001, Yacine Saidji
*/

/**
 * $Author: ysaidji $
 * $Id: List.m4,v 1.9 2001/09/12 13:25:08 ysaidji Exp $
 */

def removeMe = include[ListType]

let List: ListModule = {

  with System

  // first the static methods
  // then Nil and Cons methods

  // some constructors
  def empty[a] = Nil[a]
  def single[a] (x) = Cons(x, empty[a])
  def cons[a] (x) (xs) = Cons(x, xs)
  def snoc[a] (xs) (x)= append xs (single x)
  def append[a] (xs) (ys) = xs.foldr (cons[a]) ys
  def fromList[a] (xs) = xs
  def copy[a] (n) (x) = if (n>0) (cons[a]) x ((copy[a]) (n-1) x) else empty[a]
  def tabulate[a] (n) (f) = {
    def tab(i:int) : List[a] =
      if (i>=n)
	empty[a]
      else
	(cons[a]) (f (i)) (tab (i+1))
    tab 0
  }

  // concat and reverse
  def concat[a] xss = xss.foldr (append[a]) (empty[a])
  def reverseOnto[a] xs ys = (append[a]) (xs.reverse) ys
  
  // maps and folds
  def concatMap[a,b] f xs = (concat[b]) (xs.map f)

  // unzips
  def unzip[a,b] xys = (xys.map (x,y|x), xys.map (x,y|y))
  def unzip3[a,b,c] xyzs = (xyzs.map (x,y,z|x), xyzs.map (x,y,z|y), xyzs.map (x,y,z|z))
  def unzipWith[a,b](f)[c](g) xs = (xs.map f, xs.map g)
  def unzipWith3[a,b](f)[c](g)[d](h) xs = (xs.map f, xs.map g, xs.map h)

  def Nil[a] = {

    // Destructors
    def lview = Maybe.Nothing[(a,List[a])]
    def lhead = System.error "list empty"
    def ltail = empty
    def rview = Maybe.Nothing[(List[a],a)]   
    def rhead = System.error "list empty"
    def rtail = empty

    // visitor suppport
    def match[b] (v) = v.Nil

    // Observers
    def isEmpty = true
    def size = 0
    def toList = empty

    // Reverse
    def reverse = empty

    // Maps and folds
    def map f = empty
    def foldr f c = c
    def foldl f c = c
    def foldr1 f = System.error "list empty"
    def foldl1 f = System.error "list empty"
    def reducer f c = c
    def reducel f c = c
    def reduce1 f = System.error "list empty"
    def iterate (f) = ()


    // Subsequences
    def take i = empty
    def drop i = empty
    def splitAt i = (empty,empty)
    def subseq i len = empty

    // Predicate based operations
    def filter p = empty
    def partition p = (empty, empty)
    def takeWhile p = empty
    def dropWhile p = empty
    def splitWhile p = (empty, empty)
    def exist (p)  = false

    // Index based operations
    def inBounds i = false
    def lookupM (i) = Maybe.Nothing[a]
    def lookup i = System.error "list empty"
    def lookupWithDefault d i = d
    def update i y = empty
    def adjust f i = empty
    def mapWithIndex f = empty
    def foldrWithIndex f c = c
    def foldlWithIndex f c = c
    def iterateWithIndex (f) = ()

    // Zips and unzips
    def zip ys = empty
    def zip3 ys zs = empty
    def zipWith f ys = empty
    def zipWith3 f ys zs = empty

  }

  def Cons[a](x, xs) = {

    // Destructors
    def lview = Maybe.Just(x,xs)
    def lhead = x
    def ltail = xs
    def rview = Maybe.Just((xs.rtail),(xs.rhead))
    def rhead = if (xs.isEmpty) x else xs.rhead
    def rtail = if (xs.isEmpty) empty else cons x (xs.rtail)

    // visitor suppport
    def match[b] (v) = v.Cons(x,xs)

    // Observers
    def isEmpty = false
    def size = 1 + xs.size
    def toList = cons x xs

    // Reverse
    def reverse = foldl ( l:List[a] | x | Cons(x,l)) (empty)

    // Maps and folds
    def map f = cons (f x) (xs.map f)
    def foldr f c = f x (xs.foldr f c) 
    def foldl f c = xs.foldl f (f c x)

    def foldr1 (f) = {
      def fr (y:a) (ys:List[a]) : a= {
	if (ys.isEmpty) y
	else f (y) (fr (ys.lhead) (ys.ltail))
      }
      fr (x) (xs)
    }

    def foldl1 (f) = xs.foldl f x
    // ToDo : rewrite the following three (if necessary)
    def reducer (f) (c) = foldr (f) (c)
    def reducel (f) (c) = foldl (f) (c)
    def reduce1 (f) = foldl1 (f)
    def iterate f = (f(x); xs.iterate f)

    // Subsequences	
    def take i = if (i<=0) empty else cons x (xs.take (i-1))
    def drop i = if (i<=0) cons x xs else xs.drop (i-1)
    def splitAt i = (take i, drop i)
    def subseq i len = (drop i).take len

    // Predicate based operations
    def filter p = if (p x) cons x (xs.filter p) else (xs.filter p)
    def partition p = (filter p, filter (y: a| !(p y)))
    def takeWhile p = if (p x) cons x (xs.takeWhile p) else empty
    def dropWhile p = if (p x) xs.dropWhile p else cons x xs
    def splitWhile p = (takeWhile p, dropWhile p)
    def exist (p)  = if (p (x)) true else (xs.exist (p))

    // Index based operations
    def inBounds i = (0.toInt<=i) && (i<size)
    def lookupM (i) = if (!(inBounds i)) Maybe.Nothing[a] else Maybe.Just(lookup (i))
    def lookup i = if (!(inBounds i)) System.error "out of bounds index"
		   else if (i==0) x else xs.lookup (i-1)
    def lookupWithDefault d i = if (!(inBounds i)) d
				else lookup i
    def update i y = if (!(inBounds i)) cons x xs
		     else if (i==0) cons y xs else cons x (xs.update (i-1) y)
    def adjust f i = if (!(inBounds i)) cons x xs
		     else if (i==0) cons (f x) xs else cons x (xs.adjust f (i-1))
    def mapWithIndex[b] f =  {
      def mapwi (f: int -> a -> b) (lst: List[a]) (i: int) : List[b] =
	if (lst.isEmpty)
	    List.empty[b]
	else 
	    cons[b] (f i (lst.lhead)) (mapwi (f) (lst.ltail) (i+1))
      mapwi f (cons[a] x xs) 0
    }
    def foldrWithIndex[b] f c = {
      def foldrwi (f: int -> a -> b -> b) (lst: List[a]) (c: b) (i: int) : b =
	if (lst.isEmpty)
	  c
	else
	  f (i) (lst.lhead) (foldrwi (f) (lst.ltail) (c) (i+1))
      foldrwi (f) (cons x xs) (c) (0)
    }
    def foldlWithIndex[b] f c = {
      def foldlwi (f: b -> int -> a -> b) (lst: List[a]) (c: b) (i: int) : b = {
	if (lst.isEmpty) {
	  c
	} else {
	  foldlwi (f) (lst.ltail) (f (c) (i) (lst.lhead)) (i+1)
	}
      }
//	  f (foldlwi (f) (lst.ltail) c (i+1)) (i) (lst.lhead)
      foldlwi (f) (cons x xs) (c) (0)
    }

    def iterateWithIndex f =  {
      def iterwi (f:int->a->()) (lst:List[a]) (i:int) : () = {
	if (lst.isEmpty)
	  ()
	else {
	  f (i) (lst.lhead)
	  iterwi (f) (lst.ltail) (i+1)
	}
      }
      iterwi (f) (cons x xs) (0)
    }

    // Zips and unzips
    def zip ys = if (ys.isEmpty) empty
		 else cons (x, ys.lhead) ((xs.zip) (ys.ltail))

    def zip3[b,c] ys zs = if ((ys.isEmpty) || (zs.isEmpty)) empty
			  else cons (x, ys.lhead, zs.lhead) ((xs.zip3[b,c]) (ys.ltail) (zs.ltail))

    def zipWith[b,c] f ys = if (ys.isEmpty) empty
			    else cons (f x (ys.lhead)) ((xs.zipWith[b,c]) f (ys.ltail))

    def zipWith3[b,c,d] f ys zs = if ((ys.isEmpty) || (zs.isEmpty)) empty
			    else cons (f x (ys.lhead) (zs.lhead)) ((xs.zipWith3[b,c,d]) f (ys.ltail) (zs.ltail))

  }
}
