// -*- Funnel -*-
/*  FUNNEL LIBRARY (EDISON PORT)
**  
**  List -- support for lists in Funnel
**  
**  12/04/2001, Yacine Saidji
*/

/**
 * $Author: ysaidji $
 * $Id: ListType.m4,v 1.5 2001/07/10 01:31:55 ysaidji Exp $
 */

newtype List[a] = {

  with MaybeType
  
  // destructors
  def lview: Maybe[(a,List[a])]
  def lhead: a
  def ltail: List[a]
  def rview: Maybe[(List[a],a)]
  def rhead: a
  def rtail: List[a]

  // visitor suppport
  def match[b] ( v:ListVisitor[a,b]) : b

  // observers
  def isEmpty: boolean
  def size: int
  def toList: List[a]
  
  // concat and reverse
  def reverse: List[a]

  // maps and folds
  def map[b](f: a -> b): List[b]
  def foldr[b](f: a -> b -> b)(y: b): b
  def foldl[b](f: b -> a -> b)(y: b): b
  def foldr1(f: a -> a -> a): a
  def foldl1(f: a -> a -> a): a
  def reducer(f: a -> a -> a)(c: a): a
  def reducel(f: a -> a -> a)(c: a): a
  def reduce1(f: a -> a -> a): a
  def iterate (f:a->()) : ()

  // subsequences
  def take(n: int): List[a]
  def drop(n: int): List[a]
  def splitAt(n: int): (List[a], List[a])
  def subseq(b: int) (e: int): List[a]

  // predicate-based operations
  def filter(p: a -> boolean): List[a]
  def partition(p: a -> boolean): (List[a], List[a])
  def takeWhile(p: a -> boolean): List[a]
  def dropWhile(p: a -> boolean): List[a]
  def splitWhile(p: a -> boolean): (List[a], List[a])
  def exist(p: a -> boolean) : boolean

  // index-based operations
  def inBounds(i: int): boolean
  def lookup(i: int): a
  def lookupM(i: int): Maybe[a]
  def lookupWithDefault(d: a) (i: int): a
  def update(i: int) (x: a): List[a]
  def adjust(f: a -> a)(i: int): List[a]
  def mapWithIndex[b](f: int -> a -> b): List[b]
  def foldrWithIndex[b](f: int -> a -> b -> b)(c: b): b
  def foldlWithIndex[b](f: b -> int -> a -> b)(c: b): b
  def iterateWithIndex (f:int->a->()) : ()

  // zips and unzips
  def zip[b](ys: __Seq__[b]): List[(a, b)]
  def zip3[b,c](ys: __Seq__[b])(zs: __Seq__[c]): List[(a, b,c)]
  def zipWith[b,c](f: a -> b -> c)(ys:__Seq__[b]): List[c]
  def zipWith3[b,c,d](f: a -> b -> c -> d)(ys:__Seq__[b])(zs: __Seq__[c]): List[d]


}

newtype ListVisitor[a, +b] = {
  def Nil: b
  def Cons(x: a, xs: List[a]): b
}

type ListModule = {
  def Nil[a]: List[a]
  def Cons[a](x: a, xs: List[a]): List[a]

  // Constructors
  def cons[a] (x:a) (xs:__Seq__[a]) : List[a]
  def snoc[a] (xs:__Seq__[a]) (x:a): List[a]
  def append[a] (xs:__Seq__[a]) (ys:__Seq__[a]) : List[a]
  def fromList[a] (xs:__Seq__[a]) : List[a]
  def empty[a] : List[a]
  def single[a] (x:a) : List[a]
  def copy[a] (n:int) (x: a) : List[a]
  def tabulate[a] (n:int) (f:int->a) : List[a]

  // Static methods
  def concat[a](xss: __Seq__[__Seq__[a]]): List[a]
  def reverseOnto[a] (xs: __Seq__[a]) (ys: __Seq__[a]): List[a]
  def concatMap[a,b](f: a -> List[b]) (xs: __Seq__[a]): List[b]
  def unzip[a,b](xys: __Seq__[(a,b)]): (List[a], List[b])
  def unzip3[a,b,c](xyzs: __Seq__[(a,b,c)]): (List[a], List[b], List[c])
  def unzipWith[a,b](f: a -> b)[c](g: a -> c)(xs: __Seq__[a]):(List[b], List[c])
  def unzipWith3[a,b](f: a -> b)[c](g: a -> c)[d](h: a -> d)(xs: __Seq__[a]):(List[b], List[c],List[d])
}

newtype ListType = {
  def dummy:any
}
