// -*- Funnel -*-

/**
 * $Author: ysaidji $
 * $Id: SeqDefault.m4,v 1.1 2001/07/02 12:09:21 ysaidji Exp $
 */

def removeMe = include[__SeqDefaultType__]

// __begin_Seq__
def SeqDefault (sm:__SeqModule__) : SeqDefaultModule = {

  with Maybe

  def compose[a,b,c] (f:b->c,g:a->b) : (a->c) = (x:a|f(g(x))) 

  def flip[a,b] (f:a->b->b) : b->a->b = (y:b | x:a | f (x) (y))

  def snocUsingAppend[a] (s) (x) = sm.append (s) (sm.single (x))

  def snocUsingFoldr[a] (s) (x) = s.foldr (sm.cons[a]) (sm.single (x))

  def appendUsingFoldr[a] (s) (t) = if (t.isEmpty) s else s.foldr (sm.cons[a]) (t)

  def rviewDefault[a] (xs) = {
    if (xs.isEmpty) (Maybe.Nothing)[(__Seq__[a],a)] else (Maybe.Just) ( ((xs.rtail), (xs.rhead)) )
  }

  def rtailUsingLview[a] (xs) = {
    def rt (x:a) (xs:__Seq__[a]) : __Seq__[a]= {
      (xs.lview).match {
	def Nothing = sm.empty[a]
	def Just (y,ys) = sm.cons[a] (x) (rt (y) (ys))
      }
    }

    (xs.lview).match[__Seq__[a]] {
      def Nothing = sm.empty
      def Just(x,xs) = rt (x) (xs)
    }
  }

  def concatUsingFoldr[a] (xss) = xss.foldr (sm.append[a]) (sm.empty[a])


  def reverseUsingReverseOnto[a] (s) = sm.reverseOnto (s) (sm.empty)

  def reverseUsingLists[a] (s) = sm.fromList ((s.toList).reverse)

  def reverseOntoUsingFoldl[a] (xs) (ys) = xs.foldl (flip (sm.cons[a])) (ys)

  def reverseOntoUsingReverse[a] (xs) (ys) = sm.append (xs.reverse) (ys)

  def fromListUsingCons[a] (xs) = xs.foldr (sm.cons[a]) (sm.empty)

  def toListUsingFoldr[a] (xs) = xs.foldr (List.cons[a]) (List.empty)

  def mapUsingFoldr[a,b] (f) (xs) = xs.foldr (compose[a,b,__Seq__[b]->__Seq__[b]] (sm.cons[b],f)) (sm.empty)

  def concatMapUsingFoldr[a,b] (f) (xs) = xs.foldr[__Seq__[b]] (compose (sm.append[b],f)) (sm.empty[b])

  def foldrUsingLists[a,b] (f) (e) (xs) = (xs.toList).foldr (f) (e)

  def foldlUsingLists[a,b] (f) (e) (xs) = (xs.toList).foldl (f) (e)

  def foldr1UsingLists[a] (f) (xs) = (xs.toList).foldr1 (f)

  def foldl1UsingLists[a] (f) (xs) = (xs.toList).foldl1 (f)

  def foldr1UsingLview[a] (f) (xs) = {
    def fr1 (x:a) (xs:__Seq__[a]) : a = {
      (xs.lview).match[a] {
	def Nothing = x
	def Just (y,ys) = f (x) (fr1 (y) (ys))
      }
    }
    
    (xs.lview).match[a] {
      def Nothing = System.error "foldr1: empty sequence"
      def Just (x,xs) = fr1 (x) (xs)
    }
  }

  def foldl1UsingLview[a] (f) (xs) = {
    (xs.lview).match {
      def Nothing = System.error "foldl1: empty sequence"
      def Just (x,xs) = (xs).foldl (f) (x)
    }
  }

  def reducerUsingReduce1[a] (f) (e) (s) = {
    if (s.isEmpty) e
    else f (s.reduce1 (f)) (e)
  }

  def reducelUsingReduce1[a] (f) (e) (s) = {
    if (s.isEmpty) e
    else f (e) (s.reduce1 (f))    
  }

  def reduce1UsingLists[a] (f) (xs) = (xs.toList).reduce1 (f)

  def copyUsingLists[a] (n) (x) = sm.fromList (List.copy[a] (n) (x))

  def tabulateUsingLists[a] (n) (f) = sm.fromList (List.tabulate[a] (n) (f))

  def tabulateUsingCons[a] (n) (f) = {
    def tab (i:int) : __Seq__[a]= if (i == n) sm.empty else sm.cons (f (i)) (tab (i+1))

    if (n <= 0) sm.empty
    else tab (0)
  }

  def inBoundsUsingDrop[a] (s) (i) = (i >= 0) && !((s.drop (i)).isEmpty)

  def inBoundsUsingLookupM[a] (s) (i) = {
    (s.lookupM (i)).match {
      def Nothing = false
      def Just (x) = true
    }
  }

  def inBoundsUsingSize[a] (s) (i) = (i >= 0) && (i < (s.size))

  def lookupUsingLookupM[a] (s) (i) = {
    (s.lookupM (i)).match {
      def Nothing = System.error "lookup: bad subscript"
      def Just (x) = x
    }
  }

  def lookupUsingDrop[a] (s) (i) = {
    val s' = s.drop (i)
    if ( (i<0) || s'.isEmpty ) System.error "lookup: bad subscript"
    else s'.lhead
  }

  def lookupWithDefaultUsingLookupM[a] (d) (s) (i) = {
    (s.lookupM (i)).match {
      def Nothing = d
      def Just (x) = x
    }
  }

  def lookupWithDefaultUsingDrop[a] (d) (s) (i) = {
    val s' = s.drop (i)
    if ( (i<0) || s'.isEmpty ) d
    else s'.lhead
  }

  def lookupMUsingDrop[a] (s) (i) = {
    val s' = s.drop (i)
    if ( (i<0) || s'.isEmpty ) Maybe.Nothing
    else Maybe.Just(s'.lhead)    
  }

  def filterUsingLview[a] (p) (xs) = {
    (xs.lview).match[__Seq__[a]] {
      def Nothing = sm.empty
      def Just (x,xs) = if (p(x)) sm.cons (x) (xs.filter (p)) else xs.filter (p)
    }
  }

  def filterUsingLists[a] (p) (xs) = sm.fromList((xs.toList).filter (p))

  def filterUsingFoldr[a] (p) (xs) = {
    def pcons (x:a) (xs:__Seq__[a]) = if (p(x)) sm.cons (x) (xs) else (xs)
    
    xs.foldr (pcons) (sm.empty)
  }

  def partitionUsingLists[a] (p) (xs) = {
    val (ys,zs) = (xs.toList).partition (p)
    (sm.fromList (ys), sm.fromList (zs))
  }

  def partitionUsingFoldr[a] (p) (xs) = {
    def pcons (x:a) (xs,xs') : (__Seq__[a], __Seq__[a])= if (p(x)) (sm.cons (x) (xs), xs') else (xs, sm.cons (x) (xs'))
    
    xs.foldr (pcons) (sm.empty,sm.empty)
  }

  def updateUsingAdjust[a] (i) (y) (s) = s.adjust (x | y) (i)  

  def updateUsingSplitAt[a] (i) (x) (xs) = {
    if (i < 0) xs
    else {
      val (ys,zs) = xs.splitAt (i)
      if (zs.isEmpty) xs
      else sm.append (ys) (sm.cons (x) (zs.ltail))
    }
  }

  def adjustUsingLists[a] (f) (i) (s) = sm.fromList ((s.toList).adjust (f) (i))  

  def adjustUsingSplitAt[a] (f) (i) (xs) = {
    if (i < 0) xs
    else {
      val (ys,zs) = xs.splitAt (i)
      (zs.lview).match {
	def Nothing = xs
	def Just (z,zs') = sm.append (ys) (sm.cons (f (z)) (zs'))
      }
    }
  }

  def mapWithIndexUsingLists[a,b] (f) (xs) = sm.fromList ((xs.toList).mapWithIndex (f))

  def foldrWithIndexUsingLists[a,b] (f) (e) (xs) = (xs.toList).foldrWithIndex (f) (e)

  def foldlWithIndexUsingLists[a,b] (f) (e) (xs) = (xs.toList).foldlWithIndex (f) (e)

  def takeUsingLists[a] (i) (s) = sm.fromList ((s.toList).take (i))

  def takeUsingLview[a] (i) (s) = {
    if (i <= 0) sm.empty
    else {
      (s.lview).match[__Seq__[a]] {
	def Nothing = sm.empty
	def Just (x,xs') = sm.cons (x) ((xs').take (i-1))
      }
    }
  }

  def dropUsingLists[a] (i) (s) = sm.fromList ((s.toList).drop (i))

  def dropUsingLtail[a] (i) (xs) = {
    if ( (i <= 0) || (xs.isEmpty) ) xs
    else dropUsingLtail[a] (i-1) (xs.ltail)
  }

  def splitAtDefault[a] (i) (s) = (s.take (i), s.drop (i))

  def splitAtUsingLview[a] (i) (xs) = {
    if (i <= 0) (sm.empty,xs)
    else {
      (xs.lview).match[(__Seq__[a],__Seq__[a])] {
	def Nothing = (sm.empty,sm.empty)
	def Just(x,xs') = {
	  val (ys,zs) = splitAtUsingLview[a] (i-1) (xs')
	  (sm.cons (x) (ys),zs)
	}
      }
    }
  }

  def subseqDefault[a] (i) (len) (xs) = (xs.drop (i)).take (len)

  def takeWhileUsingLview[a] (p) (xs) = {
    (xs.lview).match[__Seq__[a]] {
      def Just (x,xs') =  if (p(x)) sm.cons (x) (takeWhileUsingLview[a] (p) (xs')) else sm.empty
      def Nothing = sm.empty
    }
  }

  def dropWhileUsingLview[a] (p) (xs) = {
    (xs.lview).match[__Seq__[a]] {
      def Just (x,xs') = if (p(x)) dropWhileUsingLview[a] (p) (xs') else xs
      def Nothing = xs
    }
  }

  def splitWhileUsingLview[a] (p) (xs) = {
    (xs.lview).match[(__Seq__[a],__Seq__[a])] {
      def Just (x,xs') = {
	if (p(x)) {
	  val (front,back) = splitWhileUsingLview[a] (p) (xs')
	  (sm.cons (x) (front), back)
	} else {
	  (sm.empty,xs)
	}
      }
      def Nothing = (sm.empty,xs)
    }
  }

  def zipUsingLview[a,b] (xs) (ys) = {
    (xs.lview).match[__Seq__[(a,b)]] {
      def Nothing = sm.empty
      def Just (x,xs') = {
	(ys.lview).match[__Seq__[(a,b)]] {
	  def Nothing = sm.empty
	  def Just (y,ys') = sm.cons (x,y) (zipUsingLview[a,b] (xs') (ys'))
	}
      }
    }
  }

  def zip3UsingLview[a,b,c] (xs) (ys) (zs) = {
    (xs.lview).match[__Seq__[(a,b,c)]] {
      def Nothing = sm.empty
      def Just (x,xs') = {
	(ys.lview).match[__Seq__[(a,b,c)]] {
	  def Nothing = sm.empty
	  def Just (y,ys') = {
	    (zs.lview).match[__Seq__[(a,b,c)]] {
	      def Nothing = sm.empty
	      def Just (z,zs') = sm.cons (x,y,z) (zip3UsingLview[a,b,c] (xs') (ys') (zs'))
	    }
	  }
	}
      }
    }
  }

  def zipWithUsingLview[a,b,c] (f) (xs) (ys) = {
    (xs.lview).match[__Seq__[c]] {
      def Nothing = sm.empty
      def Just(x,xs') = {
	(ys.lview).match[__Seq__[c]] {
	  def Nothing = sm.empty
	  def Just (y,ys') = sm.cons (f (x) (y)) (zipWithUsingLview[a,b,c] (f) (xs') (ys'))
	}
      }
    }
  }

  def zipWith3UsingLview[a,b,c,d] (f) (xs) (ys) (zs) = {
    (xs.lview).match[__Seq__[d]] {
      def Nothing = sm.empty
      def Just (x,xs') = {
	(ys.lview).match[__Seq__[d]] {
	  def Nothing = sm.empty
	  def Just (y,ys') = {
	    (zs.lview).match[__Seq__[d]] {
	      def Nothing = sm.empty
	      def Just (z,zs') = sm.cons (f (x) (y) (z)) (zipWith3UsingLview[a,b,c,d] (f) (xs') (ys') (zs'))
	    }
	  }
	}
      }
    }    
  }

  def zipUsingLists[a,b] (xs) (ys) = sm.fromList((xs.toList).zip (ys.toList))
  
  def zip3UsingLists[a,b,c] (xs) (ys) (zs) = sm.fromList((xs.toList).zip3[b,c] (ys.toList) (zs.toList))

  def zipWithUsingLists[a,b,c] (f) (xs) (ys) = sm.fromList((xs.toList).zipWith[b,c] (f) (ys.toList))

  def zipWith3UsingLists[a,b,c,d] (f) (xs) (ys) (zs) = sm.fromList((xs.toList).zipWith3[b,c,d] (f) (ys.toList) (zs.toList))

  def unzipUsingLists[a,b] (xys) = {
    val (xs,ys) = List.unzip (xys.toList)
    (sm.fromList (xs), sm.fromList (ys))
  }

  def unzipUsingFoldr[a,b] (xys) = {
    def pcons (x:a,y:b) (xs:__Seq__[a],ys:__Seq__[b])  = (sm.cons (x) (xs), sm.cons (y) (ys))
    xys.foldr (pcons) (sm.empty,sm.empty)
  }

  def unzip3UsingLists[a,b,c] (xyzs) = {
    val (xs,ys,zs) = List.unzip3 (xyzs.toList)
    (sm.fromList (xs), sm.fromList (ys), sm.fromList (zs))
  }

  def unzip3UsingFoldr[a,b,c] (xyzs) = {
    def tcons (x:a,y:b,z:c) (xs:__Seq__[a],ys:__Seq__[b],zs:__Seq__[c]) = (sm.cons (x) (xs), sm.cons (y) (ys), sm.cons (z) (zs))
    xyzs.foldr (tcons) (sm.empty,sm.empty,sm.empty)
  }

  def unzipWithUsingLists[a,b,c] (f) (g) (xys) = {
    val (xs,ys) = List.unzipWith[a,b] (f) [c] (g) (xys.toList)
    (sm.fromList (xs), sm.fromList (ys))
  }

  def unzipWithUsingFoldr[a,b,c] (f) (g) (xys) = {
    def pcons (e:a) (xs:__Seq__[b],ys:__Seq__[c]) = (sm.cons (f (e)) (xs), sm.cons (g (e)) (ys))
    xys.foldr (pcons) (sm.empty,sm.empty)   
  }

  def unzipWith3UsingLists[a,b,c,d] (f) (g) (h) (xyzs) = {
    val (xs,ys,zs) = List.unzipWith3[a,b] (f) [c] (g) [d] (h) (xyzs.toList)
    (sm.fromList (xs), sm.fromList (ys), sm.fromList (zs))    
  }

  def unzipWith3UsingFoldr[a,b,c,d] (f) (g) (h) (xyzs) = {
    def tcons (e:a) (xs:__Seq__[b],ys:__Seq__[c],zs:__Seq__[d]) = (sm.cons (f (e)) (xs), sm.cons (g (e)) (ys), sm.cons (h (e)) (zs))
    xyzs.foldr (tcons) (sm.empty,sm.empty,sm.empty)    
  }

}

// divert(0)
newtype __SeqDefault__ = {
  def empty:any
}
