// -*- Funnel -*-
/*  FUNNEL LIBRARY (EDISON PORT)
**  
**  SimpleQueue -- support for FIFO queue in Funnel
**  
**  29/06/2001, Yacine Saidji
*/

/**
 * $Author: ysaidji $
 * $Id: SimpleQueue.m4,v 1.2 2001/09/12 13:25:08 ysaidji Exp $
 */

def removeMe0 = include[SimpleQueueType]
def removeMe1 = include[__SeqDefaultType__]
def removeMe2 = include[__SeqDefault__]

let SimpleQueue : SimpleQueueModule = {

  let this = {

    // inherence
    with SeqDefault(this)

    def makeQ[a] (xs) (ys) = if (xs.isEmpty) Q[a](ys.reverse,xs) else Q[a](xs,ys)

    // some constructors
    def empty[a] = Q[a]( List.empty, List.empty )
    def single[a] (x) = Q[a] (List.single(x), List.empty)
    def cons[a] (x) (s) = Q[a] (List.cons (x) (s.xs), s.ys)
    def snoc[a] (s) (y) = if ((s.xs).isEmpty) (single (y)) else Q[a] (s.xs, List.cons (y) (s.ys))  
    def append[a] (xs) (ys) = {
      val xs1 = xs.xs
      val xs2 = ys.xs
      val ys1 = xs.ys
      val ys2 = ys.ys
      Q[a] (List.append (xs1) (List.reverseOnto (ys1) (xs2)), ys2)
    }
    def fromList[a] (xs) = Q[a] (xs, List.empty[a])
    def copy[a] (n) (x) = copyUsingLists[a] (n) (x)
    def tabulate[a] (n) (f) = tabulateUsingLists[a] (n) (f)

    // concat and reverse
    def concat[a] (xss) = concatUsingFoldr[a] (xss)
    def reverseOnto[a] (xs) (ys) = {
      val xs1 = xs.xs
      val xs2 = ys.xs
      val ys1 = xs.ys
      val ys2 = ys.ys
      Q[a] (List.append (ys1) (List.reverseOnto (xs1) (xs2)), ys2)
    }

    // maps and folds
    def concatMap[a,b] (f) (xs) = concatMapUsingFoldr[a,b] (f) (xs)

    // unzips
    def unzip[a,b] (xys) = unzipUsingLists[a,b] (xys)
    def unzip3[a,b,c] (xyzs) = unzip3UsingLists[a,b,c] (xyzs)
    def unzipWith[a,b](f)[c](g) (xs) = unzipWithUsingLists[a,b,c] (f) (g) (xs)
    def unzipWith3[a,b](f)[c](g)[d](h) (xs) = unzipWith3UsingLists[a,b,c,d] (f) (g) (h) (xs)

  

    def Q[a](xs:List[a],ys:List[a]) = {
      let this : SimpleQueue[a] = {

	// helpers
	def isSingle (xs:List[a]) : boolean = if (xs.isEmpty) false 
					      else if ((xs.ltail).isEmpty) true else false

	// Destructors
	def lview = {
	  if (xs.isEmpty) Maybe.Nothing[(a,SimpleQueue[a])]
	  else if (isSingle (xs)) Maybe.Just(xs.lhead, Q[a](ys.reverse, List.empty))
	  else Maybe.Just(xs.lhead, Q[a](xs.ltail, ys))
	}

	def lhead = {
	  if (xs.isEmpty) System.error "SimpleQueue.lhead: empty sequence"
	  else (xs.lhead)
	}

	def ltail = {
	  if (xs.isEmpty) this
	  else if (isSingle (xs)) Q[a](ys.reverse, List.empty)
	  else Q[a](xs.ltail, ys)
	}

	def rview = {
	  if (ys.isEmpty) {
	    (xs.rview).match[Maybe[(SimpleQueue[a],a)]] {
	      def Nothing = Maybe.Nothing
	      def Just(xs',x) = Maybe.Just (Q[a](xs',List.empty), x)
	    }
	  } else {
	    Maybe.Just(Q[a](xs,ys.ltail), ys.lhead)
	  }
	}

	def rhead = {
	  if (ys.isEmpty) {
	    if (xs.isEmpty) System.error "SimpleQueue.rhead: empty sequence"
	    else xs.rhead
	  } else {
	    ys.lhead
	  }
	}

	def rtail = {
	  if (ys.isEmpty) {
	    if (xs.isEmpty) this
	    else Q[a](xs.rtail, List.empty)
	  } else {
	    Q[a](xs,ys.ltail)
	  }
	}
	
	// visitor support
	def match[b] (v) = v.Q(xs,ys)

	// Observers
	def isEmpty = xs.isEmpty

	def size = xs.size + ys.size

	def toList = {
	  if (ys.isEmpty) xs
	  else List.append (xs) (ys.reverse)
	}
	
	// Reverse
	def reverse = {
	  if (ys.isEmpty) Q[a] (xs.reverse,List.empty)
	  else Q[a](ys,xs)
	}

	// local functions on lists
	def revfoldr[b] (f:a->b->b) (e:b) (s:List[a]) : b = {
	  if (s.isEmpty) e
	  else revfoldr (f) (f (s.lhead) (e)) (s.ltail)
	}

	def revfoldl[b] (f:b->a->b) (e:b) (s:List[a]) : b = {
	  if (s.isEmpty) e
	  else f (revfoldl (f) (e) (s.ltail)) (s.lhead)
	}

	// Maps and folds
	def map[b] (f) = Q[b](xs.map[b] (f), ys.map[b] (f))

	def foldr (f) (e) = xs.foldr (f) (revfoldr (f) (e) (ys))

	def foldl (f) (e) = revfoldl (f) (xs.foldl (f) (e)) (ys)

	def foldr1 (f) = {
	  if (ys.isEmpty) {
	    if (xs.isEmpty) {
	      System.error "SimpleQueue.foldr1: empty sequence"
	    } else {
	      xs.foldr1 (f)
	    }
	  } else {
	    xs.foldr (f) (revfoldr (f) (ys.lhead) (ys.ltail))
	  }
	}

	def foldl1 (f) = {
	  if (xs.isEmpty) System.error "SimpleQueue.foldl1: empty sequence"
	  else revfoldl (f) (xs.foldl1 (f)) (ys)
	}

	def reducer (f) (e) = reducerUsingReduce1[a] (f) (e) (this)
	def reducel (f) (e) = reducelUsingReduce1[a] (f) (e) (this)
	def reduce1 (f) = reduce1UsingLists (f) (this)

	def iterate (f) = {
	  val _ = xs.iterate (f)
	  val _ = (ys.reverse).iterate (f)
	  ()
	}

	// Subsequences
	def take (i) = takeUsingLists[a] (i) (this)
	def drop (i) = dropUsingLists[a] (i) (this)
	def splitAt (i) = splitAtDefault[a] (i) (this)
	def subseq (i) (len) = subseqDefault[a] (i) (len) (this)

	// Predicate based operations
	def filter (p) = makeQ (xs.filter (p)) (ys.filter (p))

	def partition (p) = {
	  val (xsT,xsF) = xs.partition (p)
	  val (ysT,ysF) = ys.partition (p)
	  (makeQ (xsT) (ysT), makeQ (xsF) (ysF))
	}

	def takeWhile (p) = takeWhileUsingLview[a] (p) (this)
	def dropWhile (p) = dropWhileUsingLview[a] (p) (this)
	def splitWhile (p) = splitWhileUsingLview[a] (p) (this)
	def exist (p) = (xs.exist (p)) || (ys.exist (p))

	// Index based operations
	def inBounds (i) = inBoundsUsingLookupM[a] (this) (i)
	def lookupM (i) = lookupMUsingDrop[a] (this) (i)
	def lookup (i) = lookupUsingLookupM[a] (this) (i)
	def lookupWithDefault (d) (i) = lookupWithDefaultUsingLookupM[a] (d) (this) (i)
	def update (i) (y) = updateUsingAdjust[a] (i) (y) (this)
	def adjust (f) (i) = adjustUsingLists[a] (f) (i) (this)
	def mapWithIndex[b] (f:int->a->b) = mapWithIndexUsingLists[a,b] (f) (this)
	def foldrWithIndex[b] (f:int->a->b->b) (c) = foldrWithIndexUsingLists[a,b] (f) (c) (this)
	def foldlWithIndex[b] (f:b->int->a->b) (c) = foldlWithIndexUsingLists[a,b] (f) (c) (this)
	def iterateWithIndex (f) = (toList).iterateWithIndex (f)

	// Zips and unzips
	def zip[b] (ys) = zipUsingLists[a,b] (this) (ys)
	def zip3[b,c] (ys) (zs) = zip3UsingLists[a,b,c] (this) (ys)  (zs)
	def zipWith[b,c] (f) (ys) = zipWithUsingLists[a,b,c] (f) (this) (ys)
	def zipWith3[b,c,d] (f) (ys) (zs) = zipWith3UsingLists[a,b,c,d] (f) (this) (ys) (zs)

      }
      this
    }
  }
  this
}
