// -*- Funnel -*-

/**
 * $Author: ysaidji $
 * $Id: SimpleQueueType.m4,v 1.2 2001/09/12 13:25:08 ysaidji Exp $
 */

def removeMe = include[MaybeType]

newtype SimpleQueue[a] = {

  // destructors
  def lview: Maybe[(a,SimpleQueue[a])]
  def lhead: a
  def ltail: SimpleQueue[a]
  def rview: Maybe[(SimpleQueue[a],a)]
  def rhead: a
  def rtail: SimpleQueue[a]

  // visitor suppport
  def match[b] ( v:SimpleQueueVisitor[a,b]) : b

  // observers
  def isEmpty: boolean
  def size: int
  def toList: List[a]
  
  // concat and reverse
  def reverse: SimpleQueue[a]

  // maps and folds
  def map[b](f: a -> b): SimpleQueue[b]
  def foldr[b](f: a -> b -> b)(y: b): b
  def foldl[b](f: b -> a -> b)(y: b): b
  def foldr1(f: a -> a -> a): a
  def foldl1(f: a -> a -> a): a
  def reducer(f: a -> a -> a)(c: a): a
  def reducel(f: a -> a -> a)(c: a): a
  def reduce1(f: a -> a -> a): a
  def iterate (f:a->()) : ()

  // subsequences
  def take(n: int): SimpleQueue[a]
  def drop(n: int): SimpleQueue[a]
  def splitAt(n: int): (SimpleQueue[a], SimpleQueue[a])
  def subseq(b: int) (e: int): SimpleQueue[a]

  // predicate-based operations
  def filter(p: a -> boolean): SimpleQueue[a]
  def partition(p: a -> boolean): (SimpleQueue[a], SimpleQueue[a])
  def takeWhile(p: a -> boolean): SimpleQueue[a]
  def dropWhile(p: a -> boolean): SimpleQueue[a]
  def splitWhile(p: a -> boolean): (SimpleQueue[a], SimpleQueue[a])
  def exist(p: a -> boolean) : boolean

  // index-based operations
  def inBounds(i: int): boolean
  def lookup(i: int): a
  def lookupM(i: int): Maybe[a]
  def lookupWithDefault(d: a) (i: int): a
  def update(i: int) (x: a): SimpleQueue[a]
  def adjust(f: a -> a)(i: int): SimpleQueue[a]
  def mapWithIndex[b](f: int -> a -> b): SimpleQueue[b]
  def foldrWithIndex[b](f: int -> a -> b -> b)(c: b): b
  def foldlWithIndex[b](f: b -> int -> a -> b)(c: b): b
  def iterateWithIndex (f:int->a->()) : ()

  // zips and unzips
  def zip[b](ys: __Seq__[b]): SimpleQueue[(a, b)]
  def zip3[b,c](ys: __Seq__[b])(zs: __Seq__[c]): SimpleQueue[(a, b,c)]
  def zipWith[b,c](f: a -> b -> c)(ys:__Seq__[b]): SimpleQueue[c]
  def zipWith3[b,c,d](f: a -> b -> c -> d)(ys:__Seq__[b])(zs: __Seq__[c]): SimpleQueue[d]

  // private methods
  def xs : List[a]
  def ys : List[a]

}

newtype SimpleQueueVisitor[a, +b] = {
  def Q(xs: List[a], xs: List[a]): b
}

type SimpleQueueModule = {

  def Q[a](xs:List[a],ys:List[a]) : SimpleQueue[a]

  def makeQ[a] (xs:List[a]) (ys:List[a]) : SimpleQueue[a]

  // Constructors
  def cons[a] (x:a) (xs:__Seq__[a]) : SimpleQueue[a]
  def snoc[a] (xs:__Seq__[a]) (x:a): SimpleQueue[a]
  def append[a] (xs:__Seq__[a]) (ys:__Seq__[a]) : SimpleQueue[a]
  def fromList[a] (xs:List[a]) : SimpleQueue[a]
  def empty[a] : SimpleQueue[a]
  def single[a] (x:a) : SimpleQueue[a]
  def copy[a] (n:int) (x: a) : SimpleQueue[a]
  def tabulate[a] (n:int) (f:int->a) : SimpleQueue[a]

  // Static methods
  def concat[a](xss: __Seq__[__Seq__[a]]): SimpleQueue[a]
  def reverseOnto[a] (xs: __Seq__[a]) (ys: __Seq__[a]): SimpleQueue[a]
  def concatMap[a,b](f: a -> SimpleQueue[b]) (xs: __Seq__[a]): SimpleQueue[b]
  def unzip[a,b](xys: __Seq__[(a,b)]): (SimpleQueue[a], SimpleQueue[b])
  def unzip3[a,b,c](xyzs: __Seq__[(a,b,c)]): (SimpleQueue[a], SimpleQueue[b], SimpleQueue[c])
  def unzipWith[a,b](f: a -> b)[c](g: a -> c)(xs: __Seq__[a]):(SimpleQueue[b], SimpleQueue[c])
  def unzipWith3[a,b](f: a -> b)[c](g: a -> c)[d](h: a -> d)(xs: __Seq__[a]):(SimpleQueue[b], SimpleQueue[c],SimpleQueue[d])
}

newtype SimpleQueueType = {
  def dummy:any
}
