// LZW compressor/decompressor
// Yacine Saidji

def removeMe0 = include[RedBlackTreeType]
def removeMe1 = include[RedBlackTree]
// include list, etc...


val StrOrd : Ord[string] = {
  def lt (x:string) (y:string): boolean = ((String.compareTo (x) (y)) < 0)
  def leq (x:string) (y:string): boolean = ((String.compareTo (x) (y)) <= 0)
}
val  StrEqu : Equ[string] = { def equ (x:string) (y:string) : boolean = (x.equals (y)) }
type StrMap = RedBlackTree[string,int]
let  StrMap : RedBlackTreeModule[string] = RedBlackTree[string](StrEqu,StrOrd)

def encode (str:string) : List[int] = {

  def code_string (w:string) (suffix:string) (table:StrMap) (output:List[int]) : List[int] = {
    val strlen = String.length (suffix)
    if ( 0 == strlen ) List.cons (table.lookup (w)) (output)
    else {
      val k       = String.substring (suffix) (0) (1)
      val suffix' = String.substring (suffix) (1) (strlen)
      val wk      = w + k
      if (table.member (wk)) code_string (wk) (suffix') (table) (output)
      else code_string (k) (suffix') (table.insert (wk) (table.size + 1)) (List.cons (table.lookup (w)) (output))      
    }
  }

  val initial_table = StrMap.fromSeq ( List.tabulate[(string,int)] (3) (i | ((((i+97).toChar)+""),(i+1)) ) )
  val strlen = String.length (str)
  val w      = String.substring (str) (0) (1)
  val suffix = String.substring (str) (1) (strlen)

  (code_string (w) (suffix) (initial_table) (List.empty[int])).reverse
}

val IntOrd : Ord[int] = {
  def lt (x:int) (y:int): boolean = (x < y)
  def leq (x:int) (y:int): boolean = (x <= y)
}
val  IntEqu : Equ[int] = { def equ (x:int) (y:int) : boolean = (x == y) }
type IntMap = RedBlackTree[int,string]
let  IntMap : RedBlackTreeModule[int] = RedBlackTree[int](IntEqu,IntOrd)

def decode (cs:List[int]) : string = {

  def decode_string (w:string) (str:string) (cs:List[int]) (table:IntMap) : string = {
    if (cs.isEmpty) str
    else {
      val code   = cs.lhead

      if (table.member (code)) {
	val k      = table.lookup (code)
	val finalk = String.substring (k) (0) (1)
	val table' = table.insert (table.size + 1) (w + finalk)
	val str'   = str + k
	decode_string (k) (str') (cs.ltail) (table')
      }
      else {
	val k      = w + String.substring (w) (0) (1)
	val table' = table.insert (code) (k)
	val str'   = str + k
	decode_string (k) (str') (cs.ltail) (table')
      }

    }
  }

  val initial_table = IntMap.fromSeq ( List.tabulate[(int,string)] (3) (i | ((i+1),(((i+97).toChar)+"")) ) )
  val w = initial_table.lookup(cs.lhead)
  decode_string (w) (w) (cs.ltail) (initial_table)
}

// the original stream
val original = "ababcbababaaaaaa"

// compress original stream
val compressed = encode (original)

// print compressed stream
def imprime (i:int) (x:int): () = Console.println ("["+i+"/"+x+"]")
val _ = compressed.mapWithIndex (imprime)

// decompress
val uncompressed = decode (compressed)

// print decompressed stream
System.println(uncompressed)

// test if okay !
uncompressed.equals(original)

