// let's compute the number pi to a couple of thousand decimal places

// Carries
def LOBYTE (x:int) : int = x % 256
def HIBYTE (x:int) : int = (x / 256) % 256
def IMAX (l:int, r:int) : int = if (l > r) l else r
def IMIN (l:int, r:int) : int = if (l < r) l else r

// List[int]
def cons (x:int) (xs:List[int]) : List[int] = List.cons[int] (x) (xs)
def single (x:int) : List[int] = List.single[int] (x)
def pad  (n:int) (u:List[int]) : List[int] = {
  if ( n > 0 ) pad (n-1) (cons 0 u) else u
}

val BI : double = (1.0/256.0)
def toDouble (u:List[int]) : double = u.foldr ( x | fv:double | (fv * BI ) + x.toDouble ) (0.0)

def fromDouble (fu:double, n:int) : List[int] = {

  def loop (j:int) (fu:double, u:List[int]): (double,List[int]) = {
    if (j == n) (fu,u)
    else {
      val i = fu.toInt
      val u' = cons (i) (u)
      val fu' = 256.0 * ( fu - (i.toDouble) )
      loop (j+1) (fu',u')
    }
  }
  val (fu',u) = loop (0) (fu,List.empty[int])
  u.reverse
}

def sqrt (x:double) : double = {
  primitive (1, "java.lang.Math", "sqrt", "(D)D") (x)
}


def zip (u:List[int]) (v:List[int]) : List[(int,int)] = {
  val ul = u.size
  val vl = v.size

  if ( ul > vl )       u.zip (pad (ul-vl) (v))
  else if ( vl > ul )  (pad (vl-ul) (u)).zip (v)
  else                 u.zip (v) 
}
def print (xs:List[int]) : () = xs.iterate ( x | Console.print ( x + " " ) )

// multiple precisio arithmetic operations done oncharacter strings,
// interpreted as radix 256 numbers.
// This set of routines collects the sipler operations.

// Adds the unsigned radix 256 integers u[] and v[]
// yielding an unsigned integer
def mpadd (u:List[int], v:List[int]) : List[int] = {

  /*
  Console.println("-- mpadd --")
  Console.print("u is ")
  print(u)
  Console.println("")
  Console.print("v is ")
  print(v)
  Console.println("")
  */
  val uv = zip (u) (v)
  def add (u:int,v:int) (w:List[int],ireg:int) : (List[int],int) = {
    val ireg' = u + v + HIBYTE(ireg)
    val w' = cons (LOBYTE(ireg')) (w)
    (w',ireg')
  }
  val (w,ireg) = uv.foldr (add) (List.empty[int],0)
  val msd = HIBYTE(ireg)
  val r = cons (msd) (w)
  //Console.print("u + v is ")
  //print(r)
  //Console.println("")
  r
}

// Substracts the unsigned radix 256 integer v[] from u[]
// yielding an unsigned integer. (second member the result)
// If the result is negative (wraps around), the first
// member of the result is -1, otherwise it is returned as 0
def mpsub (u:List[int], v:List[int]) : (int,List[int]) = {
  val uv = zip (u) (v)
  def sub (u:int,v:int) (w:List[int],ireg:int) : (List[int],int) = {
    val ireg' = 255 + (u - v) + HIBYTE(ireg)
    val w' = cons (LOBYTE(ireg')) (w)
    (w',ireg')
  }
  val (w,ireg) = uv.foldr (sub) (List.empty[int],256)
  val is = HIBYTE(ireg) - 1
  (is, w)
}

// Short addition: the integer iv (in the range 0 <= iv <= 255)
// is added to the unsigned radix 256 integer u
def mpsad (u:List[int], iv:int) : List[int] = {
  def add (u:int) (w:List[int],ireg:int) : (List[int],int) = {
    val ireg' = u + HIBYTE(ireg)
    val w' = cons (LOBYTE(ireg')) (w)
    (w',ireg')    
  }
  val (w,ireg) = u.foldr (add) (List.empty[int],256*iv)
  val msd = HIBYTE(ireg)
  cons (msd) (w)
}

// Short multiplication: the unsigned radix 256 integer u
// is multiplied by the integer iv (in the range 0 <= iv <= 255)
def mpsmu (u:List[int], iv:int) : List[int] = {
  def mul (u:int) (w:List[int],ireg:int) : (List[int],int) = {
    val ireg' = (u  * iv) + HIBYTE(ireg)
    val w' = cons (LOBYTE(ireg')) (w)
    (w',ireg')    
  }
  val (w,ireg) = u.foldr (mul) (List.empty[int],0)
  val msd = HIBYTE(ireg)
  cons (msd) (w)
}


// Short division: the unsigned radix 256 integer u
// is divided by the integer iv (in the range 0 <= iv <= 255)
// yielding a remainder ( int the range 0 <= iv <= 255)
// and an unsigned radix 256 integer
def mpsdv (u:List[int], iv:int) : (int,List[int]) = {
  def div (w:List[int],ir:int) (u:int) : (List[int],int) = {
    val i = (256 * ir) + u
    val w' = cons (i/iv) (w)
    val ir' = i % iv
    (w',ir')
  }
  val (w,ir) = u.foldl (div) (List.empty[int],0)
  (ir,w.reverse)
}


// Ones-complement negate the unsigned radix 256 integer u
def mpneg (u:List[int]) : List[int] = {
  def neg (u:int) (w:List[int],ireg:int) : (List[int],int) = {
    val ireg' = (255 - u) + HIBYTE(ireg)
    val w' = cons (LOBYTE(ireg')) (w)
    (w',ireg')    
  }
  val (w,ireg) = u.foldr (neg) (List.empty[int],256)
  w
}

// Left shift u
def mplsh (u:List[int],n:int) : List[int] = (u.ltail).take (n)

def mpmov (u:List[int], n:int) : List[int] = u.take (n)

// Uses Fast Fourier Transform to multiply
// the unsigned radix 256 integers u and v 
def mpmul (u:List[int], v:List[int]) : List[int] = {

  val RX : double = 256.0

  def p2 (n:int) = {
    def loop (n:int) (i:int) : int = if (i < n) loop (n) (i*2) else i
    loop (n) (1)
  }

  // Find the smallest usable power of two for the transform
  val ul = u.size
  val vl = v.size
  val mn = IMAX( ul, vl )
  val nn = 2 * (p2 (mn))
  // Map u and v to a Complex vector
  val a = (List.append (u) (List.copy[int] (nn-ul) (0))).map ( x | Complex.toComplex (x) )
  val b = (List.append (v) (List.copy[int] (nn-vl) (0))).map ( x | Complex.toComplex (x) )

  // Perform the convolution
  // First the two fourier transforms
  val a' = Fft.fft(a)
  val b' = Fft.fft(b)

  // Them multiply the complex results (Convolution)
  val c = b'.zipWith[Complex,Complex] ( b | a | b.mul (a)) (a')
  
  // Then do the inverse Fourier transform
  val d = Fft.scale(Fft.ifft(c))

  // Get the "real" vector
  val e = d.map ( c | c.re )
 
  
  // make a final pass to do all the carries
  def carries (b:double) (v:List[int],cy:double) : (List[int],double) = {
    val t = b + cy + 0.5                // The 0.5 allows for roundoff error
    val cy' = ((t / RX).toInt).toDouble
    val v' = List.cons ((t - (cy' * RX)).toInt) (v)
    (v',cy')
  }
  val (b'',cy) = e.foldr (carries) (List.empty[int],0.0)

  if ( cy >= RX  ) {
   System.error ("cannot happen in fftmul : cy is " + cy)
  } else {
    val msd = cy.toInt
    val w = (cons (msd) (b'')).take(ul+vl)

    w
  }
}

// v is interpreted as a radix 256 number with radix point after v[1]
// u is set to the most significant digits of its reciprocal,
// with radix point before u[1]
def mpinv (v:List[int], n':int) : List[int] = {

  val MF : int = 4
  val n = n' + 1

  // Use ordinary floating arithmetic to get an initial approximation
  val u' = fromDouble( 1.0 / toDouble(v.take (MF)) , n )

  // Newton-Raphson Functional Iteration
  // iterate Newton's rule to convergence
  def newton (u:List[int]) : List[int]= {

    // construct 2 - UV
    val s' = mpneg (mplsh(mpmov( mpmul(u,v), n),n))
    val s = cons ((s'.lhead) - 254) (s'.ltail)

    // multiply SU into U
    val u' = mplsh (mpmul(s,u), n)

    // if fractional part of S is not zero, it has not converged
    val cont = ((s'.ltail)).rtail.exist (x | (x != 0))

    if (cont) newton (u') else (u')
  }
  val r = newton (u')
  mplsh(r,n')

}

// Divides unsigned radix 256 integers u by v (with v.size <= u.size)
// yielding a remainder and a quotient
def mpdiv (u:List[int], v:List[int]) : (List[int], List[int]) = {

  val MACC : int = 3
  
  val n = u.size
  val m = v.size

  val s = mpinv (v, (n - m) + MACC)   // Set S = 1 / V
  val q = mpmov (mpmul(s,u), (n-m)+1) // Set Q = SU

  val rr = mplsh(mpmul(q,v),n)        // Multiply and substract to get the remainder
  val (is,r) = mpsub(u,mpmov(rr,n))
  
  if ( is != 0 ) System.error("MACC too small in mpdiv. is = " + is)
  else {
    (mpmov (r.drop(n-m),m),q)
  }
  
}




def mpsqrt (v:List[int], n:int) : (List[int], List[int]) = {

  val MF = 3

  val m = v.size
  val mm = IMIN(m,MF)

  // Use ordinary floating arithmetic to get an initial approximation
  val u' = fromDouble( 1.0 / sqrt (toDouble(v.take (MF))) , n )

  //def mplsh(u:List[int],n:int) : List[int] = mpmov(u:List[int],n:int)

  def newton (u:List[int]) : (List[int],List[int]) = {

    // Construct S = (3 - V*U^2)/2
    val r   = mplsh( mpmul(u,u), n )
    val s''' =  mplsh ( mpmul( r,v ), n )
    val s'' = mpneg ( s''' )
    val s'  = cons ((s''.lhead) - 253) (s''.ltail)

    val (ir,s) = mpsdv(s',2)

    // if fractional part of S is not zero, it has not converged
    val cont = ((s.ltail)).rtail.exist (x | (x != 0))

 
   
    if (cont) {
      // replace U by SU
      newton ( mplsh(mpmul(s,u),n) )
    } else {
      // Get square root from reciprocal and return
      val w = ( mplsh(mpmul(u,v),n) )
      (w,u)
    }

  }
  val r = newton (u')
  r
}

def mp2dfr (a:List[int]) : List[int] = {

  val IAZ = 48 // ascii (unicode) for 0

  val n = a.size

  val m = (2.408 * n.toDouble).toInt

  def loop (j:int) (a:List[int],s:List[int]) : (List[int],List[int]) = {
    if (j>m) (a,s)
    else {
      val a' = mpsmu( a, 10 )
 
      val s' = cons (a'.lhead) (s)
      val a'' = mplsh(a',n)
      loop (j+1) (a'',s')
    }
  }
  val (_,s) = loop (1) (a,List.empty[int])
  s.reverse
}

def printfloat(x:List[int]) : () = {
  Console.print("" + x.lhead + ".")
  mp2dfr(x.ltail).iterate ( c | Console.print(""+c) )
}

def mpi(n:int) : List[int] = {

  val t = cons (2) (List.copy[int] (n-1) (0)) // set T = 2

  val (x,_)    = mpsqrt(t,n)
  val pi       = mplsh(mpadd(t,x),n)
  val (sx,sxi) = mpsqrt(x,n)
  val y        = mpmov(sx,n)

  def loop (pi:List[int], sx:List[int], sxi:List[int], y:List[int]) :  (List[int], List[int], List[int], List[int]) = {

    val (_,x)      = mpsdv ( mpadd(sx,sxi).ltail, 2)    // x(i+1) = (sx + sxi)/2

    val (sx,sxi)   = mpsqrt (x,n)                                      // sx(i+1) = x^(1/2), sxi(i+1) = x^(-1/2)
    val t          = mplsh( mpadd ( mplsh(mpmul(y,sx),n) , sxi ), n )  // t = y(i) * sx(i+1) + sxi(i+1)


    val x' = x
    val x          = cons ((x'.lhead)+1) (x'.ltail) // x = x + 1
    val y' = y
    val y          = cons ((y'.lhead)+1) (y'.ltail) // y = y + 1
    
    val s          = mpinv (y,n)
    val y          = mpmov ( mpmul(t,s), n )  // y(i+1) = t / y

    val t          = mpmul (x,s)                    // t = ( (x(i+1) + 1) / (y(i) + 1) )
    
    // if fractional part of S is not zero, it has not converged
    val cont = ((t.take(n)).ltail).rtail.exist (x | (x != 0))

    if (cont) {
      val pi' = mplsh ( mpmul(pi,t) , n )
      //Console.print("\ntemp pi " )
      //printfloat(pi)
      //Console.println("")
      loop (pi',sx,sxi,y)
    } else {
      (pi, sx, sxi, y)
    }
  }
  val (pi',_,_,_) = loop (pi,sx,sxi,y)
  pi'
}


val Pi = mpi(1000)
Console.print("\npi is close to " )
printfloat(Pi)
Console.println("")

